package de.wbi.ppi.weka;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.TreeSet;

import weka.classifiers.Classifier;
import weka.classifiers.bayes.BayesNet;
import weka.classifiers.bayes.NaiveBayes;
import weka.classifiers.functions.Logistic;
import weka.classifiers.lazy.KStar;
import weka.classifiers.meta.FilteredClassifier;
import weka.classifiers.rules.JRip;
import weka.classifiers.rules.PART;
import weka.classifiers.trees.J48;
import weka.classifiers.trees.LADTree;
import weka.classifiers.trees.RandomForest;
import weka.core.Attribute;
import weka.core.Instances;
import weka.core.UnsupportedAttributeTypeException;
import weka.core.converters.ConverterUtils.DataSource;
import weka.filters.Filter;
import weka.filters.MultiFilter;
import weka.filters.unsupervised.attribute.ClassAssigner;
import weka.filters.unsupervised.attribute.RemoveType;
import weka.filters.unsupervised.attribute.Reorder;
import weka.filters.unsupervised.instance.RemoveWithValues;

public class FeatureLearner {
	
	static final String GROUND_TRUTH 	= "gt";
	static final String FOLD 			= "fold";
	static final String HARDNESS 		= "hardness";
	static final String CORPUS 			= "corpus";
	static final String ID 				= "pairlabel";
	static final String FIRST_INDEPENDENT_FEATURE = "wcount";
	
	static final String attributeRange(Instances data, String attributeName)
	{
		return String.valueOf(data.attribute(attributeName).index());
	}
	
	/** Class to split weka Instances based on value of an attribute (e.g., "fold").
	 *
	 */
	static class FoldWithValues
	{
		int attributeIndex;
		final RemoveWithValues removeWithValues;
		Filter outputFilter;
		Filter filter;
		String value;
		
		/** Mandatory default c'tor. 
		 *
		 *
		 */
		public FoldWithValues() throws Exception {
			
			this.attributeIndex = -1; // invalid
			this.removeWithValues = new RemoveWithValues();
			this.filter = this.removeWithValues;
		}
		
		/** Set the attribute index of the attribute to split on.
		 */
		public void setFoldAttribute(int index)	{
			
			this.attributeIndex = index;
			this.removeWithValues.setAttributeIndex(String.valueOf(attributeIndex+1));
		}
		
		/** Set the fold attribute's value (defines test set).
		 */
		public void setValue(String value)
		{
			this.value = value;
		}
		
		/** Find the fold attribute within a dataset.
		 */
		private Attribute getAttribute(Instances data) {
			Attribute att = data.attribute(attributeIndex);
			if (att == null)
				throw new NoSuchElementException("attribute #" + attributeIndex + " does not exist");
			if (!att.isNominal())
				throw new IllegalArgumentException("Attribute '" + att + "' is not nominal");
			return att;
		}
		
		/** Helper method to represent an Attribute as a single-element vector.
		 */
		private int[] toValueIndexArray(Attribute attribute)
		{
			int valueIndex = attribute.indexOfValue(value);
			if (valueIndex < 0)
				throw new NoSuchElementException("no such value: '" + value + "'");
			return new int[] {valueIndex};
		}
		
		/** Convenience method that allows to set an outputfilter applied after this filter.
		 */
		public void setOutputFilter(Filter outputFilter) {
			if (outputFilter == null) {
				this.filter = removeWithValues;
			} else {
				MultiFilter mf = new MultiFilter();
				mf.setFilters(new Filter[] {removeWithValues, outputFilter});
				this.filter = mf;
			}
		}
		
		public Filter getOutputFilter() {
			return outputFilter;
		}
				
		/** Split and return train Instances.
		 */
		Instances trainFolds(Instances data) throws Exception
		{
			return split(data, false);
		}
		
		
		/** Split and return test Instances.
		 */
		Instances testFold(Instances data) throws Exception
		{
			return split(data, true);
		}
		
		/** Split and return test or train instances.
		 */
		Instances split(Instances data, boolean isTest) throws Exception {
			removeWithValues.setInvertSelection(isTest);
			removeWithValues.setNominalIndicesArr(toValueIndexArray(getAttribute(data)));
			filter.setInputFormat(data);
			return Filter.useFilter(data, filter);
		}
	}
	
	/** Weka hlper method.
	 */
	static String toAttStr(Instances data, String attributeName)
	{
		return String.valueOf(data.attribute(attributeName).index()+1);		
	}
	
	/** Convenience method to represent an attribute's range as an Iterable. 
	 * Note: the implementation collects and sorts the values.
	 */
	static Iterable<String> toCollection(Instances data, String attributeName)
	{
		Collection<String> values = new TreeSet<String>();
		for (Enumeration<?> e = data.attribute(attributeName).enumerateValues(); e.hasMoreElements(); )
			values.add((String)e.nextElement());
		return values;
	}
	
	public static void main(String[] args) throws Exception {
		if (args.length < 2)
		{
			System.out.println("Usage: ... <classifier> <pairfeatures.arff>");
		}
		String classifierName = args[0];
		String dataFile = args[1];
		System.err.println("INFO: Loading '"+dataFile +"' ...");
		DataSource source = new DataSource(dataFile);
		Instances data = source.getDataSet();
		
		// setting class attribute if the data format does not provide this information
		// E.g., the XRFF format saves the class attribute information as well
//		if (data.classIndex() == -1)
		 //data.setClassIndex(data.numAttributes() - 1);
		// data.setClassIndex(data.attribute(GROUND_TRUTH).index());
		System.err.println("INFO: Instantiating filters and classifiers ...");		
		if (!data.attribute(FOLD).isNominal())
		{
			throw new UnsupportedAttributeTypeException("attribute '" + FOLD + "' is not nominal");
		}
		
		// Show folds
		Attribute foldAttr = data.attribute(FOLD);
		Enumeration<?> values = foldAttr.enumerateValues();
		for (int i = 0; values.hasMoreElements(); i++)
		{
			Object o = values.nextElement();
			System.err.println("" + i +": " + o + " | " + foldAttr.indexOfValue(o.toString()));
		}
		
		// filter
		//	Remove rm = new Remove();
		//	rm.setAttributeIndices(REMOVE_RANGE);  // remove 1st attribute
		
		
		
		// initialize some classifiers
		Collection<Classifier> classifiers = new ArrayList<Classifier>();
		if ("J48".equals(classifierName))
		{
			J48 c = new J48();
			//c.setUnpruned(true);        // using an unpruned J48
			classifiers.add(c);
		}
		else if ("BayesNet".equals(classifierName))
		{
			BayesNet c = new BayesNet();
			classifiers.add(c);
		}
		else if ("KStar".equals(classifierName))
		{
			KStar c = new KStar();
			classifiers.add(c);
		}
		else if ("Logistic".equals(classifierName))
		{
			Logistic c = new Logistic();
			classifiers.add(c);
		}
		else if ("JRip".equals(classifierName))
		{
			JRip c = new JRip();
			classifiers.add(c);
		}
		else if ("LADTree".equals(classifierName))
		{
			LADTree c = new LADTree();
			classifiers.add(c);
		}
		else if ("PART".equals(classifierName))
		{
			PART c = new PART();
			classifiers.add(c);
		}
		else if ("NaiveBayes".equals(classifierName))
		{
			NaiveBayes c = new NaiveBayes();
			classifiers.add(c);
		}
		else if ("RandomForest".equals(classifierName))
		{
			RandomForest c = new RandomForest();
			//c.setMaxDepth(10);
			c.setNumTrees(100);
			classifiers.add(c);
		}
		else
		{
			throw new NoSuchElementException("Unrecognized classifier '" + classifierName +"'");
		}
		
		
		// Reorder attributes so that GROUND_TRUTH becomes last as expected by most classifiers
		Reorder ro = new Reorder();
		ro.setAttributeIndices(
				toAttStr(data, ID) + "," + 
				toAttStr(data, FIRST_INDEPENDENT_FEATURE) + "-last," + 
				toAttStr(data, GROUND_TRUTH));
		System.err.println("INFO: Reordering range: '" + ro.getAttributeIndices() +"'");
		
		// set the 'class' flag on the last attribute (now GROUND_TRUTH)
		ClassAssigner ca = new ClassAssigner();
		ca.setClassIndex("last");
		
		// build a pipeline
		MultiFilter mf = new MultiFilter();
		mf.setFilters(new Filter[] {ro, ca});
		
		// debug
		printAttributes(data);
		
		FoldWithValues corpora = new FoldWithValues();
		corpora.setFoldAttribute(data.attribute(CORPUS).index());		
		
		FoldWithValues folds = new FoldWithValues();
		folds.setOutputFilter(mf);
		folds.setFoldAttribute(data.attribute(FOLD).index());
		
		// remove string attributes
		RemoveType removeId = new RemoveType();
		
		// classifier trainer
		FilteredClassifier fc = new FilteredClassifier();
		fc.setFilter(removeId);
		
		
		// iterate classifiers
		for (Classifier c : classifiers)
		{
			final String shortName = c.getClass().getSimpleName();
			final String description = c.getClass().getCanonicalName() + " " + Arrays.toString(c.getOptions());
			System.err.println("INFO: Using classifier: " + shortName +" (" + description +")");						 

			fc.setClassifier(c);
			// iterate over corpora
			for (String corpus : toCollection(data, CORPUS))
			{
				System.err.println("INFO: Corpus " + corpus);
				
				Writer out = new BufferedWriter(new FileWriter("prediction." + corpus + "." + shortName + ".sql"), 8*1024);
				
				// split on corpus, keep only crrent corpus
				corpora.setValue(corpus);
				Instances corpusInstances = corpora.testFold(data);
				
				// iterate folds
				for (String fold : toCollection(data, FOLD))
				{
					System.err.println("INFO: Starting fold " + fold);
					//	INSERT INTO ppiCV (parsertype, parser, corpus, fold, kernel, kernel_script) VALUES ('POS tagger', 'TextPRO (Version06?)', 'LLL', 8, 'SL', 'jsre; n=3 (n-gram), w=2 (window)');
					out.write("INSERT INTO ppiCV (parsertype, parser, corpus, fold, kernel, kernel_script) VALUES (NULL, NULL, '" + corpus + "', " + fold + ", $$" + shortName + "$$, $$" + description + "$$);\n");
					
					// split on fold value
					folds.setValue(fold);
					System.err.println("INFO: Generating splits ...");
					final Instances train = folds.trainFolds(corpusInstances);
					final Instances test = folds.testFold(corpusInstances);
					
					System.err.println("DEBUG: train[0] = " + train.instance(0));
					System.err.println("DEBUG: test[0] = " + test.instance(0));
					
					// train and make predictions
					System.err.println("INFO: Training classifier ...");
					fc.buildClassifier(train);
					System.err.println("INFO: Applying classifier ...");
					for (int i = 0; i < test.numInstances(); i++)
					{
						String id = test.attribute(0).value(i);
						double pred = fc.classifyInstance(test.instance(i));
						String predLabel = test.classAttribute().value((int) pred);
						
						System.out.print("ID: " + id);
						System.out.print(", actual: " + test.classAttribute().value((int) test.instance(i).classValue()));
						System.out.println(", predicted: " + predLabel);
						// INSERT INTO ppiCVoutput (expId, pair, output, prediction) VALUES (currval('ppiCV_ppiCVid_seq'), 'LLL.d3.s0.p0', '0', -0.423565281318);
						out.write("INSERT INTO ppiCVoutput (expId, pair, output, prediction) VALUES (currval('ppiCV_ppiCVid_seq'), $$" + id + "$$, '"+ predLabel +"', "+ pred +");\n");
					}
				}
				out.close();
			}
		}
	}
	
	/** Helper method.
	 */
	static void printAttributes(Instances data)
	{
		for (int i = 0; i < data.numAttributes(); i++)
		{
			Attribute a = data.attribute(i);
			System.out.println("ATTRIBUTE: " + i + "\t" + a.name() + "\t" + a.isNumeric());
		}
		
	}
}
