#!/usr/bin/perl
# Sript to parse corpora in XML, and write them to relational database.
# Input format:
# 
# <corpus source="LLL">
#   <document id="LLL.d0" origId="11069677">
#     <sentence id="LLL.d0.s0" origId="11069677-3" text="In vivo studies of the activity of four of the kinases, KinA, KinC, KinD (ykvD) and KinE (ykrQ), using abrB transcription as an indicator of Spo0A~P level, revealed that KinC and KinD were responsible for Spo0A~P production during the exponential phase of growth in the absence of KinA and KinB.">
#       <entity charOffset="170-173" id="LLL.d0.s0.e0" origId="29" text="KinC" type="agent" />
#       <entity charOffset="179-182" id="LLL.d0.s0.e1" origId="31" text="KinD" type="agent" />
#       <entity charOffset="205-211" id="LLL.d0.s0.e2" origId="35" text="Spo0A~P" type="target" />
#       <pair e1="LLL.d0.s0.e0" e2="LLL.d0.s0.e1" id="LLL.d0.s0.p0" interaction="False" />
#       <pair e1="LLL.d0.s0.e0" e2="LLL.d0.s0.e2" id="LLL.d0.s0.p1" interaction="True" />
#       <pair e1="LLL.d0.s0.e1" e2="LLL.d0.s0.e2" id="LLL.d0.s0.p2" interaction="True" />
# ...
#       <tokenizations>
#	       <tokenization tokenizer="Charniak-Lease">
#   	       <token POS="IN" charOffset="0-1" id="clt_1" text="In" />
#              <token POS="FW" charOffset="3-6" id="clt_2" text="vivo" />
#              <token POS="NNS" charOffset="8-14" id="clt_3" text="studies" />
# ...

use strict;
use warnings;

use XML::Parser;
use File::Basename;
use FileHandle;
use IPC::Open3;
#use Data::Dumper;

sub porter($);
require 'porter.pm';

# main

if ($#ARGV < 2)
{
	print "Usage $0 <path/to/geniatagger> <tokenization> <corpus.xml> ...\n";
	exit -1;
}

# path to geniatagger executable
my $geniatagger = shift (@ARGV); # '../Parsing/geniatagger/geniatagger-3.0.1/geniatagger';

# Tokens will be taken from this tokenization
my $tokenizer = shift (@ARGV); # 'Charniak-Lease';


# run GENIA Tagger
my $cmd = '( cd ' . dirname($geniatagger) . ' ; ./' . basename($geniatagger) . ' -nt )'; 
print STDERR "Executing `$cmd'\n";
my $pid = open3(*Writer, *Reader, *Error, $cmd );
print STDERR "GENIA $_ of 4: " . <Error> foreach (1..4);


my $parser = new XML::Parser;

$parser->setHandlers(
	Start => \&startElement,
	End => \&endElement,
	Char => \&characterData,
	Default => \&default);

# quoting
my $q = '$XX$';


# current values
my ($corpus, $document, $sentence, $entity, $pair, $collectTokens, @tokens);

foreach my $xmlfile (@ARGV)
{
	die "Cannot find file '$xmlfile'"
		unless -f $xmlfile;

	$parser->parsefile($xmlfile);
}


# subs

sub startElement 
{

	my( $parseinst, $element, %attrs ) = @_;
	SWITCH: {
		# <corpus source="LLL">
		if ($element eq 'corpus') {
			$corpus = $attrs{'source'};
#			print "INSERT INTO corpus VALUES ($q$corpus$q);\n";
            print STDERR "Corpus '$corpus'\n";
			last SWITCH;
		}
		# <document id="LLL.d0" origId="11069677">
		if ($element eq 'document') {
			$document = $attrs{'id'} or warn();
			my $origId = $attrs{'origId'} || "";# or warn();
			my ($position) = ($document =~ m/(\d+)$/) or warn();

#			print "INSERT INTO doc VALUES ($q$document$q, $q$corpus$q, $position, $q$origId$q);\n";
            print STDERR "Doc: $document, Corpus: $corpus, Pos: $position, OrigId: $origId\n";
			last SWITCH;
		}
		# <sentence id="LLL.d0.s0" origId="11069677-3" text="In vivo studies of the activity of four of the kinases, KinA, KinC, KinD (ykvD) and KinE (ykrQ), using abrB transcription as an indicator of Spo0A~P level, revealed that KinC and KinD were responsible for Spo0A~P production during the exponential phase of growth in the absence of KinA and KinB.">
		if ($element eq 'sentence') {
			$sentence = $attrs{'id'} or warn();
			my $origId = $attrs{'origId'} || "";# or warn();
			my $text = $attrs{'text'} or warn();
			my ($position) = ($sentence =~ m/(\d+)$/) or warn();

#			print "INSERT INTO sentence VALUES ($q$sentence$q, $q$document$q, $position, $q$origId$q, $q$text$q);\n";
			last SWITCH;
		}
		# <entity charOffset="205-211" id="LLL.d0.s0.e2" origId="35" text="Spo0A~P" type="target" />
# 		if ($element eq 'entity') {
# 			$entity = $attrs{'id'} or warn("entity:id");
# 			my $origId = $attrs{'origId'} || "";# or warn("entity:origId");
# 			my $text = $attrs{'text'} or warn();
# 			my $type = $attrs{'type'} || "";# warn();
# 			my ($position) = ($entity =~ m/(\d+)$/) or warn();
# 			my ($begin, $end, $begin2, $end2) = ($attrs{'charOffset'} =~ m/^(\d+)-(\d+)(?:,(\d+)-(\d+))?(.+)?$/) or warn($attrs{'charOffset'});
# 			warn($attrs{'charOffset'}) if defined $5;
# 			if (!defined $begin2) {$begin2='NULL';$end2='NULL';}
# 			print "INSERT INTO entity VALUES ($q$entity$q, $q$sentence$q, $position, $q$origId$q, $q$text$q, $q$type$q, $begin, $end, $begin2, $end2);\n";
# 			last SWITCH;
# 		}
		#       <pair e1="LLL.d0.s0.e0" e2="LLL.d0.s0.e1" id="LLL.d0.s0.p0" interaction="False" />
# 		if ($element eq 'pair') {
# 			$pair = $attrs{'id'} or warn();
# 			my $e1 = $attrs{'e1'} or warn();
# 			my $e2 = $attrs{'e2'} or warn();
# 			my ($interaction) = $attrs{'interaction'} or warn();
# 
# 			my $groundTruth;
# 			$groundTruth = 0 if $interaction =~ /False/i;
# 			$groundTruth = 1 if $interaction =~ /True/i;
# 			warn() unless defined $groundTruth;
# 
# 			print "INSERT INTO pair VALUES ($q$pair$q, $q$e1$q, $q$e2$q, $q$groundTruth$q);\n";
# 			last SWITCH;
# 		}
        if ($element eq 'tokenization') {
            $collectTokens = 1 if $attrs{'tokenizer'} eq $tokenizer;
            last SWITCH;
        }
        if ($element eq 'token') {
            last SWITCH unless $collectTokens;
            # <token POS="IN" charOffset="0-1" id="clt_1" text="In" /> 
            my $token = \%attrs;
            push(@tokens, $token);
        }

	}
}

sub endElement 
{
    my( $parseinst, $element ) = @_;
    if ($element eq 'tokenization')
    {
        if ($collectTokens)
        {
            geniaAnalyze(\@tokens);
            
            for my $i (0..$#tokens)
            {
                my $token = $tokens[$i];
                my $text = $token->{'text'};
                my ($begin, $end) = ($token->{'charOffset'} =~ /^(\d+)\D+(\d+)$/);
                defined ($begin) && defined($end) 
                    or die "Unable to parse charOffset: '" . $token->{'charOffset'} . "'";
                my $posTag = $token->{'POS'};
                my $lemma = $token->{'geniaLemma'};
                my $stem = porter($text);
                print 'INSERT INTO sentenceToken (sentence, position, beginOffset, endOffset, tokenText, lemma, stem, posTag)' .
                        " VALUES ($q$sentence$q, $i, $begin, $end, $q$text$q, $q$lemma$q, $q$stem$q, $q$posTag$q);\n";
            }
        }

        # reset
        $collectTokens = 0;
        @tokens = ();
    }
}

sub characterData 
{
	my( $parseinst, $data ) = @_;
}

sub default 
{	
	my( $parseinst, $data ) = @_;
        # do nothing, but stay quiet
}

# Takes a sentence (arrayref) as input, and sets the key 'geniaLemma' for each array element
sub geniaAnalyze($)
{
    my $tokens = shift;
    
    #print STDERR Dumper($tokens);

    # re-assemble a sentence
    my @tokenTexts;
    for my $token (@$tokens)
    {
        my $tokenText = $token->{'text'};
        $tokenText =~ s/\s+/_/g;
        push(@tokenTexts, $tokenText);
    }
    my $sentence = join(' ',  @tokenTexts);
    print STDERR "Sending: '$sentence'\n";
    print Writer "$sentence\n";

    # read output from Genia (a line per each token plus an empty line)
    for my $i (0..$#{$tokens}+1)
    {
        ! eof(Reader)
            or die "premature end of output.";
        defined( $_ = <Reader> )
            or die "readline failed: $!";
        my $line = $_;
        chomp($line);
        #print "GOT: $line\n";
        if ($i <= $#{$tokens}) ## not the end of sentence newline
        {
            # KinB       KinB    NN      I-NP    B-protein
            my ($text, $lemma, $posTag, $chunkTag, $neTag) = split(/\s+/, $line);
            $tokens->[$i]->{'geniaLemma'} = $lemma;
            # $tokens->[$i]->{'geniaPosTag'} = $posTag;
            # ...
        }
    }
}
