DROP TYPE IF EXISTS PosNegCorpusResult CASCADE;

CREATE TYPE PosNegCorpusResult AS 
(
	ppicvid INTEGER,
	corpus TEXT,
	threshold DOUBLE PRECISION,
	f_measure DOUBLE PRECISION,
	outputposrate DOUBLE PRECISION,
	trainposrate DOUBLE PRECISION	
);


drop function if exists PosNegCorpus(Text);
create or replace function PosNegCorpus(corpus Text) returns PosNeg as 
$$
select cast(t.pos as integer) as pos,cast(t.neg as integer) as neg,
(case when t.neg+t.pos = 0 then null else cast(t.pos as double precision)/cast(t.neg+t.pos as double precision) end) as ratio from
(
select sum(case when c.groundTruth = true then 1 else 0 end) as pos,
	sum(case when c.groundTruth = false then 1 else 0 end) as neg
from pair c
	inner join entity e1 on e1.entityLabel = c.entity1
	inner join entity e2 on e2.entityLabel = c.entity2
	inner join sentence s on s.label = e1.sentence
	inner join doc d on d.doclabel = s.doc
where d.corpus = $1) as t
$$
language SQL stable returns null on null input;
;
comment on function PosNegCorpus(Text) is 
'Calculates the number of positive and negative training examples for a corpus and the ratio of posrate (pos/pos+neg).';


drop function if exists PosNegOutsideCorpus(Text,integer);
create or replace function PosNegOutsideCorpus(corpus Text) returns PosNeg as 
$$
select cast(t.pos as integer) as pos,cast(t.neg as integer) as neg,
(case when t.neg+t.pos = 0 then null else cast(t.pos as double precision)/cast(t.neg+t.pos as double precision) end) as ratio from
(
select sum(case when c.groundTruth = true then 1 else 0 end) as pos,
	sum(case when c.groundTruth = false then 1 else 0 end) as neg
from pair c
	inner join entity e1 on e1.entityLabel = c.entity1
	inner join entity e2 on e2.entityLabel = c.entity2
	inner join sentence s on s.label = e1.sentence
	inner join doc d on d.doclabel = s.doc
where d.corpus <> $1 and d.corpus is not null) as t
$$
language SQL stable returns null on null input;
;
comment on function PosNegOutsideCorpus(Text) is 
'Calculates the number of positive and negative training examples for the ensemble of all corpora outside the given corpus
and the ratio of posrate (pos/pos+neg).';

drop function if exists PosNegOutputCorpus(integer);
create or replace function PosNegOutputCorpus(ppiclid integer) returns PosNeg as 
$$
select cast(t.pos as integer) as pos,cast(t.neg as integer) as neg,
(case when t.neg+t.pos = 0 then null else cast(t.pos as double precision)/cast(t.neg+t.pos as double precision) end) as ratio from
(
select sum(case when c.output = true then 1 else 0 end) as pos,
	sum(case when c.output = false then 1 else 0 end) as neg
from ppicloutput c
	inner join ppicl p on c.expid=p.ppiclid 
where p.ppiclid = $1) as t
$$
language SQL stable returns null on null input;
;
comment on function PosNegOutputCorpus(integer) is 
'Calculates the number of positive and negative predictions for a given corpus and the ratio of posrate (pos/pos+neg).
The experiment is specified by ppiclid.';

drop function if exists PosNegOutputThresholdCorpus(integer,double precision);
create or replace function PosNegOutputThresholdCorpus(ppiclid integer,threshold double precision) returns PosNeg as 
$$
select cast(t.pos as integer) as pos,cast(t.neg as integer) as neg,
(case when t.neg+t.pos = 0 then null else cast(t.pos as double precision)/cast(t.neg+t.pos as double precision) end) as ratio from
(
select sum(case when c.prediction>=$2 then 1 else 0 end) as pos,
	sum(case when c.prediction>=$2 then 0 else 1 end) as neg
from ppicloutput c
	inner join ppicl p on c.expid=p.ppiclid 
where p.ppiclid = $1) as t
$$
language SQL stable returns null on null input;
;
comment on function PosNegOutputThresholdCorpus(integer,double precision) is 
'Calculates the number of positive and negative predictions for a corpus at a given threshold and the ratio of posrate (pos/pos+neg).
The experiment is specified by ppiclid.
The threshold should be in the range of [0,1]';

drop function if exists PosNegOutputThresholdCLF1(integer,double precision,double precision,integer);
create or replace function PosNegOutputThresholdCLF1(
ppiclid integer,lower double precision,upper double precision,step integer
) returns setof PosNegCorpusResult as
$$
select b.ppiclid,
b.corpus,
t.threshold,
cl_my_f1_threshold(b.ppiclid,t.threshold),
(PosNegOutputThresholdCorpus(b.ppiclid,t.threshold)).posrate as outputposrate, 
(PosNegOutsideCorpus(b.corpus)).posrate as trainposrate
from (select ppiclid,corpus from ppicl p 
where ppiclid=$1) as b
cross join (select (cash)/cast($4 as double precision)*($3-$2)+$2 as threshold from generate_series(0,$4) cash) as t
$$
language SQL stable returns null on null input;
;
comment on function PosNegOutputThresholdCLF1(integer,double precision,double precision,integer) is 
'For a given ppiclid the function calculates the thresholded F1 using cl_my_f1_threshold function
The thresholds are calculated from lower($2) to upper($3) extremes with step $4.
Usually, the threshold should be in the range of [0,1]';

drop function if exists PosNegOutputOptimalThresholdCL(integer,double precision,double precision,integer);
create or replace function PosNegOutputOptimalThresholdCL(
ppiclid integer,lower double precision,upper double precision,step integer
) returns PosNegCorpusResult as
$$
select * from PosNegOutputThresholdCLF1($1,$2,$3,$4) t where
abs(t.outputposrate-t.trainposrate) = (select min(abs(t1.outputposrate-t1.trainposrate)) from PosNegOutputThresholdCLF1($1,$2,$3,$4) as t1) 
limit 1
$$
language SQL stable returns null on null input;
;
comment on function PosNegOutputOptimalThresholdCL(integer,double precision,double precision,integer) is 
'For a given ppiclid the function selects the threshold with the best F1 using the PosNegOutputThresholdCLF1 function. 
The thresholds are calculated from lower($2) to upper($3) extremes with step $4.
Usually, the threshold should be in the range of [0,1]';

-- select *,cl_add_threshold_version(p.ppiclid,(PosNegOutputOptimalThresholdCL(p.ppiclid,0,1,20)).threshold) as newppiclid 
-- from ppicl p
-- where kernel = 'Kim-shallow' and kernel_script = 'shallow -w0 1 -w1 16 -nu 0.5' and c=0.5 and corpus='BioInfer'
-- ;
-- -- 
