DROP TYPE IF EXISTS PosNeg CASCADE;

CREATE TYPE PosNeg AS 
(
	pos INTEGER,
	neg INTEGER,
	posrate DOUBLE PRECISION
);
 
DROP TYPE IF EXISTS PosNegFoldResult CASCADE;

CREATE TYPE PosNegFoldResult AS 
(
	ppicvid INTEGER,
	ppicvfoldsid INTEGER,
	corpus TEXT,
	fold INTEGER,
	threshold DOUBLE PRECISION,
	f_measure DOUBLE PRECISION,
	outputposrate DOUBLE PRECISION,
	trainposrate DOUBLE PRECISION	
);

drop function if exists PosNegInsideFold(Text,integer);
create or replace function PosNegInsideFold(corpus Text, fold integer) returns PosNeg as 
$$
select cast(t.pos as integer) as pos,cast(t.neg as integer) as neg,
(case when t.neg+t.pos = 0 then null else cast(t.pos as double precision)/cast(t.neg+t.pos as double precision) end) as ratio from
(
select sum(case when c.groundTruth = true then 1 else 0 end) as pos,
	sum(case when c.groundTruth = false then 1 else 0 end) as neg
from pair c
	inner join entity e1 on e1.entityLabel = c.entity1
	inner join entity e2 on e2.entityLabel = c.entity2
	inner join sentence s on s.label = e1.sentence
	inner join doc d on d.doclabel = s.doc
where d.corpus = $1 and d.fold = $2) as t
$$
language SQL stable returns null on null input;
;
comment on function PosNegInsideFold(Text,integer) is 
'Calculates the number of positive and negative training examples inside a given fold and the ratio of posrate (pos/pos+neg).
The fold is specified by (corpus,fold).';


drop function if exists PosNegOutsideFold(Text,integer);
create or replace function PosNegOutsideFold(corpus Text, fold integer) returns PosNeg as 
$$
select cast(t.pos as integer) as pos,cast(t.neg as integer) as neg,
(case when t.neg+t.pos = 0 then null else cast(t.pos as double precision)/cast(t.neg+t.pos as double precision) end) as ratio from
(
select sum(case when c.groundTruth = true then 1 else 0 end) as pos,
	sum(case when c.groundTruth = false then 1 else 0 end) as neg
from pair c
	inner join entity e1 on e1.entityLabel = c.entity1
	inner join entity e2 on e2.entityLabel = c.entity2
	inner join sentence s on s.label = e1.sentence
	inner join doc d on d.doclabel = s.doc
where d.corpus = $1 and d.fold <> $2 and d.fold is not null) as t
$$
language SQL stable returns null on null input;
;
comment on function PosNegOutsideFold(Text,integer) is 
'Calculates the number of positive and negative training examples outside a given fold and the ratio of posrate (pos/pos+neg).
The fold is specified by (corpus,fold).';

drop function if exists PosNegOutputFold(integer);
create or replace function PosNegOutputFold(ppicvid integer) returns PosNeg as 
$$
select cast(t.pos as integer) as pos,cast(t.neg as integer) as neg,
(case when t.neg+t.pos = 0 then null else cast(t.pos as double precision)/cast(t.neg+t.pos as double precision) end) as ratio from
(
select sum(case when c.output = true then 1 else 0 end) as pos,
	sum(case when c.output = false then 1 else 0 end) as neg
from ppicvoutput c
	inner join ppicv p on c.expid=p.ppicvid 
where p.ppicvid = $1) as t
$$
language SQL stable returns null on null input;
;
comment on function PosNegOutputFold(integer) is 
'Calculates the number of positive and negative predictions for a given fold and the ratio of posrate (pos/pos+neg).
The fold is specified by the identifier of the experiment (ppicvid).';

drop function if exists PosNegOutputThresholdFold(integer,double precision);
create or replace function PosNegOutputThresholdFold(ppicvid integer,threshold double precision) returns PosNeg as 
$$
select cast(t.pos as integer) as pos,cast(t.neg as integer) as neg,
(case when t.neg+t.pos = 0 then null else cast(t.pos as double precision)/cast(t.neg+t.pos as double precision) end) as ratio from
(
select sum(case when c.prediction>=$2 then 1 else 0 end) as pos,
	sum(case when c.prediction>=$2 then 0 else 1 end) as neg
from ppicvoutput c
	inner join ppicv p on c.expid=p.ppicvid 
where p.ppicvid = $1) as t
$$
language SQL stable returns null on null input;
;
comment on function PosNegOutputThresholdFold(integer,double precision) is 
'Calculates the number of positive and negative predictions for a given fold at a given threshold and the ratio of posrate (pos/pos+neg).
The fold is specified by the identifier of the experiment (ppicvid).
The threshold should be in the range of [0,1]';

drop function if exists PosNegOutputThresholdF1(integer,double precision,double precision,integer);
create or replace function PosNegOutputThresholdF1(
ppicvid integer,lower double precision,upper double precision,step integer
) returns setof PosNegFoldResult as
$$
select b.ppicvid,
b.ppicvfoldsid,
b.corpus,
b.fold,
t.threshold,
my_f1_threshold(b.ppicvid,t.threshold),
(PosNegOutputThresholdFold(b.ppicvid,t.threshold)).posrate as outputposrate, 
(PosNegOutsideFold(b.corpus,b.fold)).posrate as trainposrate
from (select ppicvid,ppicvfoldsid,corpus,fold from ppicv p 
where ppicvid=$1) as b
cross join (select (cash)/cast($4 as double precision)*($3-$2)+$2 as threshold from generate_series(0,$4) cash) as t
$$
language SQL stable returns null on null input;
;
comment on function PosNegOutputThresholdF1(integer,double precision,double precision,integer) is 
'For a given ppicvid the function calculates the thresholded F1 using my_f1_threshold function
The thresholds are calculated from lower($2) to upper($3) extremes with step $4.
Usually, the threshold should be in the range of [0,1]';


drop function if exists PosNegOutputOptimalThreshold(integer,double precision,double precision,integer);
create or replace function PosNegOutputOptimalThreshold(
ppicvid integer,lower double precision,upper double precision,step integer
) returns PosNegFoldResult as
$$
select * from PosNegOutputThresholdF1($1,$2,$3,$4) t where
abs(t.outputposrate-t.trainposrate) = (select min(abs(t1.outputposrate-t1.trainposrate)) from PosNegOutputThresholdF1($1,$2,$3,$4) as t1) 
limit 1
$$
language SQL stable returns null on null input;
;
comment on function PosNegOutputOptimalThreshold(integer,double precision,double precision,integer) is 
'For a given ppicvid the function selects the threshold with the best F1 using the PosNegOutputThresholdF1 function. 
The thresholds are calculated from lower($2) to upper($3) extremes with step $4.
Usually, the threshold should be in the range of [0,1]';

-- select *,add_threshold_version(p.ppicvid,(PosNegOutputOptimalThreshold(p.ppicvid,0,1,20)).threshold) as newppicvid 
-- from ppicv p
-- where kernel = 'Kim-shallow' and kernel_script = 'shallow -w0 1 -w1 16 -nu 0.5 (thresholded)' and c=0.5
-- ;
-- -- 

-- 
-- select *,my_f1_threshold(b.ppicvid,t.threshold),
-- (PosNegOutputThresholdFold(b.ppicvid,t.threshold)).posrate as outputposrate, 
-- (PosNegOutsideFold('LLL',2)).posrate as trainposrate
-- from (select ppicvid from ppicv p 
-- where corpus='LLL' and fold = 2 limit 1) as b
-- cross join (select (cash)/10.0 as threshold from generate_series(0,10) cash) as t
-- ;
-- -- 


