-------------------------------------- README ---------------------------------------
-- Requires best_CV_settings.sql to be already executed				   --
--										   --
-- First part defines functions to select the best CL settings                     --
-- Second part defines a function for thresholding CL experiments 		   --
-- Third part selects the best settings and creates table for the Best_CL_settings --
-------------------------------------------------------------------------------------



-- First part

drop function if exists results_CL(Text);

create or replace function results_CL
(
IN kernel Text
) 
returns setof kernelsettingresult as
$$
SELECT 
kernel,avg(auc) AS auc, avg(precision_) AS prec, avg(recall) AS recall, avg(f_measure) AS f_measure,
parsertype,parser,
kernel_script,
c_normal,avg(c_spec),coalesce(c_normal,avg(c_spec)) as c,j,k,lmin,lmax,match_,normalized,input_format,led,
cast(count(*) as integer) as cnt 
from ppiCLwithC
where kernel = $1 
GROUP BY kernel,parsertype,parser,c_normal,
kernel_script,j,k,lmin,lmax,match_,normalized,input_format,led
$$
LANGUAGE SQL 
IMMUTABLE RETURNS NULL ON NULL INPUT
;
comment on function results_CL(Text) is 
'Calculates for a kernel the average goodness of the kernel measured accross the available corpora.
The parameter is the name of the kernel. 
Calculation is based on ppiCLwithC (CL evaluation with extended C values).';

drop function if exists best_CL_setting(Text);

create or replace function best_CL_setting
(
IN kernel Text
) 
returns setof kernelsettingresult as
$$
select * from results_CL($1) ORDER BY f_measure DESC limit 1;
$$
LANGUAGE SQL 
IMMUTABLE RETURNS NULL ON NULL INPUT
;
comment on function best_CL_setting(Text) is 
'Calculates the best setting according to F1 using results_CL function.
The parameter is the name of the kernel.';

drop function if exists best_CL_ids(Text);

create or replace function best_CL_ids
(
IN kernel Text
) 
returns setof integer as
$$
SELECT 
ppiclid
from ppiCLwithC p
join best_CL_setting($1) s on
p.kernel is not distinct from s.kernel and
p.kernel_script is not distinct from s.kernel_script and
p.j is not distinct from s.j and 
p.k is not distinct from s.k and 
p.lmin is not distinct from s.lmin and
p.lmax is not distinct from s.lmax and
p.match_ is not distinct from s.match_ and
p.normalized is not distinct from s.normalized and
p.input_format is not distinct from s.input_format and
p.led is not distinct from s.led and
s.c_normal is not distinct from p.c_normal
$$
LANGUAGE SQL 
IMMUTABLE RETURNS NULL ON NULL INPUT
;
comment on function best_CL_ids(Text) is 
'Returns the ppiclids that belong to the best CL setting (F-measure based). 
The number of returned rows should be equal to the number of corpora.
The parameter is the name of the kernel.';


-- Second part

DROP FUNCTION IF EXISTS cl_add_threshold_version(integer,double precision);
CREATE FUNCTION cl_add_threshold_version
(
  myppiclid int,
  th double precision
)
RETURNS INT AS
$$
DECLARE
    mykernel text;
    mycorpus text;
    myparsertype text;
    myparser text;
    mykernelscript text;
    myc double precision;
    myj double precision;
    myled Text;
    myauc double precision;

    combinedKernel text;
    combinedKernelScript text;
    newppiclid integer;

BEGIN
  SELECT INTO mykernel kernel FROM ppicl WHERE ppiclid=myppiclid;
  SELECT INTO mycorpus corpus FROM ppicl WHERE ppiclid=myppiclid;
  SELECT INTO myparser parser FROM ppicl WHERE ppiclid=myppiclid;
  SELECT INTO myparsertype parsertype FROM ppicl WHERE ppiclid=myppiclid;
  SELECT INTO mykernelscript kernel_script FROM ppicl WHERE ppiclid=myppiclid;
  SELECT INTO myc c FROM ppicl WHERE ppiclid=myppiclid;  
  SELECT INTO myj j FROM ppicl WHERE ppiclid=myppiclid;
  SELECT INTO myled led FROM ppicl WHERE ppiclid=myppiclid;
  SELECT INTO myauc auc FROM ppicl WHERE ppiclid=myppiclid; -- auc does not change!
      
  SELECT INTO combinedKernel (SELECT mykernel || ' (thresholded)');
  SELECT INTO combinedKernelScript (SELECT mykernelscript || ' (thresholded)');
   
  IF EXISTS (SELECT ppiclid FROM ppicl WHERE 
	kernel is not distinct from combinedKernel 
	and corpus is not distinct from mycorpus 
	and parsertype is not distinct from myparsertype 
	and c is not distinct from myc 
	and j is not distinct from myj 
	and led is not distinct from myled
	and parser is not distinct from myparser 
	and kernel_script is not distinct from combinedKernelScript) 
    THEN 
    RETURN (
    SELECT ppiclid FROM ppicl WHERE 
	kernel is not distinct from combinedKernel 
	and corpus is not distinct from mycorpus 
	and parsertype is not distinct from myparsertype 
	and c is not distinct from myc 
	and j is not distinct from myj 
	and led is not distinct from myled
	and parser is not distinct from myparser 
	and kernel_script is not distinct from combinedKernelScript); /* already exists */
  END IF;

  INSERT INTO ppicl (corpus,parsertype,parser,kernel, normalized,c,j,kernel_script,auc,forced_threshold) VALUES 
 	(mycorpus, myparsertype,myparser,combinedKernel,'t',myc,myj,combinedKernelScript,myauc,th);
	
  SELECT INTO newppiclid currval('ppiCL_ppiCLid_seq');

  INSERT INTO
    ppicloutput (expId, pair, output, prediction)
  SELECT
    newppiclid         as expId, 
    sub.pair	      as pair,
    (case when sub.prediction>=$2 then TRUE else FALSE end) as output,
    prediction
  FROM
  (SELECT
     prediction,
     pair
     FROM
    ppiCLoutput
    WHERE expid = myppiclid
   ) as sub
;

  UPDATE ppicl set 
	tp=s.tp,
	fp=s.fp, 
	fn=s.fn, 
	tn=s.tn,
	total=s.n,
	precision_=s.p,
	recall=s.r,
	f_measure=s.f
  FROM cl_get_statistics(newppiclid) as s
  where ppiclid=newppiclid;

  RETURN newppiclid; /* just created */
  
END
$$
LANGUAGE 'plpgsql'
;
comment on function cl_add_threshold_version(integer, double precision) is 
'Adds a new row into ppicl (and many into ppicloutput) with thresholding on the experiment given by ppiclid.
Warning: if the thresholded version already exists, its ppiclid is returned';

-- Third part

drop table if exists Best_CL;

create table Best_CL as 
select * from ppicl p where ppiclid in 
(select * from best_CL_ids('SL')
union all
select * from best_CL_ids('ST')
union all
select * from best_CL_ids('SST')
union all
select * from best_CL_ids('SpT')
union all
select * from best_CL_ids('kBSPS')
union all
select * from best_CL_ids('cosine')
union all
select * from best_CL_ids('edit')
union all
select * from best_CL_ids('APG')
union all
select * from best_CL_ids('Kim-shallow (thresholded)')
union all
select * from best_CL_ids('Kim-lexical (thresholded)')
union all
select * from best_CL_ids('Kim-syntactic (thresholded)')
union all
select * from best_CL_ids('Kim-combined (thresholded)')
);

create or replace view best_CL_output as 
select p.kernel,o.* from best_CL p
join ppicloutput o on o.expid=p.ppiclid
;
