-------------------------------------- README ---------------------------------------
-- First part defines functions to select the best CL settings                     --
-- Second part defines a function for thresholding CL experiments 		   --
-- Third part selects the best settings and creates table for the Best_CL_settings --
-------------------------------------------------------------------------------------

-- First part

DROP TYPE IF EXISTS kernelsettingresult CASCADE;

CREATE TYPE kernelsettingresult AS 
(
	kernel TEXT,
	auc DOUBLE PRECISION,
	prec DOUBLE PRECISION, 
	recall DOUBLE PRECISION, 
	f_measure DOUBLE PRECISION,
	parsertype TEXT,
	parser TEXT,
	kernel_script TEXT,
	c_normal DOUBLE PRECISION,
	c_spec DOUBLE PRECISION,
	c DOUBLE PRECISION,
	j DOUBLE PRECISION,
	k INTEGER,
	lmin INTEGER,
	lmax INTEGER,
	match_ TEXT,
	normalized BOOLEAN,
	input_format TEXT,
	led TEXT,
	cnt INTEGER
);
 

drop function if exists results_CV(Text);

create or replace function results_CV
(
IN kernel Text
) 
returns setof kernelsettingresult as
$$
SELECT 
kernel,avg(auc) AS auc, avg(precision_) AS prec, avg(recall) AS recall, avg(f_measure) AS f_measure,
parsertype,parser,
kernel_script,
c_normal,avg(c_spec),coalesce(c_normal,avg(c_spec)) as c,j,k,lmin,lmax,match_,normalized,input_format,led,
cast(count(*) as integer) as cnt 
from foldCVcombined 
where kernel = $1 
GROUP BY kernel,parsertype,parser,c_normal,
kernel_script,j,k,lmin,lmax,match_,normalized,input_format,led
$$
LANGUAGE SQL 
IMMUTABLE RETURNS NULL ON NULL INPUT
;
comment on function results_CV(Text) is 
'Calculates for a kernel the average goodness of the kernel measured accross the available corpora.
The parameter is the name of the kernel. 
Calculation is based on foldCVcombined that contains both experiments obtained with fixed and automatic c value.';

drop function if exists best_CV_setting(Text);

create or replace function best_CV_setting
(
IN kernel Text
) 
returns setof kernelsettingresult as
$$
select * from results_CV($1) ORDER BY f_measure DESC limit 1;
$$
LANGUAGE SQL 
IMMUTABLE RETURNS NULL ON NULL INPUT
;
comment on function best_CV_setting(Text) is 
'Calculates the best setting according to F1 using results_CV function.
The parameter is the name of the kernel.';


drop function if exists best_CV_folds(Text);

create or replace function best_CV_folds
(
IN kernel Text
) 
returns setof integer as
$$
SELECT 
ppicvfoldsid
from foldCVcombined p
join best_CV_setting($1) s on
p.kernel is not distinct from s.kernel and
p.kernel_script is not distinct from s.kernel_script and
p.j is not distinct from s.j and 
p.k is not distinct from s.k and 
p.lmin is not distinct from s.lmin and
p.lmax is not distinct from s.lmax and
p.match_ is not distinct from s.match_ and
p.normalized is not distinct from s.normalized and
p.input_format is not distinct from s.input_format and
p.led is not distinct from s.led and
s.c_normal is not distinct from p.c_normal
-- GROUP BY parsertype,parser,c_normal,
-- kernel_script,j,k,lmin,lmax,match_,normalized,input_format,led
$$
LANGUAGE SQL 
IMMUTABLE RETURNS NULL ON NULL INPUT
;
comment on function best_CV_folds(Text) is 
'Returns the ppicvfoldsids that belong to the best CV setting (F-measure based). 
The number of returned rows should be equal to the number of corpora.
The parameter is the name of the kernel.';

-- Second part

DROP FUNCTION IF EXISTS add_threshold_version(integer,double precision);
CREATE FUNCTION add_threshold_version
(
  myppicvid int,
  th double precision
)
RETURNS INT AS
$$
DECLARE
    mykernel text;
    mycorpus text;
    myfold INT;
    myparsertype text;
    myparser text;
    mykernelscript text;
    myc double precision;
    myj double precision;
    myled Text;
    myauc double precision;

    combinedKernel text;
    combinedKernelScript text;
    newppicvid integer;

BEGIN
  SELECT INTO mykernel kernel FROM ppicv WHERE ppicvid=myppicvid;
  SELECT INTO mycorpus corpus FROM ppicv WHERE ppicvid=myppicvid;
  SELECT INTO myfold fold FROM ppicv WHERE ppicvid=myppicvid;
  SELECT INTO myparser parser FROM ppicv WHERE ppicvid=myppicvid;
  SELECT INTO myparsertype parsertype FROM ppicv WHERE ppicvid=myppicvid;
  SELECT INTO mykernelscript kernel_script FROM ppicv WHERE ppicvid=myppicvid;
  SELECT INTO myc c FROM ppicv WHERE ppicvid=myppicvid;  
  SELECT INTO myj j FROM ppicv WHERE ppicvid=myppicvid;
  SELECT INTO myled led FROM ppicv WHERE ppicvid=myppicvid;
  SELECT INTO myauc auc FROM ppicv WHERE ppicvid=myppicvid; -- auc does not change!
      
  SELECT INTO combinedKernel (SELECT mykernel || ' (thresholded)');
  SELECT INTO combinedKernelScript (SELECT mykernelscript || ' (thresholded)');
   
  IF EXISTS (SELECT ppicvid FROM ppicv WHERE 
	kernel is not distinct from combinedKernel 
	and corpus is not distinct from mycorpus 
	and fold is not distinct from myfold 
	and parsertype is not distinct from myparsertype 
	and c is not distinct from myc 
	and j is not distinct from myj 
	and led is not distinct from myled
	and parser is not distinct from myparser 
	and kernel_script is not distinct from combinedKernelScript) 
    THEN 
    RETURN (
    SELECT ppicvid FROM ppicv WHERE 
	kernel is not distinct from combinedKernel 
	and corpus is not distinct from mycorpus 
	and fold is not distinct from myfold 
	and parsertype is not distinct from myparsertype 
	and c is not distinct from myc 
	and j is not distinct from myj 
	and led is not distinct from myled
	and parser is not distinct from myparser 
	and kernel_script is not distinct from combinedKernelScript); /* already exists */
  END IF;

  INSERT INTO ppicv (corpus,parsertype,parser,kernel, fold, normalized,c,j,kernel_script,auc,forced_threshold) VALUES 
 	(mycorpus, myparsertype,myparser,combinedKernel,myfold,'t',myc,myj,combinedKernelScript,myauc,th);
	
  SELECT INTO newppicvid currval('ppiCV_ppiCVid_seq');

  INSERT INTO
    ppicvoutput (expId, pair, output, prediction)
  SELECT
    newppicvid         as expId, 
    sub.pair	      as pair,
    (case when sub.prediction>=$2 then TRUE else FALSE end) as output,
    prediction
  FROM
  (SELECT
     prediction,
     pair
     FROM
    ppiCVoutput
    WHERE expid = myppicvid
   ) as sub
;

  UPDATE ppicv set 
	tp=s.tp,
	fp=s.fp, 
	fn=s.fn, 
	tn=s.tn,
	total=s.n,
	precision_=s.p,
	recall=s.r,
	f_measure=s.f
  FROM get_statistics(newppicvid) as s
  where ppicvid=newppicvid;

  RETURN newppicvid; /* just created */
  
END
$$
LANGUAGE 'plpgsql'
;
comment on function add_threshold_version(integer, double precision) is 
'Adds a new row into ppicv (and many into ppicvoutput) with thresholding on the experiment given by ppicvid
Warning: if the thresholded version already exists, the function returns its ppicvid';


-- Third part

drop table if exists Best_CV cascade;

create table Best_CV as 
select * from ppicv p where ppicvfoldsid in 
(select * from best_CV_folds('SL')
union all
select * from best_CV_folds('ST')
union all
select * from best_CV_folds('SST')
union all
select * from best_CV_folds('PT')
union all
select * from best_CV_folds('SpT')
union all
select * from best_CV_folds('kBSPS')
union all
select * from best_CV_folds('cosine')
union all
select * from best_CV_folds('edit')
union all
select * from best_CV_folds('APG')
union all
select * from best_CV_folds('Kim-shallow (thresholded)')
union all
select * from best_CV_folds('Kim-lexical (thresholded)')
union all
select * from best_CV_folds('Kim-syntactic (thresholded)')
union all
select * from best_CV_folds('Kim-combined (thresholded)')
-- non-kernel-based classifiers
union all
select * from best_CV_folds('KStar')
union all
select * from best_CV_folds('J48')
union all
select * from best_CV_folds('BayesNet')
union all
select * from best_CV_folds('Logistic')
union all
select * from best_CV_folds('JRip')
union all
select * from best_CV_folds('LADTree')
union all
select * from best_CV_folds('NaiveBayes')
union all
select * from best_CV_folds('PART')
);

alter table best_cv add constraint pk_best_cv primary key (ppicvid);

create or replace view best_CV_output as 
select p.kernel,o.* from best_CV p
join ppicvoutput o on o.expid=p.ppicvid
;

drop type if exists clcvoutput cascade;

create type clcvoutput as 
(
	pair Text,
	cloutput Boolean,
	clpred double precision,
	cvoutput Boolean,
	cvpred double precision	
);

create or replace function best_outputs(kernel Text) returns setof clcvoutput as
$$
select p.pairLabel as pair,
	cl.output as cloutput, cl.prediction as clpred,
	cv.output as cvoutput, cv.prediction as cvpred 
from pair p
left join best_CL_output cl on cl.pair = p.pairLabel and cl.kernel = $1
left join best_CV_output cv on cv.pair = p.pairLabel and cv.kernel = $1
$$
LANGUAGE 'SQL'
STABLE RETURNS NULL ON NULL INPUT
;

