#!/usr/bin/perl
# Sript to parse corpora in XML, and write them to relational database.
# Input format:
# 
# <corpus source="LLL">
#   <document id="LLL.d0" origId="11069677">
#     <sentence id="LLL.d0.s0" origId="11069677-3" text="In vivo studies of the activity of four of the kinases, KinA, KinC, KinD (ykvD) and KinE (ykrQ), using abrB transcription as an indicator of Spo0A~P level, revealed that KinC and KinD were responsible for Spo0A~P production during the exponential phase of growth in the absence of KinA and KinB.">
#       <entity charOffset="170-173" id="LLL.d0.s0.e0" origId="29" text="KinC" type="agent" />
#       <entity charOffset="179-182" id="LLL.d0.s0.e1" origId="31" text="KinD" type="agent" />
#       <entity charOffset="205-211" id="LLL.d0.s0.e2" origId="35" text="Spo0A~P" type="target" />
#       <pair e1="LLL.d0.s0.e0" e2="LLL.d0.s0.e1" id="LLL.d0.s0.p0" interaction="False" />
#       <pair e1="LLL.d0.s0.e0" e2="LLL.d0.s0.e2" id="LLL.d0.s0.p1" interaction="True" />
#       <pair e1="LLL.d0.s0.e1" e2="LLL.d0.s0.e2" id="LLL.d0.s0.p2" interaction="True" />
# ...

use strict;
use warnings;

# UTF-8 stdio
use open qw( :std :encoding(UTF-8) );

use XML::Parser;
 

# main

if ($#ARGV < 0)
{
	print "Usage $0 <corpus.xml> ...\n";
	exit -1;
}

my $parser = new XML::Parser;

$parser->setHandlers(
	Start => \&startElement,
#	End => \&endElement,
#	Char => \&characterData,
#	Default => \&default
);

# quoting
my $q = '"';

# current values
my ($corpus, $document, $sentence, $entity, $pair);
my $count = 0;
my $start = time;
 
my %elementsToHandle = map { $_ => 1 } qw( corpus document sentence entity pair );

foreach my $xmlfile (@ARGV)
{
	die "Cannot find file $q$xmlfile$q"
		unless -f $xmlfile;

	$parser->parsefile($xmlfile);
}


# subs



sub startElement 
{
	return unless exists $elementsToHandle{$_[1]};
	my( $parseinst, $element, %attrs ) = @_;
	SWITCH: {
		# <corpus source="LLL">
		if ($element eq 'corpus') {
			$corpus = $attrs{'source'};
			print "INSERT INTO corpus VALUES ($q$corpus$q);\n";
			last SWITCH;
		}
		# <document id="LLL.d0" origId="11069677">
		if ($element eq 'document') {
			$document = $attrs{'id'} or warn();
			my $origId = $attrs{'origId'} || $attrs{'origID'} or warn();
			my ($position) = ($document =~ m/(\d+)$/) or warn();

			print "INSERT INTO doc VALUES ($q$document$q, $q$corpus$q, $position, $q$origId$q);\n";
			last SWITCH;
		}
		# <sentence id="LLL.d0.s0" origId="11069677-3" text="In vivo studies of the activity of four of the kinases, KinA, KinC, KinD (ykvD) and KinE (ykrQ), using abrB transcription as an indicator of Spo0A~P level, revealed that KinC and KinD were responsible for Spo0A~P production during the exponential phase of growth in the absence of KinA and KinB.">
		if ($element eq 'sentence') {
			$sentence = $attrs{'id'} or warn();
			my $origId = $attrs{'origId'} || $attrs{'origID'} || $attrs{'seqId'} || "";
			my $text = $attrs{'text'} or warn();
			my ($position) = ($sentence =~ m/(\d+)$/) or warn();
			my $escapedText= $text;
			$escapedText =~ s/"/\\"/g;

			$count++;
			if ($count % 500 == 0)
			{
				print STDERR "INFO: Processed $count sentences (" . (time == $start ? '?' : int($count / (time - $start)))  . " #/s)...\n";
			}


			print "INSERT INTO sentence VALUES ($q$sentence$q, $q$document$q, $position, $q$origId$q, $q$escapedText$q);\n";
			last SWITCH;
		}
		# <entity charOffset="205-211" id="LLL.d0.s0.e2" origId="35" text="Spo0A~P" type="target" />
		if ($element eq 'entity') {
			$entity = $attrs{'id'} or warn("entity:id");
			my $origId = $attrs{'origId'} || $attrs{'origID'} || "";
			my $text = $attrs{'text'} or warn();
			my $type = $attrs{'type'} || "";# warn();
			my ($position) = ($entity =~ m/(\d+)$/) or warn();
			my ($begin, $end, $begin2, $end2) = ($attrs{'charOffset'} =~ m/^(\d+)-(\d+)(?:,(\d+)-(\d+))?(.+)?$/) or warn($attrs{'charOffset'});
			warn($attrs{'charOffset'}) if defined $5;
			if (!defined $begin2) {$begin2='NULL';$end2='NULL';}
			print "INSERT INTO entity VALUES ($q$entity$q, $q$sentence$q, $position, $q$origId$q, $q$text$q, $q$type$q, $begin, $end, $begin2, $end2);\n";
			last SWITCH;
		}
		#       <pair e1="LLL.d0.s0.e0" e2="LLL.d0.s0.e1" id="LLL.d0.s0.p0" interaction="False" />
		if ($element eq 'pair') {
			$pair = $attrs{'id'} or warn();
			my $e1 = $attrs{'e1'} or warn();
			my $e2 = $attrs{'e2'} or warn();
			my ($interaction) = $attrs{'interaction'} or warn();

			my $groundTruth;
			$groundTruth = 0 if $interaction =~ /False/i;
			$groundTruth = 1 if $interaction =~ /True/i;
			warn() unless defined $groundTruth;

			print "INSERT INTO pair VALUES ($q$pair$q, $q$e1$q, $q$e2$q, $q$groundTruth$q);\n";
			last SWITCH;
		}
	}
}

sub endElement 
{
      my( $parseinst, $element ) = @_;
}

sub characterData 
{
	my( $parseinst, $data ) = @_;
}

sub default 
{	
	my( $parseinst, $data ) = @_;
        # do nothing, but stay quiet
}
