#!/usr/bin/perl
# Sript to parse corpora in XML, and write them to relational database.
# Input format:
# 
# <corpus source="LLL">
#   <document id="LLL.d0" origId="11069677">
#     <sentence id="LLL.d0.s0" origId="11069677-3" text="In vivo studies of the activity of four of the kinases, KinA, KinC, KinD (ykvD) and KinE (ykrQ), using abrB transcription as an indicator of Spo0A~P level, revealed that KinC and KinD were responsible for Spo0A~P production during the exponential phase of growth in the absence of KinA and KinB.">
#       <entity charOffset="170-173" id="LLL.d0.s0.e0" origId="29" text="KinC" type="agent" />
#       <entity charOffset="179-182" id="LLL.d0.s0.e1" origId="31" text="KinD" type="agent" />
#       <entity charOffset="205-211" id="LLL.d0.s0.e2" origId="35" text="Spo0A~P" type="target" />
#       <pair e1="LLL.d0.s0.e0" e2="LLL.d0.s0.e1" id="LLL.d0.s0.p0" interaction="False" />
#       <pair e1="LLL.d0.s0.e0" e2="LLL.d0.s0.e2" id="LLL.d0.s0.p1" interaction="True" />
#       <pair e1="LLL.d0.s0.e1" e2="LLL.d0.s0.e2" id="LLL.d0.s0.p2" interaction="True" />
# ...

use strict;
use warnings;

use XML::Parser;
use Graph::Undirected;
use Graph::Directed;
use Data::Dumper;

require 'pstree_align.pl';
use constant { NODE_ID => 8, TOKEN_INX => 9, OFFSET_BEGIN => 10, OFFSET_END => 11 };
 

# main

if ($#ARGV < 0)
{
	print "Usage $0 <corpus.xml> ...\n";
	exit -1;
}

my $parser = new XML::Parser;

$parser->setHandlers(
	Start => \&startElement,
	End => \&endElement,
	Char => \&characterData,
	Default => \&default);

# quoting
my $q = '$$';

# current values
my ($corpus, $document, $sentence, $entity, $pair);

my ($g, $s, %pairs, %entities, %tokens, %constituents, $sentencetext);

my $tokenizer = 'Charniak-Lease';
my $depParser = 'Charniak-Lease';
my $synParser = 'Charniak-Johnson-McClosky';

my $collectTokens       = 0;
my $collectDependencies = 0;
my $collectBrackets     = 0;

my %dep_type_frequency = ();
my %const_type_frequency = ();

my %dep_type_frequency_per_pair = ();
my %const_type_frequency_per_pair = ();

my %dep_type_frequency_per_sentence = ();
my %const_type_frequency_per_sentence = ();


foreach my $xmlfile (@ARGV)
{
	die "Cannot find file $q$xmlfile$q"
		unless -f $xmlfile;

	$parser->parsefile($xmlfile);
}


# subs

sub startElement 
{

	my( $parseinst, $element, %attrs ) = @_;
	SWITCH: {
		# <corpus source="LLL">
		if ($element eq 'corpus') {
			$corpus = $attrs{'source'};
# 			print "INSERT INTO corpus VALUES ($q$corpus$q);\n";
			last SWITCH;
		}
		# <document id="LLL.d0" origId="11069677">
		if ($element eq 'document') {
			$document = $attrs{'id'} or warn();
			my $origId = $attrs{'origId'} || $attrs{'origID'} or warn();
			my ($position) = ($document =~ m/(\d+)$/) or warn();

# 			print "INSERT INTO doc VALUES ($q$document$q, $q$corpus$q, $position, $q$origId$q);\n";
			last SWITCH;
		}
		# <sentence id="LLL.d0.s0" origId="11069677-3" text="In vivo studies of the activity of four of the kinases, KinA, KinC, KinD (ykvD) and KinE (ykrQ), using abrB transcription as an indicator of Spo0A~P level, revealed that KinC and KinD were responsible for Spo0A~P production during the exponential phase of growth in the absence of KinA and KinB.">
		elsif ($element eq 'sentence') {
            $g = Graph::Undirected->new; # dependency graph
            $s = undef; # syntax tree
            %pairs = ();
            %entities = ();
            %constituents = ();
            $sentencetext = undef;
            
			$sentence = $attrs{'id'} or warn();
			my $origId = $attrs{'origId'} || $attrs{'origID'} || $attrs{'seqId'} || "";
			my $text = $attrs{'text'} or warn();
			my ($position) = ($sentence =~ m/(\d+)$/) or warn();
            
            $sentencetext = $text;

# 			print "INSERT INTO sentence VALUES ($q$sentence$q, $q$document$q, $position, $q$origId$q, $q$text$q);\n";
			last SWITCH;
		}
		# <entity charOffset="205-211" id="LLL.d0.s0.e2" origId="35" text="Spo0A~P" type="target" />
		elsif ($element eq 'entity') {
			$entity = $attrs{'id'} or warn("entity:id");
			my $origId = $attrs{'origId'} || $attrs{'origID'} || "";
			my $text = $attrs{'text'} or warn();
			my $type = $attrs{'type'} || "";# warn();
			my ($position) = ($entity =~ m/(\d+)$/) or warn();
			my ($begin, $end, $begin2, $end2) = ($attrs{'charOffset'} =~ m/^(\d+)-(\d+)(?:,(\d+)-(\d+))?(.+)?$/) or warn($attrs{'charOffset'});
			warn($attrs{'charOffset'}) if defined $5;
			if (!defined $begin2) {$begin2='NULL';$end2='NULL';}
			
			$entities{$entity} = {'id' => $entity, 'begin' => $begin, 'end' => $end+1, 'text' => $text};
# 			print "INSERT INTO entity VALUES ($q$entity$q, $q$sentence$q, $position, $q$origId$q, $q$text$q, $q$type$q, $begin, $end, $begin2, $end2);\n";
			last SWITCH;
		}
		#       <pair e1="LLL.d0.s0.e0" e2="LLL.d0.s0.e1" id="LLL.d0.s0.p0" interaction="False" />
		elsif ($element eq 'pair') {
			$pair = $attrs{'id'} or warn();
			my $e1 = $attrs{'e1'} or warn();
			my $e2 = $attrs{'e2'} or warn();
			my ($interaction) = $attrs{'interaction'} or warn();

			my $groundTruth;
			$groundTruth = 0 if $interaction =~ /False/i;
			$groundTruth = 1 if $interaction =~ /True/i;
			warn() unless defined $groundTruth;

            $pairs{$pair} = {'id' => $pair, 'e1' => $e1, 'e2' => $e2};
#			print "INSERT INTO pair VALUES ($q$pair$q, $q$e1$q, $q$e2$q, $q$groundTruth$q);\n";
			last SWITCH;
		}
        elsif ($element eq 'tokenization') {
            $collectTokens =  ($attrs{'tokenizer'} eq $tokenizer);
            last SWITCH;
        }
        elsif ($element eq 'token') {
            last SWITCH unless $collectTokens;
            # <token POS="IN" charOffset="0-1" id="clt_1" text="In" /> 
            my $token = $attrs{'id'} or warn("token:id");
            my ($begin, $end, $begin2, $end2) = ($attrs{'charOffset'} =~ m/^(\d+)-(\d+)(?:,(\d+)-(\d+))?(.+)?$/) or warn($attrs{'charOffset'});
            my $text = $attrs{'text'} or warn("token:text");
            my $POS = $attrs{'POS'} || "";# warn();
            $tokens{$token} = {'id' => $token,  'begin' => $begin, 'end' => $end + 1, 'text' => $text, 'POS' => $POS};
            $g->add_vertex($token);
        }
        elsif ($element eq 'parse') {
            $collectDependencies = ($attrs{'parser'} eq $depParser);
            last SWITCH;
        }
        elsif ($element eq 'dependency') {
            last SWITCH unless $collectDependencies;
            # <dependency id="clp_1" t1="clt_3" t2="clt_2" type="nn" />
            my $dep = $attrs{'id'} or warn("dep:id");
            my $type = $attrs{'type'} or warn();
            my $t1 = $attrs{'t1'} or warn("dep:t1");
            my $t2 = $attrs{'t2'} or warn("dep:t2");
            $g->add_edge($t1, $t2);
            $g->set_edge_attribute($t1, $t2, 'type', $type);
            $dep_type_frequency{$type}++;            
            $dep_type_frequency_per_sentence{$sentence}{$type}++;
        }
        elsif ($element eq 'bracketing') {
            $collectBrackets = ($attrs{'parser'} eq $synParser);
            if ($collectBrackets) 
            {
                my $bracketing = $attrs{'bracketing'};
#                 print "$sentencetext\n$bracketing\n\n";
                $s = build_syntax_tree($bracketing, $sentencetext);                
                if (defined $s) {
                    $s->walk(sub 
                        {
                            my ($self, $state) = @_;
                            $const_type_frequency{$self->tag()}++; 
                            $const_type_frequency_per_sentence{$sentence}{$self->tag()}++; 
                        }, undef, undef);
                } else {
                    print STDERR "ERROR: $bracketing\n";
                }
            }
            last SWITCH;
        }
        elsif ($element eq 'charOffsetMapEntry') {
            last SWITCH unless $collectBrackets;
			my ($begin, $end, $begin2, $end2) = ($attrs{'sentenceTextCharOffset'} =~ m/^(\d+)-(\d+)(?:,(\d+)-(\d+))?(.+)?$/) or warn($attrs{'sentenceTextCharOffset'});
            my $constituent = $attrs{'sentenceTextCharOffset'};
            #$constituents{$constituent} = {'id' => $constituent,  'begin' => $begin, 'end' => $end, 'text' => substr($sentencetext, $begin, $end-$begin)};
			# <dependency id="clp_1" t1="clt_3" t2="clt_2" type="nn" />
            # TODO
#             my $dep = $attrs{'id'} or warn("dep:id");
#             my $type = $attrs{'type'} or warn();
#             my $t1 = $attrs{'t1'} or warn("dep:t1");
#             my $t2 = $attrs{'t2'} or warn("dep:t2");
#             $g->add_edge($t1, $t2);
        }
	}
}

sub overlaps($$)
{
    my $a = shift; 
    my $b = shift;
    return !($a->{begin} >= $b->{end} || $a->{end} <= $b->{begin});
}

sub overlaps_constituent($$)
{
    my $a = shift; 
    my $b = shift;
    return undef unless defined $a->[OFFSET_BEGIN];
    return undef unless defined $a->[OFFSET_END];
    return !($a->[OFFSET_BEGIN] >= $b->{end} || $a->[OFFSET_END] <= $b->{begin});
}

sub contains($$)
{
    my $a = shift; 
    my $b = shift;
    return $a->{begin} <= $b->{begin} && $a->{end} >= $b->{end};
}

sub min_shortest_path($$)
{
    my $g = shift;
    my $pair = shift;
    
    my $e1 = $entities{$pairs{$pair}{e1}};
    my $e2 = $entities{$pairs{$pair}{e2}};
    
#     print "$pair $e1 $e2\n";
    my $min = undef;
    my @min_path = ();
    my $no_such_path = 0;
    for my $tok1 ($g->vertices)
    {
        if (overlaps($tokens{$tok1}, $e1))
        {
            for my $tok2 ($g->vertices)
            {
                if (overlaps($tokens{$tok2}, $e2))
                {
                    my $u = $tok1;
                    my $v = $tok2;
                    my @path = $g->SP_Dijkstra($u, $v);
                    if (scalar(@path) == 0)
                    {
                        $no_such_path = 1;
                    } elsif ((! defined $min) || ($min > scalar(@path))) {
                        $min = scalar(@path);
                        @min_path = @path;
                    }
                }
            }
        }
    }
    return defined $min 
        ? ($min, \@min_path) 
        : ($no_such_path
            ? (0, \@min_path)
            : (undef, undef));
}

sub min_shortest_path_syntax_tree($$)
{
    my $st = shift;
    my $pair = shift;
    
    return undef unless defined $st;
    die("ERROR: pair undefined") unless defined $pair;

    die("ERROR: e1 undefined") unless exists $pairs{$pair}{e1};
    die("ERROR: e2 undefined") unless exists $pairs{$pair}{e2};
    
    die("ERROR: ent1 undefined") unless exists $entities{$pairs{$pair}{e1}};
    die("ERROR: ent2 undefined") unless exists $entities{$pairs{$pair}{e2}};
    
    my $e1 = $entities{$pairs{$pair}{e1}};
    my $e2 = $entities{$pairs{$pair}{e2}};
    
#     print "$pair $e1 $e2\n";
    my $min = undef;
    my @min_path = undef;
    for my $tok1 ($st->get_all_terminals)
    {
        if (overlaps_constituent($tok1, $e1))
        {
            for my $tok2 ($st->get_all_terminals)
            {
                if (overlaps_constituent($tok2, $e2))
                {
                    my $cons =  $tok1->find_common_ancestor($tok2);
                    next unless defined $cons;
                    if (defined $cons)
                    {
                        my @path1 = $tok1->path_up_to($cons);
                        my @path2 = $tok2->path_up_to($cons);
                        next unless @path1;
                        next unless @path2;
#                         print "1: ".($tok1->tag)."\n@path1\n2:".($tok2->tag)."\n@path2\n\n";
                        my $path_length = scalar(@path1) + scalar(@path2) - 1;
                        if ( (! defined $min) || $min > $path_length)
                        {
                            $min = $path_length;
                            my $last = pop(@path2);
                            die "NODE MISMATCH: $last <> $path1[$#path1]\n"
                                unless $last == $path1[$#path1];
                            push(@path1, reverse(@path2));
                            @min_path = @path1;
                            die "MISMATCH: $path_length, $#min_path\n"
                                unless scalar(@min_path) == $path_length;
                        }
                    }
                }
            }
        }
    }
    return defined $min 
        ? ($min, \@min_path)
        : (undef, undef);
}

sub build_syntax_tree($$)
{
    my $ptb = shift;
    my $sentencetext = shift;
    
    my $st = to_aligned_parse_tree($ptb, $sentencetext);
    my %tokens = ();
    
    my $sg = Graph::Directed->new;
        
    return $st;
}

sub log2($) {
    my $n = shift;
    return log($n)/log(2);
}

# hash: ? -> num
# Kullback–Leibler divergence (information gain)
sub bit_dkl($$)
{
    my $p = shift;
    my $q = shift;
    my $sum = 0;
    
    
    my $e = 0.0;
    for my $i (keys(%$p))
    {
        my $pi = $p->{$i};
        next unless $pi > 0;
        my $qi = $q->{$i};
        $e += $pi*log2($pi/$qi);
    }
    return $e;
}

sub dep_path_first_last_edge($$)
{
    my $g = shift;
    my $path = shift;
    
    my %labels = ();
    
    my $first = $g->get_edge_attribute($path->[0], $path->[1], 'type');
    my $last  = $#{$path} > 1 ? $g->get_edge_attribute($path->[$#{$path}-1], $path->[$#{$path}], 'type') : '';
    
    my @a = ($first, $last);
    my $subj = scalar( grep {/subj/} @a );
    my $obj  = scalar( grep {/obj/ } @a );
        
    return ($subj, $obj);
}


sub dep_path_to_frequency($$)
{
    my $g = shift;
    my $path = shift;
    
    my %labels = ();
    
    for my $i (1..$#{$path})
    {
        my $u = $path->[$i-1];
        my $v = $path->[$i];
        my $type = $g->get_edge_attribute($u, $v, 'type');
        $labels{$type}++;
    }
    return \%labels;
}


sub syn_path_to_frequency($)
{
    my $path = shift;
    
    my %labels = ();
    for my $c (@$path)
    {
        print "UNDEF: ".Dumper($path) unless defined $c;
        my $type = $c->tag;
        $labels{$type}++;
    }
    return \%labels;
}


sub endElement 
{
    my( $parseinst, $element ) = @_;
    
    if ($element eq 'sentence')
    {
        # dependency graph
#         print "$g\n";
        for $pair (keys(%pairs))
        {
            my ($min, $sp) = min_shortest_path($g, $pair);
            
            my $min_val = defined $min ? $min : -1;
            print "UPDATE pair SET depShortestPath = $min_val WHERE pairLabel = $q$pair$q;\n";
            
            if (defined $sp)
            {
                my $p = dep_path_to_frequency($g, $sp);
                $dep_type_frequency_per_pair{$pair} = $p;
            } else {
                warn("WARNING: missing dep not implemented");
            }
            
            if (defined $sp && scalar($sp) > 0)
            {            
                my ($subj, $obj) = dep_path_first_last_edge($g, $sp);
                print "UPDATE pair SET border_subj = $subj, border_obj = $obj WHERE pairLabel = $q$pair$q;\n";
            }
		}
		
		# syntax tree
        for $pair (keys(%pairs))
        {
            my ($min, $sp) = min_shortest_path_syntax_tree($s, $pair);
            my $min_val = defined $min ? $min : -1;
            print "UPDATE pair SET syntShortestPath = $min_val WHERE pairLabel = $q$pair$q;\n";
            if (defined $sp)
            {
                my $p = syn_path_to_frequency($sp);
                $const_type_frequency_per_pair{$pair} = $p;
            } else {
                warn("WARNING: missing const not implemented");
            }
		}
    }
    elsif ($element eq 'parse')
    {
        $collectDependencies = 0;
    }
    elsif ($element eq 'tokens')
    {
        $collectDependencies = 0;
    }
}

sub characterData
{
# 	my( $parseinst, $data ) = @_;
}

sub default 
{
# 	my( $parseinst, $data ) = @_;
        # do nothing, but stay quiet
}

sub absolute2relative($)
{
    my $p = shift;
    my $sum = 0;
    $sum += $_ foreach values(%$p);
    $p->{$_} = ($p->{$_}/(0.0 + $sum)) foreach keys(%$p);
}

sub esacpe_sql_identifier($)
{
    my $s = shift;
    $s =~ s/(^[0-9]|[^a-zA-Z0-9_])/"_x".ord($1)."_"/ge;
    return $s;
}

sub create_table($$$$$$)
{
    my $tbl = shift;
    my $id = shift;
    my $prefix = shift;
    my $cols = shift;
    my $type = shift;
    my $data = shift;
    
    my $csv = "$tbl.csv";
    my $sep = ',';
    my $null = '';
    
    print "\n--\n-- Table: '$tbl', file: $csv, prefix: $prefix, cols: ".scalar(@$cols)."+1\n--\n";
    print "DROP TABLE IF EXISTS $tbl;\n";
    print "CREATE TABLE $tbl (\n";
    print "  $prefix$id text PRIMARY KEY";
    print ",\n  $prefix".esacpe_sql_identifier($_)." $type NOT NULL DEFAULT 0" foreach @$cols;
    print "\n);\n";
    
    open(my $fh, '>', $csv) or die $!;
    print $fh "$prefix$id";
    print $fh "$sep$prefix".(esacpe_sql_identifier($_)) foreach @$cols;
    print $fh "\n";
    while (my ($key, $val) = each(%$data))
    {
        print $fh $key;
        print $fh $sep . (exists $val->{$_} ? $val->{$_} : '0') foreach @$cols;
        print $fh "\n";
    }
    close($fh);
    print '\copy'." $tbl FROM $csv WITH DELIMITER AS '$sep' NULL AS '$null' CSV HEADER\n";
}



###############################
# dep frequencies

my @dep_types = keys(%dep_type_frequency);

# global
create_table('dep_type_occurrence', 'dummy', 'dto_', \@dep_types, 'integer', {'ALL' => \%dep_type_frequency });
absolute2relative(\%dep_type_frequency);
create_table('dep_type_frequency', 'dummy', 'dtf_', \@dep_types, 'double precision', {'ALL' => \%dep_type_frequency });

# pairs
create_table('dep_type_occurrence_pair', 'pairLabel', 'dtop_', \@dep_types, 'integer', \%dep_type_frequency_per_pair);
for my $pair (keys(%dep_type_frequency_per_pair))
{
    my $freq = $dep_type_frequency_per_pair{$pair};
    absolute2relative($freq);
    my $dkl = bit_dkl($freq, \%dep_type_frequency);
    print "UPDATE pair SET dkl_dep_shortest_path = $dkl WHERE pairLabel = $q$pair$q;\n";
}
create_table('dep_type_frequency_pair', 'pairLabel', 'dtfp_', \@dep_types, 'double precision', \%dep_type_frequency_per_pair);

# sentences
create_table('dep_type_occurrence_sentence', 'sentenceLabel', 'dtos_', \@dep_types, 'integer', \%dep_type_frequency_per_sentence);
for my $sentence (keys(%dep_type_frequency_per_sentence))
{
    my $freq = $dep_type_frequency_per_sentence{$sentence};
    absolute2relative($freq);
    my $dkl = bit_dkl($freq, \%dep_type_frequency);
    print "UPDATE sentence SET dkl_dep_graph = $dkl WHERE label = $q$sentence$q;\n";
}
create_table('dep_type_frequency_sentence', 'sentenceLabel', 'dtfs_', \@dep_types, 'double precision', \%dep_type_frequency_per_sentence);

###############################
# const type (POS) frequencies
absolute2relative(\%const_type_frequency);
for my $pair (keys(%const_type_frequency_per_pair))
{
    my $freq = $const_type_frequency_per_pair{$pair};
    absolute2relative($freq);
    my $dkl = bit_dkl($freq, \%const_type_frequency);
    print "UPDATE pair SET dkl_synt_shortest_path = $dkl WHERE pairLabel = $q$pair$q;\n";
}

for my $sentence (keys(%const_type_frequency_per_sentence))
{
    my $freq = $const_type_frequency_per_sentence{$sentence};
    absolute2relative($freq);
    my $dkl = bit_dkl($freq, \%const_type_frequency);
    print "UPDATE sentence SET dkl_synt_tree = $dkl WHERE label = $q$sentence$q;\n";
}