drop function if exists threshold_stats(integer, double precision);
create or replace function threshold_stats (CVexpID integer, threshold double precision) returns stats as 
'select 
	newStats(
	sum(case when a.prediction>=$2 and c.groundTruth is TRUE then 1 else 0 end),
	sum(case when a.prediction>=$2 and c.groundTruth is FALSE then 1 else 0 end),
	sum(case when a.prediction<$2 and c.groundTruth is TRUE then 1 else 0 end),
	sum(case when a.prediction<$2 and c.groundTruth is FALSE then 1 else 0 end)
	)
from ppiCVoutput a
join ppiCV b on a.expID=b.ppiCVid 
join pair c on a.pair = c.pairLabel 
where a.expid=$1 
group by a.expid
;
' language SQL stable returns null on null input;
comment on function threshold_stats (integer, double precision)
is 'Calculates tp,fp,fn,tn (in this order) where the decision hyperplane is given by threshold for a CV experiment:
under threshold is a negative prediction, over threshold is a positive prediction
on Kims kernels the default threshold is 0.5-epsilon=0.4999999999
on APG kernel the default threshold is specified by the kernel and not attached to the experiment results, 
therefore it is not advised to use this function
on other kernels the default threshold is 0.0';

-- test 
-- select ppicvid,tp,fp,fn,tn,threshold_stats(ppicvid,0.5),precision_,recall,f_measure from ppicv where ppicvfoldsid=11272;

DROP TYPE IF EXISTS PRF CASCADE;
CREATE TYPE PRF AS 
(
    prec DOUBLE PRECISION,
    recall DOUBLE PRECISION,
    f1 DOUBLE PRECISION
);

DROP FUNCTION IF EXISTS newPRF(DOUBLE PRECISION,DOUBLE PRECISION,DOUBLE PRECISION);
CREATE FUNCTION newPRF
(
    IN prec DOUBLE PRECISION,
    IN recall DOUBLE PRECISION,
    IN f1 DOUBLE PRECISION
)
RETURNS PRF AS
$$
    SELECT $1, $2, $3;
$$ 
LANGUAGE SQL 
IMMUTABLE RETURNS NULL ON NULL INPUT;

DROP FUNCTION IF EXISTS precPRF(PRF);
CREATE FUNCTION precPRF
(
    IN myprf PRF
)
RETURNS DOUBLE PRECISION AS
$$
    SELECT $1.prec;
$$ 
LANGUAGE SQL 
IMMUTABLE RETURNS NULL ON NULL INPUT;

DROP FUNCTION IF EXISTS recallPRF(PRF);
CREATE FUNCTION recallPRF
(
    IN myprf PRF
)
RETURNS DOUBLE PRECISION AS
$$
    SELECT $1.recall;
$$ 
LANGUAGE SQL 
IMMUTABLE RETURNS NULL ON NULL INPUT;

DROP FUNCTION IF EXISTS f1PRF(PRF);
CREATE FUNCTION f1PRF
(
    IN myprf PRF
)
RETURNS DOUBLE PRECISION AS
$$
    SELECT $1.f1;
$$ 
LANGUAGE SQL 
IMMUTABLE RETURNS NULL ON NULL INPUT;


drop function if exists calculatePRF(integer);
create or replace function calculatePRF(CVexpID integer) returns prf as 
'select 
newPRF(
	rr(prec),
	rr(recall),
	(case when prec + recall>0 then rr(2 * prec * recall/( prec + recall )) else 0 end))
from
(Select *,
(case when tp+fp>0 then (cast(tp as double precision)/(cast(tp as double precision)+cast(fp as double precision))) else 0 end) as prec,
(case when tp+fn>0 then (cast(tp as double precision)/(cast(tp as double precision)+cast(fn as double precision))) else 0 end) as recall 
from (Select a.expid,count(*) as cnt,b.total,
	sum(case when a.output=FALSE and c.groundTruth != a.output then 1 else 0 end) as fn,
	sum(case when a.output=FALSE and c.groundTruth = a.output then 1 else 0 end) as tn,
	sum(case when a.output=TRUE and c.groundTruth != a.output then 1 else 0 end) as fp,
	sum(case when a.output=TRUE and c.groundTruth = a.output then 1 else 0 end) as tp, count(*)
from ppiCVoutput a 
join ppiCV b on a.expID=b.ppiCVid 
join pair c on a.pair = c.pairLabel where expid=$1 group by a.expid, b.total) as sub1)
as sub2;'
language SQL stable returns null on null input;
comment on function calculatePRF (integer)
is 'Calculates precision,recall,f1 (in this order) for the CV experiment $1';



drop function if exists PRF_threshold(integer, double precision);
create or replace function PRF_threshold (CVexpID integer, threshold double precision) returns prf as 
'select 
newPRF(
	rr(prec),
	rr(recall),
	(case when prec + recall>0 then rr(2 * prec * recall/( prec + recall )) else 0 end))
from
(Select *,
(case when tp+fp>0 then (cast(tp as double precision)/(cast(tp as double precision)+cast(fp as double precision))) else 0 end) as prec,
(case when tp+fn>0 then (cast(tp as double precision)/(cast(tp as double precision)+cast(fn as double precision))) else 0 end) as recall 
from threshold_stats($1,$2)
) as sub1;'
language SQL stable returns null on null input;
comment on function PRF_threshold (integer, double precision)
is 'Calculates precision,recall,f1 (in this order) where the decision hyperplane is given by threshold for a CV experiment:
under threshold is a negative prediction, over threshold is a positive prediction
on Kims kernels the default threshold is 0.5-epsilon=0.4999999999
on APG kernel the default threshold is specified by the kernel and not attached to the experiment results, 
therefore it is not advised to use this function
on other kernels the default threshold is 0.0';


-- test
-- select * from PRF_threshold(36924,0.25);

drop function if exists foldsPRF (integer);
create or replace function foldsPRF (foldID integer) 
returns PRF as 
'select avg(t.prec),avg(t.recall),avg(t.f1) from
(select (calculatePRF(ppicvid)).prec as prec,
(calculatePRF(ppicvid)).recall as recall,
(calculatePRF(ppicvid)).f1 as f1
from ppicv where ppicvfoldsid=$1) as t;' 
language SQL stable returns null on null input;
comment on function foldsPRF (integer) is
'Calculates the average precision,recall,f1 of the fold $1 using the PRF_threshold function';

drop function if exists foldsPRF_threshold (integer,double precision);
create or replace function foldsPRF_threshold (foldID integer,threshold double precision) 
returns PRF as 
'select avg(t.prec),avg(t.recall),avg(t.f1) from
(select (PRF_threshold(ppicvid, $2)).prec as prec,
(PRF_threshold(ppicvid, $2)).recall as recall,
(PRF_threshold(ppicvid, $2)).f1 as f1
from ppicv where ppicvfoldsid=$1) as t;'  
language SQL stable returns null on null input;
comment on function foldsPRF_threshold (integer,double precision) is
'Calculates the average precision,recall,f1 of the fold $1 with threshold $2 using the PRF_threshold function';

-- test
-- select * from foldsPRF_threshold(11766,0.25);

-- test

DROP FUNCTION IF EXISTS compute_auc_cv(integer);
CREATE OR REPLACE FUNCTION compute_auc_cv(CVexpID integer) RETURNS double precision AS 
$$
SELECT
  ( CASE 
    WHEN tn + fp = 0 THEN (CAST (tp AS NUMERIC) / (tp + fn)) -- TP rate
    WHEN tp + fn = 0 THEN (CAST (tn AS NUMERIC) / (tn + fp)) -- TN rate
    ELSE 
      ( SELECT 
          SUM(SIGN(pos.prediction - neg.prediction)+1)/2 AS errorsum -- pos>neg: 1, pos=neg: 0.5, pos<neg: 0
        FROM
          ppiCVoutput pos 
            CROSS JOIN ppiCVoutput neg
            JOIN pair ppos ON pos.pair = ppos.pairLabel
            JOIN pair pneg ON neg.pair = pneg.pairLabel
        WHERE pos.expid = $1
          AND neg.expid = $1
          AND ppos.groundTruth IS     TRUE
          AND pneg.groundTruth IS NOT TRUE
      )
      /CAST ((tn + fp)*(tp + fn) AS double precision) END
  ) AS auc
FROM
  ( SELECT
--       SUM(CASE WHEN a.prediction >  0 AND c.groundTruth IS     TRUE THEN 1 ELSE 0 END) AS tp,
--       SUM(CASE WHEN a.prediction >  0 AND c.groundTruth IS NOT TRUE THEN 1 ELSE 0 END) AS fp,
--       SUM(CASE WHEN a.prediction <= 0 AND c.groundTruth IS     TRUE THEN 1 ELSE 0 END) AS fn,
--       SUM(CASE WHEN a.prediction <= 0 AND c.groundTruth IS NOT TRUE THEN 1 ELSE 0 END) AS tn
      SUM(CASE WHEN a.output is TRUE AND c.groundTruth IS     TRUE THEN 1 ELSE 0 END) AS tp,
      SUM(CASE WHEN a.output is TRUE AND c.groundTruth IS NOT TRUE THEN 1 ELSE 0 END) AS fp,
      SUM(CASE WHEN a.output is NOT TRUE AND c.groundTruth IS     TRUE THEN 1 ELSE 0 END) AS fn,
      SUM(CASE WHEN a.output is NOT TRUE AND c.groundTruth IS NOT TRUE THEN 1 ELSE 0 END) AS tn
    FROM
      ppiCVoutput a
        JOIN ppiCV b ON a.expID = b.ppiCVid
        JOIN pair  c ON a.pair  = c.pairLabel
    WHERE
      expid = $1
    GROUP BY
      a.expid
  ) AS sub1;
$$ 
LANGUAGE SQL 
STABLE
RETURNS NULL ON NULL INPUT;
COMMENT ON FUNCTION compute_auc_cv (integer)
IS 'Calculates AUC for given CV fold';

DROP FUNCTION IF EXISTS compute_auc_cl(integer);
CREATE OR REPLACE FUNCTION compute_auc_cl(CLexpID integer) RETURNS double precision AS 
$$
SELECT
  ( CASE 
    WHEN tn + fp = 0 THEN (CAST (tp AS NUMERIC) / (tp + fn)) -- TP rate
    WHEN tp + fn = 0 THEN (CAST (tn AS NUMERIC) / (tn + fp)) -- TN rate
    ELSE 
      ( SELECT 
          SUM(SIGN(pos.prediction - neg.prediction)+1)/2 AS errorsum -- pos>neg: 1, pos=neg: 0.5, pos<neg: 0
        FROM
          ppiCLoutput pos 
            CROSS JOIN ppiCLoutput neg
            JOIN pair ppos ON pos.pair = ppos.pairLabel
            JOIN pair pneg ON neg.pair = pneg.pairLabel
        WHERE pos.expid = $1
          AND neg.expid = $1
          AND ppos.groundTruth IS     TRUE
          AND pneg.groundTruth IS NOT TRUE
      )
      /CAST ((tn + fp)*(tp + fn) AS double precision) END
  ) AS auc
FROM
  ( SELECT
--       SUM(CASE WHEN a.prediction >  0 AND c.groundTruth IS     TRUE THEN 1 ELSE 0 END) AS tp,
--       SUM(CASE WHEN a.prediction >  0 AND c.groundTruth IS NOT TRUE THEN 1 ELSE 0 END) AS fp,
--       SUM(CASE WHEN a.prediction <= 0 AND c.groundTruth IS     TRUE THEN 1 ELSE 0 END) AS fn,
--       SUM(CASE WHEN a.prediction <= 0 AND c.groundTruth IS NOT TRUE THEN 1 ELSE 0 END) AS tn
      SUM(CASE WHEN a.output is TRUE AND c.groundTruth IS     TRUE THEN 1 ELSE 0 END) AS tp,
      SUM(CASE WHEN a.output is TRUE AND c.groundTruth IS NOT TRUE THEN 1 ELSE 0 END) AS fp,
      SUM(CASE WHEN a.output is NOT TRUE AND c.groundTruth IS     TRUE THEN 1 ELSE 0 END) AS fn,
      SUM(CASE WHEN a.output is NOT TRUE AND c.groundTruth IS NOT TRUE THEN 1 ELSE 0 END) AS tn
    FROM
      ppiCLoutput a
        JOIN ppiCL b ON a.expID = b.ppiCLid
        JOIN pair  c ON a.pair  = c.pairLabel
    WHERE
      expid = $1
    GROUP BY
      a.expid
  ) AS sub1;
$$ 
LANGUAGE SQL 
STABLE
RETURNS NULL ON NULL INPUT;
COMMENT ON FUNCTION compute_auc_cl (integer)
IS 'Calculates AUC for given CL experiment';


