package corpus;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import java.util.logging.Level;
import java.util.logging.Logger;
import javax.naming.OperationNotSupportedException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import parser.Parser;
import parser.StanfordParser;
import preprocessing.Preprocessor;

import deeper.Interaction;
import deeper.InteractionType;
import deeper.Protein;
import exceptions.GraphException;
import exceptions.NoSuchWordException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.io.StringReader;
import java.util.Collections;
import jung.entities.Graph;
import parser.StructuredSentence;
import util.Params;

/**
 * Leaf class of the Composite pattern. COncentrate the most useful functionality
 * of the application.
 * 
 * @author tfayruzo
 * 
 */
public class Sentence extends Text implements Serializable {

    private static Log _log = LogFactory.getLog(Sentence.class);
    public String text;
    public String modText;

    // fields that should be filled during corpus construction
    private List<Protein> proteins;
    //private List<Protein> nestedProteins;
    private List<Interaction> interactions;
    public transient Graph depTree;
    public transient Graph parseTree;
    public StructuredSentence struct;

    // preprocessing data. Preprocessing step is necessary, because
    // it assigns unique ids to tagged proteins. This allows to avoid confusion
    // if proteins are mentioned in the sentence several times.
    // private Preprocessor prep = new SimplePreprocessor();
    // private Map<String, String> abbr;
    public Sentence(String text, String id, String origId) {
        this.id = id;
        this.originalId = origId;
        this.text = text;
        proteins = new ArrayList<Protein>();
        interactions = new ArrayList<Interaction>();
    }

    @Override
    public Collection<Interaction> interactions() {
        return interactions;
    }

    @Override
    public void parse(Parser parser, String... params) throws Exception {
        _log.info(text);
        Params.addLog( 3 , "Sentence: " + text);
        boolean load = false, save = false;
        String path = "";
        for (int i = 0; i < params.length; i++) {
            if (params[i].equals("serialized")) {
                load = true;
                path = params[++i];
            } else if (params[i].equals("-serialize")) {
                save = true;
                path = params[++i];
            }
        }
//        System.err.println( "Load/save: " + load + " " + save );
        if (load) {
        	Params.addLog(3, String.valueOf("Sentence.class: reading sentence "+id));
            try {
                ObjectInputStream in = new ObjectInputStream(
                        new FileInputStream(path + File.separator + id));
                System.err.println( "File opened: " + path + File.separator + id );
                Sentence full = (Sentence) in.readObject();
                this.depTree = full.depTree;
                this.parseTree = full.parseTree;
                this.struct = full.struct;
                for (Interaction i : interactions) {
                    i.initInteraction();
                }
            } catch (UnsupportedOperationException e) {
                _log.error(e.getMessage());
                return;
            } catch (GraphException ex) {
                ex.printStackTrace();
            } catch (NoSuchWordException ex) {
                ex.printStackTrace();
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            } catch (ClassNotFoundException e) {
                e.printStackTrace();
            }
        } else {
            try {
                parser.parse(modText); // parses sentence modified with entities as ccoterm[x
                struct = parser.getStructuredSentence();

                depTree = parser.getDependencyTree();

                parseTree = parser.getParseTree();
                if (save) {
                    ObjectOutputStream out = null;
                    try {
                        File file = new File(path + File.separator + id);
                        if (file.exists()) {
                            file.delete();
                        }
                        System.out.println(file.getAbsolutePath());
                        file.createNewFile();
                        out = new ObjectOutputStream(new FileOutputStream(file));
                        out.writeObject(this);
                    } catch (IOException e) {
                        e.printStackTrace();
                    } finally {
                        if (out != null) {
                            out.close();
                        }
                    }
                }
                for (Iterator<Interaction> i = interactions.iterator(); i.hasNext();) {
                    try{
                        i.next().initInteraction();
                    }catch(NoSuchWordException e){
                        Logger.getLogger(Sentence.class.getName()).log(Level.SEVERE, id + ": " + text, e);
                        i.remove();
						System.err.println("Dependency graph:" );
						depTree.printGraph();
						System.err.println("Parse tree:" );
						parseTree.printGraph();
                    }catch (GraphException e ) {
						Logger.getLogger(Sentence.class.getName()).log(Level.SEVERE, id + ": " + text, e);
                        i.remove();
						System.err.println("Dependency graph:" );
						depTree.printGraph();
						System.err.println("Parse tree:" );
						parseTree.printGraph();
						}
					finally {
					}

                }
            } catch (StanfordParser.ParseException ex) {
                Logger.getLogger(Sentence.class.getName()).log(Level.SEVERE, id + ": " + text, ex);
                // try to continue
 
            }
        }

    }

    /* Old version
    @Override
    public void parse(Parser parser, String... params) {
    _log.info(text);
    boolean load = false;
    String path = "";
    for (int i = 0; i < params.length; i++) {
    if (params[i].equals("serialized")) {
    load = true;
    path = params[++i];
    }
    }

    for (Iterator<Interaction> it = interactions.iterator(); it.hasNext();) {
    // sentence parse could be taken from the file
    // or be parsed directly by parser, dependent on provided params
    Interaction i = it.next();
    //special case of interactions, that defined to be negative by default
    if (i.predefinedStatus != null) {
    i.status = i.predefinedStatus;
    continue;
    }
    if (load) {
    try {
    ObjectInputStream in = new ObjectInputStream(
    new FileInputStream(path.concat("/").concat(i.id)));
    Interaction full = (Interaction) in.readObject();
    //saving memory
    //i.parseTree = full.parseTree;
    //i.depTree = full.depTree;
    //i.struct = full.struct;
    i.depPath = full.depPath;
    i.parsePath = full.parsePath;
    } catch (UnsupportedOperationException e) {
    _log.error(e.getMessage());
    return;
    } catch (FileNotFoundException e) {
    e.printStackTrace();
    } catch (IOException e) {
    e.printStackTrace();
    } catch (ClassNotFoundException e) {
    e.printStackTrace();
    }
    } else {
    try {
    parser.parse(i.sentence);
    i.parseTree = parser.getParseTree();
    i.depTree = parser.getDependencyTree();
    i.struct = parser.getStructuredSentence();
    i.initInteraction();
    } catch (NoSuchWordException e) {
    _log.error(e.getMessage());
    it.remove();
    } catch (GraphException e) {
    _log.error("Can't define path for interaction between " + i.p1.name + " and " + i.p2.name);
    _log.error("Nested exception:" + e.getClass() + ":" + e.getMessage());
    it.remove();
    } catch (Exception e1) {
    _log.error(e1.getMessage());
    return;
    }
    }
    }
    }
     */
    @Override
    public void add(Text text) throws OperationNotSupportedException {
        throw new OperationNotSupportedException(
                "Sentence could not act as a container");
    }

    // -------------NON-INTERFACE PUBLIC METHODS-----------
    public void addProtein(Protein p) {
        proteins.add(p);
    }

    public Protein getProtein(String id) {
        for (Protein p : proteins) {
            if (p.getId().equals(id)) {
                return p;
            }
        }
        return null;
    }

    public List<Protein> getProteins() {
        return proteins;
    }

    public void addInteraction(Interaction i) {
        interactions.add(i);
    }
    //int interId = 0;

    /*This method is invoked when the sentence is extracted from xml, and positive
     *interaction instances are initialized. This method initializes negative
     *instances and prepare the sentence for parsing.
     *   Preparation for parsing means that protein names (quite complex sometimes)
     * are being substituted by shorter artificial words with predefined /NP tag.
     * The problem that occurs here is that some proteins are `nested' in others,
     * which requires some tweaking for interactions that involve them.
     *
     * Currently only the most outer protein name is taken into account. If
     * protein A is nested in protein B and interacts with a protein C then this
     * interaction is replaced with B-C interaction.
     *
     */
    public void populateInteractions(Preprocessor prep) {
        List<Protein> list = new ArrayList<Protein>(proteins);
        Collections.sort(list);
        Protein[] pArr = list.toArray(new Protein[0]);
        int i = 0;
        //marking nested proteins and removing them from a list
        while (i < (pArr.length-1)) {
            int j = i + 1;
            while (j<pArr.length && pArr[i].includes(pArr[j])) {
                if(pArr[j].equalsTo == null ||
                   pArr[j].equalsTo != null && pArr[i].includes(pArr[j].equalsTo)){
                    pArr[j].equalsTo = pArr[i];
                    list.remove(pArr[j]);
                }                
                j++;
            }
            i = j;
        }
        //the list was reduced, convert ot to array once again
        pArr = list.toArray(new Protein[0]);
        //iterate over all possible protein pairs (excluding self-pairing)
        //to populate the set of negative instances
        for (int k = 0; k < pArr.length - 1; k++) {
            for (int j = k + 1; j < pArr.length; j++) {
                Interaction ir = getInteraction(pArr[k], pArr[j]);
                if (ir == null) {
                    // if it's not found, then make a negative instance
                    ir = new Interaction(this);
                    ir.p1 = pArr[k];
                    ir.p2 = pArr[j];
                    ir.status = InteractionType.FALSE;
                    ir.id = id + ".i" + interactions.size();
                    ir.type = "no";
                    interactions.add(ir);
                }
            }
        }
        //process the sentence by replasing protein names with artificial ones
        prep.preprocess(this);
    }

    /* old version
    public void populateInteractions(Preprocessor prep) {
    //populating nested proteins array
    nestedProteins = new ArrayList<Protein>();
    for(Protein pp1 : proteins){
    for(Protein pp2 : proteins){
    if(!pp1.equals(pp2))
    if(pp1.includes(pp2)){
    nestedProteins.add(pp2);
    }
    }
    }
    interId = interactions.size();
    for (int i = 0, size = proteins.size(); i < size - 1; i++) {
    for (int j = i + 1; j < size; j++) {
    //cloning is necessary, because proteins list is shared among different interactions
    //and char offsets for each interaction are different
    List<Protein> localProteins = new ArrayList<Protein>();
    for(Protein p : proteins)
    localProteins.add((Protein)p.clone());
    Protein p1 = (Protein) localProteins.get(i);
    Protein p2 = (Protein) localProteins.get(j);
    //check if one of the proteins is nested, then it is a special type of interaction
    if(nestedProteins.contains(p1)||nestedProteins.contains(p2)){
    Interaction inter = getInteraction4Nested(p1, p2);
    inter.sentence = text;
    inter.sentId = id;
    inter.origSentId = originalId;
    continue;
    }

    //if (p1.startOffset[0] <= p2.startOffset[0]
    //		&& p2.endOffset[p2.endOffset.length - 1] <= p1.endOffset[p1.endOffset.length - 1]
    //		|| p2.startOffset[0] <= p1.startOffset[0]
    //		&& p1.endOffset[p1.endOffset.length - 1] <= p2.endOffset[p2.endOffset.length - 1]) {
    //	_log.error("Proteins "+ p1.name+ " and "+ p2.name+ " are nested, and interaction between them couldn't be established.");
    //	continue;
    //}

    Set<Protein> pSet = new HashSet<Protein>();
    pSet.add(p1);
    pSet.add(p2);
    _log.debug("Processing protein pair " + p1.name + " - "	+ p2.name);
    localProteins.removeAll(nestedProteins);
    Map<String, String> prepSent = prep.preprocessSentence(text, pSet, localProteins);
    String prepText = prepSent.get("sentence");
    prepSent.remove("sentence");
    Interaction inter = getInteraction(p1, p2);
    inter.abbr = prepSent;
    inter.sentence = prepText;
    inter.sentId = id;
    inter.origSentId = originalId;
    //statistics code
    //for(Protein pp1 : proteins){
    //	for(Protein pp2 : pSet){
    //		if(!pp1.equals(pp2))
    //			if(pp1.includes(pp2)){
    //				System.out.println(inter.id);
    //			}
    //	}
    //}
    //end statistics code
    }
    }
    }*/

    // -------------------PRIVATE METHOS-------------------
    /*private Interaction getInteraction4Nested(Protein p1, Protein p2) {
    Interaction i = new Interaction(this);
    i.p1 = p1;
    i.p2 = p2;
    i.predefinedStatus = InteractionType.FALSE;
    i.id = id + ".i" + interactions.size();
    i.type = "no";
    interactions.add(i);
    return i;
    }*/
    private Interaction getInteraction(Protein p1, Protein p2) {
        List<Protein> pList = new ArrayList<Protein>();
        pList.add(p1);
        pList.add(p2);
        // first try to find the protein pair among positive instances
        for (Interaction i : interactions) {
            if (pList.contains(i.p1) && pList.contains(i.p2)) {
                i.p1 = pList.get(pList.indexOf(i.p1));
                i.p2 = pList.get(pList.indexOf(i.p2));
                return i;
            }
        }
        return null;
    }

    private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
		if(depTree!=null){
			out.writeObject(depTree.serialize());
		}else{
			out.writeObject(null);
		}
		if(parseTree!=null){
			out.writeObject(parseTree.serialize());
		}else{
			out.writeObject(null);
		}
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		String str = (String)in.readObject();
		if(str!=null){
			BufferedReader reader = new BufferedReader(new StringReader(str));
			depTree = Graph.deserialize(reader);
			reader.close();
		}
		str = (String)in.readObject();
		if(str!=null){
			BufferedReader reader = new BufferedReader(new StringReader(str));
			parseTree = Graph.deserialize(reader);
			reader.close();
		}
    }
}
