package deeper;

import corpus.Sentence;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.io.StringReader;
import graph.Keys;
import edu.uci.ics.jung.graph.Vertex;

import util.Params;

import exceptions.GraphException;
import exceptions.NoSuchWordException;

import jung.entities.Graph;

public class Interaction implements Comparable<Interaction>, Serializable{
	
	private static final long serialVersionUID = 2L;
	
	//proteins that take part in interaction
	public Protein p1;
	public Protein p2;
	//interaction type as defined in xml dataset (physical, bind etc.)
	public String type;
	//interaction ID as defined in xml dataset
	public String id;
	//interaction class (positive, negative, straight, inverse)
	public InteractionType status;
	//for some interactions we know that they are negative in advance, so keep it
	public InteractionType predefinedStatus;

    public Sentence origin;
	
	//all parsing information
	public transient Graph depPath;
	public transient Graph parsePath;
	
	//interaction class, as predicted by classifier (used in UI)
	public transient String predClass;

    public Interaction(Sentence s){
        origin = s;
    }
	
	public void initInteraction() throws GraphException, NoSuchWordException {
		int ind1 = origin.struct.getIndex(p1.getName_mod())+1;
		int ind2 = origin.struct.getIndex(p2.getName_mod())+1;
		if(ind1==0 || ind2==0){
            System.err.println(origin.id()+" "+origin.text);
			throw new NoSuchWordException("Protein was not found. Indices "+ind1+", "+ind2);
        }else if(ind1==ind2){
            System.err.println(origin.id()+" "+origin.text);
			throw new NoSuchWordException("Proteins occur in one word. No path exists.");
		}
        Params.addLog( 3, "Interaction.class: "+id);
        Params.addLog( 3, "Protein1: "+p1 + ", id in Graph: " + ind1);
        Params.addLog( 3, "Protein2: "+p2 + ", id in Graph: " + ind2);
                
        
//        Graph.printVertex("origin.depTree.findVertex(ind1)", origin.depTree.findVertex(ind1));
//		Graph.printVertex("origin.depTree.findVertex(ind2)", origin.depTree.findVertex(ind2));
		Vertex v1 = origin.depTree.findVertex(ind1);
		Vertex v2 = origin.depTree.findVertex(ind2);
		depPath = origin.depTree.getPath(v1, v2);
		if ( depPath == null )
			throw new GraphException("Could not find an undirected path in dependency tree from node '" + 
				v1.getUserDatum(Keys.LABEL) + "' ("+v1.getUserDatum(Keys.INDEX)+
				") to node '" + v2.getUserDatum(Keys.LABEL) + "'("+v2.getUserDatum(Keys.INDEX)+
				") corresponding to interaction of id " + id + ".");	
		depPath.checkUniqueVertexIndex();
		
		v1 = origin.parseTree.findVertex(ind1);
		v2 = origin.parseTree.findVertex(ind2);
//		Graph.printVertex("origin.parseTree.findVertex(ind1)", origin.parseTree.findVertex(ind1));
//		Graph.printVertex("origin.parseTree.findVertex(ind2)", origin.parseTree.findVertex(ind2));
		parsePath = origin.parseTree.getPath(v1, v2);
		if ( parsePath == null )
			throw new GraphException("Could not find an undirected path in parse tree from node '" + 
				v1.getUserDatum(Keys.LABEL) + "' ("+v1.getUserDatum(Keys.INDEX)+
				") to node '" + v2.getUserDatum(Keys.LABEL) + "'("+v2.getUserDatum(Keys.INDEX)+
				") corresponding to interaction of id " + id + ".");
		parsePath.checkUniqueVertexIndex();

	}
	
/*	public void erase(){
		p1 = null;
		p2 = null;
		depPath = null;
		parsePath = null;
		depTree = null;
		parseTree = null;
		struct = null;
	}
*/
	public int compareTo(Interaction i) {
		if(status.ordinal()>i.status.ordinal()){
			return 1;
		}else if(status.ordinal()<i.status.ordinal()){
			return -1;
		}
		return 0;
	}
	
/*	public String getClearSentence(){
		return sentence.replace(p1.name_mod, p1.name).replace(p2.name_mod, p2.name);
	}
*/
	public String toString(){
		return id;
	}
	
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		String str = (String)in.readObject();
		if(str!=null){
			BufferedReader reader = new BufferedReader(new StringReader(str));
			depPath = Graph.deserialize(reader);
			reader.close();
		}
		str = (String)in.readObject();
		if(str!=null){
			BufferedReader reader = new BufferedReader(new StringReader(str));
			parsePath = Graph.deserialize(reader);
			reader.close();
		}
		//Limiting memory usage
		/*str = (String)in.readObject();
		if(str!=null){
			BufferedReader reader = new BufferedReader(new StringReader(str));
			depTree = Graph.deserialize(reader);
			reader.close();
		}
		str = (String)in.readObject();
		if(str!=null){
			BufferedReader reader = new BufferedReader(new StringReader(str));
			parseTree = Graph.deserialize(reader);
			reader.close();
		}*/
	}

	private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
		if(depPath!=null){
			out.writeObject(depPath.serialize());
		}else{
			out.writeObject(null);
		}
		if(parsePath!=null){
			out.writeObject(parsePath.serialize());
		}else{
			out.writeObject(null);
		}
		/*if(depTree!=null){
			out.writeObject(depTree.serialize());
		}else{
			out.writeObject(null);
		}
		if(parseTree!=null){
			out.writeObject(parseTree.serialize());
		}else{
			out.writeObject(null);
		}*/
	}
	
	public String getType(){
		String type = "";
		int real = status.ordinal();
		int pred = Integer.parseInt(predClass);
		if(real == InteractionType.FALSE.ordinal()){
			if(real == pred)
				type = "TN";
			else
				type = "FP";
		}else{
			if(real == pred)
				type = "TP";
			else
				if(pred==InteractionType.FALSE.ordinal())
					type = "FN";
				else
					type = "WD";
		}
		return type;
	}
}
