package jung.entities;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import edu.uci.ics.jung.graph.DirectedEdge;
import edu.uci.ics.jung.graph.DirectedGraph;
import edu.uci.ics.jung.graph.Edge;
import edu.uci.ics.jung.graph.Vertex;
import edu.uci.ics.jung.graph.impl.DirectedSparseEdge;
import edu.uci.ics.jung.graph.impl.SimpleUndirectedSparseVertex;
import edu.uci.ics.jung.graph.impl.SparseGraph;
import edu.uci.ics.jung.graph.impl.UndirectedSparseEdge;
import edu.uci.ics.jung.utils.Pair;
import edu.uci.ics.jung.utils.UserData;
import exceptions.GraphException;
import graph.Keys;

/**
 * Special kind of graph for storing dependency trees. It is a directed graph
 * with possible parallel arcs.
 * 
 * @author tfayruzo
 * 
 */
public class Graph extends SparseGraph implements DirectedGraph, Keys {
	private static Log _log = LogFactory.getLog(Graph.class);

	//optional root for parse trees
	//private Vertex root = null;

	// TODO: make it weak referenced
	private UndirectedGraph counterpart = null;

	/**
	 * Creates an instance of a sparse directed graph.
	 */
	public Graph() {
		super();
		Collection edge_predicates = getEdgeConstraints();
		edge_predicates.add(DIRECTED_EDGE);
	}
	
	/*public Graph(Vertex root){
		super();
		Collection edge_predicates = getEdgeConstraints();
		edge_predicates.add(DIRECTED_EDGE);
		this.root = root;
		addVertex(root);
	}*/

	// ------------------------NON INTERFACE PUBLIC METHODS----------------------
	public Vertex findVertex(Vertex v) {
		final String index = v.getUserDatum(INDEX).toString();
		for (Object o : getVertices()) {
			Vertex _vert = (Vertex) o;
			if (index.equals(_vert.getUserDatum(INDEX).toString())) {
				return _vert;
			}
		}
		return null;
	}

	public Vertex findVertex(int index) {
		for (Object o : getVertices()) {
			Vertex _vert = (Vertex) o;
			if (Integer.parseInt(_vert.getUserDatum(INDEX).toString())==index) {
				return _vert;
			}
		}
		return null;
	}

	public DirectedEdge findEdge(int i) {
		for (Object edge : mEdges) {
			DirectedEdge _edge = (DirectedEdge) edge;
			if (Integer.parseInt(_edge.getUserDatum(INDEX).toString()) == i) {
				return _edge;
			}
		}
		return null;
	}
	
	public static int getVertexIndex( Vertex v )
	{
		return Integer.parseInt(v.getUserDatum(Keys.INDEX).toString());
	}
	
	public static boolean isSourceOf(final Vertex v, final Edge e)
	{
		final int i = getVertexIndex(v);
		Pair endPair = e.getEndpoints();
		final int j = getVertexIndex((Vertex) endPair.getFirst());
		return ( i == j );			
	}
	
	public static boolean isTargetOf(final Vertex v, final Edge e)
	{
		final int i = getVertexIndex(v);
		Pair endPair = e.getEndpoints();
		final int j = getVertexIndex((Vertex) endPair.getSecond());
		return ( i == j );			
	}
	
	public void checkUniqueVertexIndex()
	{
		Set<Integer> vertexIndices = new HashSet<Integer>();
		for (Object o: getVertices())
			if ( !vertexIndices.add(getVertexIndex((Vertex)o)))
					throw new IllegalStateException( "Duplicate vertex index:" + getVertexIndex((Vertex)o) );
	}
	
	public DirectedEdge findEdge(final Vertex v1, final Vertex v2) {
		final int i1 = getVertexIndex(v1);
		final int i2 = getVertexIndex(v2);
//		System.err.println( "findEdge:" + ", i1: " + i1 + ", i2: " + i2 );
		for (Object o : v1.getIncidentEdges()) {
			DirectedSparseEdge e = (DirectedSparseEdge) o;
			Pair endPair = e.getEndpoints();
			final int j1 = getVertexIndex((Vertex) endPair.getFirst());
			final int j2 = getVertexIndex((Vertex) endPair.getSecond());
//			System.err.println( "Edge:" + e + ", j1: " + j1 + ", j2: " + j2  );
//			printEdge("Incident edge: ", e);
			if ( i1 == j1 && i2 == j2 )
				return e;
		}
		return null;
	}

    public Graph getPath(Vertex v1, Vertex v2) throws GraphException{
        if (v1 == null || v2 == null) {
			throw new NullPointerException("One of the vertices is null");
		}

		// both vertices should belong to dep tree
		if (v1.getGraph() != this || v2.getGraph() != this) {
			System.err.println("Given vertices does not belong to the graph. Trying to search them in the graph");
			boolean fail = false;
			if(v1.getUserDatum(INDEX)!=null && v2.getUserDatum(INDEX)!=null){
				v1 = findVertex(Integer.parseInt(v1.getUserDatum(INDEX).toString()));
				v2 = findVertex(Integer.parseInt(v2.getUserDatum(INDEX).toString()));
				if(v1==null || v2==null)
					fail = true;
			}else{
				fail = true;
			}
			if(fail){
				System.err.println("Path constructor: Vertices should belong to graph");
				return null;
			}
		}

        //build an undirected counterpart in order to get the shortest path
		if (counterpart == null) {
			counterpart = transform();
			checkCounterpart();
		}
		else
			checkCounterpart();
		// edges = undirected shortest path
		List<Edge> edges = counterpart.getPath(v1, v2);
		if ( edges == null )
	    	return null; // no path
		Graph result = new Graph();
        for(Edge e : edges){
            final Pair pair = e.getEndpoints();
            
            Vertex vOld_ = (Vertex) pair.getFirst();
            Vertex _vOld = (Vertex) pair.getSecond();

//            printVertex("vOld_ (1)", vOld_);
//            printVertex("_vOld (2)", _vOld );
            
            String direction = getDirection(Integer.parseInt(vOld_.getUserDatum(INDEX).toString()), Integer.parseInt(_vOld.getUserDatum(INDEX).toString()));
            if ( direction == null )
            	throw new IllegalStateException("Directed edge not found");
            if(direction.equals("in")){
                Vertex tmp = vOld_;
                vOld_ = _vOld;
                _vOld = tmp;
            }
            Vertex v_ = new DependencyVertex();
                        
            String oldIndexDatum = vOld_.getUserDatum(Keys.INDEX).toString();
            String oldLabelDatum = vOld_.getUserDatum(Keys.LABEL).toString();
            
            Object oldPOSDatum = vOld_.getUserDatum(Keys.POS);
            
            if ( v_ == null )
            	throw new IllegalStateException();
            v_.setUserDatum(Keys.INDEX, oldIndexDatum, UserData.SHARED);
            v_.setUserDatum(Keys.LABEL, oldLabelDatum, UserData.SHARED);
            if(oldPOSDatum!=null) //it is null for parse paths, which causes an exception in Jung
                v_.setUserDatum(Keys.POS, oldPOSDatum, UserData.SHARED);
            
            oldIndexDatum = _vOld.getUserDatum(Keys.INDEX).toString();
            oldLabelDatum = _vOld.getUserDatum(Keys.LABEL).toString();
            oldPOSDatum = _vOld.getUserDatum(Keys.POS);
            
            Vertex _v = new DependencyVertex();
            _v.setUserDatum(Keys.INDEX, oldIndexDatum, UserData.SHARED);
            _v.setUserDatum(Keys.LABEL, oldLabelDatum, UserData.SHARED);
            if(oldPOSDatum!=null) //it is null for parse paths, which causes an exception in Jung
                _v.setUserDatum(Keys.POS, oldPOSDatum, UserData.SHARED);
            Vertex tmp = result.findVertex(_v);
            if(tmp==null)
                result.addVertex(_v);
            else
                _v = tmp;
            tmp = result.findVertex(v_);
            if(tmp==null)
                result.addVertex(v_);
            else
                v_ = tmp;
            Edge eNew = new DirectedSparseEdge(v_,_v);
            eNew.setUserDatum(Keys.INDEX, (Integer)e.getUserDatum(Keys.INDEX), UserData.SHARED);
            if(e.getUserDatum(Keys.LABEL)!=null) //it is null for parse paths, which causes an exception in Jung
                eNew.setUserDatum(Keys.LABEL, (String)e.getUserDatum(Keys.LABEL), UserData.SHARED);
            result.addEdge(eNew);
        }
        return result;
    }
	
	public NewPath getNewPath(Vertex v1, Vertex v2) throws GraphException{
		if (v1 == null || v2 == null) {
			throw new NullPointerException("One of the vertices is null");
		}
		
		// both vertices sould belong to dep tree
		if (v1.getGraph() != this || v2.getGraph() != this) {
			_log.info("Given vertices does not belong to the graph. Trying to search them in the graph");
			boolean fail = false;
			if(v1.getUserDatum(INDEX)!=null && v2.getUserDatum(INDEX)!=null){
				v1 = findVertex((Integer)v1.getUserDatum(INDEX));
				v2 = findVertex((Integer)v2.getUserDatum(INDEX));
				if(v1==null || v2==null)
					fail = true;
			}else{
				fail = true;
			}
			if(fail){
				_log.error("Path constructor: Vertices should belong to graph");
				return null;
			}				
		}

        //build an undirected counterpart in order to get the shortest path
		if (counterpart == null) {
			counterpart = transform();
		}		
		List<Edge> edges = counterpart.getPath(v1, v2);		
		NewPath path = new NewPath(edges.size()+1);
		List<Vertex> vertices = new ArrayList<Vertex>();
		int i = 0;
		boolean first = true, last = false;
		Vertex tmp = null;
		for(Iterator<Edge> it = edges.iterator(); it.hasNext();){
			DirectedEdge edge = findEdge((Integer)it.next().getUserDatum(INDEX));
			if(!it.hasNext()){
				last = true;
			}
			path.edges[i] = edge;
			if(first){				
				vertices.add(v1);
				tmp = edge.getOpposite(v1);
				vertices.add(tmp);
				path.eMarks[i] = getDirection(Integer.parseInt(v1.getUserDatum(INDEX).toString()), Integer.parseInt(tmp.getUserDatum(INDEX).toString()));
				first = false;
			}else if(last){
				vertices.add(v2);
				path.eMarks[i] = getDirection(Integer.parseInt(tmp.getUserDatum(INDEX).toString()), Integer.parseInt(v2.getUserDatum(INDEX).toString()));
			}else{
				Vertex tmp2 = tmp;
				tmp = edge.getOpposite(tmp);				
				path.eMarks[i] = getDirection(Integer.parseInt(tmp2.getUserDatum(INDEX).toString()), Integer.parseInt(tmp.getUserDatum(INDEX).toString()));
				vertices.add(tmp);
			}
			i++;
		}
		
		i = 0;
		for(Vertex v : vertices){
			path.vertices[i] = v;
			if(v.equals(v1)){
				if(path.eMarks[i].equals("in"))
					path.vMarks[i] = "bottom";
				else
					path.vMarks[i] = "top";
			}else if(v.equals(v2)){
				if(path.eMarks[i-1].equals("in"))
					path.vMarks[i] = "top";
				else
					path.vMarks[i] = "bottom";
			}else{
				if(path.eMarks[i-1].equals(path.eMarks[i])){
					path.vMarks[i] = "inner";
				}else if(path.eMarks[i-1].equals("in")){
					path.vMarks[i] = "top";
				}else{
					path.vMarks[i] = "bottom";
				}
			}
			i++;
		}	
		return path;
	}

	// -------------------------PRIVATE METHODS-------------------------------
	/*
	 * Specific graph transformation, transforms directed graph to undirected
	 * counterpart. As far as I understood could not be performed by copy() or
	 * whatever methods in JUNG.
	 */
	private UndirectedGraph transform() {
		UndirectedGraph graph = new UndirectedGraph();
		for (Object o : getVertices()) {
			Vertex v = (Vertex) o;
			Vertex newV = new SimpleUndirectedSparseVertex();
			newV.setUserDatum(INDEX, v.getUserDatum(INDEX), UserData.SHARED);
            newV.setUserDatum(LABEL, v.getUserDatum(LABEL),	UserData.SHARED);
			if(v.getUserDatum(POS)!=null){
				newV.setUserDatum(POS, v.getUserDatum(POS),	UserData.SHARED);
			}
			graph.addVertex(newV);
		}
		for (Object o : getEdges()) {
			Edge e = (Edge) o;
			Vertex v1 = (Vertex) e.getEndpoints().getFirst();
			Vertex v2 = (Vertex) e.getEndpoints().getSecond();
			Edge newE = new UndirectedSparseEdge(graph.findVertex(v1), graph
					.findVertex(v2));
			newE.setUserDatum(INDEX, e.getUserDatum(INDEX), UserData.SHARED);
			if( e.getUserDatum(LABEL)!=null){
				newE.setUserDatum(LABEL, e.getUserDatum(LABEL), UserData.SHARED);
			}			
			Edge newEdge = graph.addEdge(newE);
//			Graph.printVertex("orig Edge1", v1 );
//			Graph.printVertex("orig Edge2", v2 );
//			Graph.printVertex("new Edge1", graph.findVertex(v1));
//			Graph.printVertex("new Edge2", graph.findVertex(v2));
		}
		return graph;
	}
	
	public void checkCounterpart( ){
	if ( this.getVertices().size() != counterpart.getVertices().size())
		throw new IllegalStateException( "Vertex count mismatch");
	if ( this.getEdges().size() != counterpart.getEdges().size())
		throw new IllegalStateException( "Edge count mismatch");
	}
	
	// returns "out" if v1 dominates v2, "in" - otherwise  
	private String getDirection(int vInd1, int vInd2){
//		System.err.println("getDirection(int vInd1, int vInd2)" + vInd1 + ", " + vInd2);
		Vertex v1, v2;
		DirectedEdge e;

		v1 = findVertex(vInd1);
		v2 = findVertex(vInd2);
		
//		printVertex("findVertex(vInd1)", v1);
//		printVertex("findVertex(vInd2)", v2);
		e = this.findEdge(v1, v2);
		if ( e != null )
		{
			if(e.getSource().equals(v1))
				return "out";
			else
				return "in";
		}
			
		e = this.findEdge(v2, v1);
		if ( e != null )
		{
			if(!e.getSource().equals(v1))
				return "out";
			else	
				return "in";
		}
		//edge not found
		return null;
	}

	public String serialize() {
		StringBuffer sb = new StringBuffer();
		/*if(root!=null){
			sb.append("\nROOT\n");
			sb.append(root.getUserDatum(INDEX));
			sb.append("\t");
			if(root.getUserDatum(LABEL)==null)
				sb.append(root.getUserDatum("null"));
			else
				sb.append(root.getUserDatum(LABEL));
			sb.append("\t");
			if(root.getUserDatum(POS)==null)
				sb.append(root.getUserDatum("null"));
			else
				sb.append(root.getUserDatum(POS));
			sb.append("\n");
		}*/
		sb.append("\nVERTICES\n");
		for (Object o : mVertices) {
			Vertex v = (Vertex) o;
			sb.append(v.getUserDatum(INDEX));
			sb.append("\t");
			if(v.getUserDatum(LABEL)==null)
				sb.append(v.getUserDatum("\\null"));
			else
				sb.append(v.getUserDatum(LABEL));
			sb.append("\t");
			if(v.getUserDatum(POS)==null)
				sb.append(v.getUserDatum("\\null"));
			else
				sb.append(v.getUserDatum(POS));			
			sb.append("\n");
		}
		sb.append("\nEDGES\n");
		for (Object o : mEdges) {
			DirectedEdge e = (DirectedEdge) o;
			sb.append(e.getSource().getUserDatum(INDEX));
			sb.append("\t");
			sb.append(e.getDest().getUserDatum(INDEX));
			sb.append("\t");
			sb.append(e.getUserDatum(INDEX));
			sb.append("\t");
			if(e.getUserDatum(LABEL)==null)
				sb.append(e.getUserDatum("\\null"));
			else
				sb.append(e.getUserDatum(LABEL));
			sb.append("\n");
		}
		sb.append("\n");
		return sb.toString();
	}

	public static Graph deserialize(BufferedReader stream) throws IOException {
		Graph g = null;
		String s;
		boolean vertices = false;
		boolean edges = false;
		//boolean root = false;
		while ((s = stream.readLine()) != null) {
			if(s.equals("")){
				continue;
			}
			if (s.equals("EDGES")) {
				edges = true;
				vertices = false;
				//root = false;
				continue;
			}else if(s.equals("VERTICES")){
				if(g == null)
					g = new Graph();
				edges = false;
				vertices = true;
				//root = false;
				continue;
			}/*else if(s.equals("ROOT")){
				edges = false;
				vertices = false;
				root = true;
				continue;
			}*/
			String[] split = s.split("\t");
			if (vertices) {
				Vertex v = new DependencyVertex();
				v.setUserDatum(INDEX, new Integer(split[0]), UserData.SHARED);
				if(!split[1].equals("\\null"))
					v.setUserDatum(LABEL, split[1], UserData.SHARED);
				if(!split[2].equals("\\null"))
					v.setUserDatum(POS, split[2], UserData.SHARED);
				g.addVertex(v);
			} else if(edges){
				Vertex v1 = g.findVertex(Integer.parseInt(split[0]));
				Vertex v2 = g.findVertex(Integer.parseInt(split[1]));
				DirectedEdge e = new DirectedSparseEdge(v1, v2);
				e.setUserDatum(INDEX, new Integer(split[2]), UserData.SHARED);
				if(!split[3].equals("\\null"))
					e.setUserDatum(LABEL, split[3], UserData.SHARED);
				g.addEdge(e);
			}/*else if(root){
				Vertex r = new DependencyVertex();
				r.setUserDatum(INDEX, new Integer(split[0]), UserData.SHARED);
				if(!split[1].equals("null"))
					r.setUserDatum(LABEL, split[1], UserData.SHARED);
				if(!split[2].equals("null"))
					r.setUserDatum(POS, split[2], UserData.SHARED);
				g = new Graph(r);
			}*/
		}
		return g;
	}
	
	public static void printVertex(String name, Vertex v)
	{
        System.err.println("vertex: " + name + ", id: " + v.toString());
        for ( Iterator i = v.getUserDatumKeyIterator(); i.hasNext(); )
        {
        	Object Key = i.next();
        	System.err.println("vertex: " + name + ", id: " + v.toString() + 
        			" " + Key + "=" + v.getUserDatum(Key));
        }
	}

	public static void printEdge(String name, Edge e )
	{
        System.err.println("edge: " + name + ", id: " + e.toString());
        for ( Iterator i = e.getUserDatumKeyIterator(); i.hasNext(); )
        {
        	Object Key = i.next();
        	System.err.println("edge: " + name + ", id: " + e.toString() + 
        			" " + Key + "=" + e.getUserDatum(Key));
        }
	}

	/**
	 * Ensure an edge connecting two vertices of <code>e</code> exists in graph.
	 * @note edge is looked up based on vertex indices and not edge index
	 * @param e
	 * @return an edge in graph
	 * @throws IllegalArgumentException if either vertices is not present in graph
	 */
	public Edge ensureEdge( Edge e ) {
		Vertex v, w;
		v = (Vertex) e.getEndpoints().getFirst();
		w = (Vertex) e.getEndpoints().getSecond();		
		v = findVertex(v);
		w = findVertex(w);
		if ( v == null || w == null )
			throw new IllegalArgumentException( "Vertices must be present in graph!");
		Edge f = findEdge( v, w );
		// not f, because f is in another graph (jung feature)
		return f != null? f: addEdge(e);
	}
	
    public Vertex ensureVertex(Vertex v)
    {
    	Vertex w = findVertex(v);
    	if ( w != null ) {
    		if ( !v.getUserDatum(Keys.LABEL).equals(w.getUserDatum(Keys.LABEL)))
    			throw new IllegalStateException("Vertex labels do not match!" + w.getUserDatum(Keys.LABEL) + ", "  + v.getUserDatum(Keys.LABEL) );
    		// found
    		return w;
    	}
    	return addVertex(v);
    }

    public void printVertices()
    {
    	int i=0;
    	for ( Object o: getVertices())
    		printVertex("vertex" + (++i), (Vertex) o);
    }

    public void printEdges()
    {
    	int i=0;
    	for ( Object o: getEdges())
    		printEdge("edge" + (++i), (Edge) o);
    }

    public void printGraph()
    {
    	printVertices();
		printEdges();
    }

}