/* tree.c
 *
 *  Written by Mark Johnson, 27th May 1997
 *         and Peter Palaga 2009-04-13
 *             create_q_profile() based on Tetsui Kuboyama's Ruby implementation
 *             see also
 *             Kuboyama, T., Kashima, H., Aoki-Kinoshita, K.F., Hirata, K., Yasuda, H. (2007):
 *             A Spectrum Tree Kernel,
 *             Journal of Japanese Society of Artificial Intelligence, Vol.22, No.2, 2007.
 *
 */

#include <stdlib.h>
#include <assert.h>
#include <ctype.h>
#include "tree.h"
#include "svm_common.h"
#include "hashtable/hashtable_itr.h"

void
consumespaces(char* src, int* pos)
{
  int c;
  /* skip spaces */
  while (1)
    {
      c = src[*pos];
      if (!isspace(c))
        {
          break;
        }
      *pos = (*pos) + 1;
    }
}

string_index_key_t
parselabel(char* src, int* pos, unsigned char is_terminal, STRING_INDEX *string_index)
{
  int c, start_pos, len;

  start_pos = *pos;

  while (1)
    {
      c = src[*pos];
      if (c == '\0' || isspace(c) || c == '(' || c == ')')
        {
          break;
        }
      *pos = (*pos) + 1;
    };

  len = (*pos) - start_pos;
  assert(len> 0);

  consumespaces(src, pos);

  return index_for_value(src + start_pos, len, is_terminal, string_index);
}

TREE *
parsetree(char *src, int *pos, node_count_t *node_count, tree_depth_t depth,
    tree_depth_t* max_depth, STRING_INDEX *string_index)
{
  int c;
  TREE* t;
  TREE* p;

  c = src[*pos];
  assert(c != '\0'); /* file ended w/out closing ')' */

  switch (c)
    {

  case ')':
    *pos = (*pos) + 1;
    consumespaces(src, pos);
    return (NULL);
    break; /* empty tree */

  case '(': /* nonterminal */
    *pos = (*pos) + 1;
    consumespaces(src, pos);
    t = (TREE*) my_malloc(sizeof(TREE));
    t->subtrees = NULL;
    t->sibling = NULL;
    if ((*node_count) + 1 > node_count_max)
      {
        perror("node_count overflow.");
      }
    *node_count = (*node_count) + 1;
    t->label = parselabel(src, pos, 0, string_index);
    t->depth = depth;
    if ((*max_depth) < depth)
      {
        *max_depth = depth;
      }

    t->subtrees = p = parsetree(src, pos, node_count, depth + 1, max_depth,
        string_index);

    while (p)
      {
        p->sibling = parsetree(src, pos, node_count, depth + 1, max_depth,
            string_index);
        p = p->sibling;
      }
    return (t);
    break;

  default:
    /* terminal */
    t = (TREE*) my_malloc(sizeof(TREE));
    t->subtrees = NULL;
    t->sibling = NULL;
    if ((*node_count) + 1 > node_count_max)
      {
        perror("node_count overflow.");
      }
    *node_count = (*node_count) + 1;
    t->label = parselabel(src, pos, 1, string_index);
    t->depth = depth;
    if ((*max_depth) < depth)
      {
        *max_depth = depth;
      }

    consumespaces(src, pos);

    /* check the closing bracket */
    c = src[*pos];
    *pos = (*pos) + 1;
    assert(c==')');
    *pos = (*pos) - 1;
    return (t);
    }
}

void
free_tree(TREE* t)
{
  if (t != NULL)
    {
      free_tree(t->sibling);
      free_tree(t->subtrees);
      free(t);
    }
}

void
add_tree_nodes(TREE* t, TREE** nodes, node_count_t* node_index)
{
  if (t->subtrees != NULL)
    {
      TREE* subtree = t->subtrees;

      while (subtree != NULL)
        {
          add_tree_nodes(subtree, nodes, node_index);
          subtree = subtree->sibling;
        }
    }
  nodes[*node_index] = t;
  *node_index = (*node_index) + 1;
  /*printf("array_element = %u\n", *node_index);*/
}
void
print_tree(TREE* t, STRING_INDEX* string_index)
{
  char *label;
  label = value_for_index(t->label, string_index);

  if (t->subtrees)
    {
      TREE* p;

      printf("(%s{%d} ", label, t->depth);
      print_tree(p = t->subtrees, string_index);

      while ((p = p->sibling))
        {
          printf(" ");
          print_tree(p, string_index);
        }
      printf(")");
    }
  else
    {
      printf("%s{%d}", label, t->depth);
    }
}


Q_PATH *
concat_q_path(Q_PATH* q_path, string_index_key_t next_element, unsigned char q_gram_length)
{
  int i;
  Q_PATH *result;
  if (q_path == NULL)
    {
      result = create_q_path(q_gram_length);
      result->labels[0] = next_element;
      result->length = 1;
      return result;
    }
  else
    {
      if (q_path->length >= q_gram_length)
        {
          result = create_q_path(q_gram_length + 1);
        }
      else
        {
          result = create_q_path(q_path->length);
        }
      for (i = 0; i < q_path->length; ++i)
        {
          result->labels[i] = q_path->labels[i];
        }
      result->labels[i++] = next_element;
      result->length = i;
      return result;
    }
}

Q_PATH *
q_subpath(unsigned char start, unsigned char end, Q_PATH* q_path, unsigned char q_gram_length)
{
  Q_PATH* result;
  int i, n;
  assert(q_path != NULL);
  assert(start < end);
  /*assert(end <= q_path->length || end <= q_gram_length TODO: check if this is not too liberal );*/
  n = end - start;

  result = create_q_path(n > q_gram_length ? n : q_gram_length);
  result->length = n;
  for (i = 0; i < n; ++i)
    {
      result->labels[i] = q_path->labels[start + i];
    }
  return result;
}
void print_q_gram(Q_GRAM* q_gram, STRING_INDEX* string_index)
{
  printf("%u|", q_gram->pattern_id);
  print_q_path(q_gram->q_path, string_index);
}
void
print_q_path(Q_PATH* q_path, STRING_INDEX* string_index)
{
  int i;
  if (q_path != NULL)
    {
      for (i = 0; i < q_path->length; ++i)
        {
          printf("%u", q_path->labels[i]);
        }
    }
}
Q_PATH *
q_path_dup(Q_PATH* q_path, unsigned char q_gram_length)
{
  Q_PATH* result;
  int i;
  assert(q_path != NULL);
  result = create_q_path(q_path->length > q_gram_length ? q_path->length
      : q_gram_length);
  result->length = q_path->length;
  for (i = 0; i < q_path->length; ++i)
    {
      result->labels[i] = q_path->labels[i];
    }
  return result;
}

Q_PATH *
create_q_path(unsigned char q_path_capacity)
{
  Q_PATH *result;
  int i;
  result = (Q_PATH*) my_malloc(sizeof(Q_PATH));
  result->length = 0;
  result->labels = (string_index_key_t*) my_malloc(q_path_capacity
      * sizeof(string_index_key_t));
  for (i = 0; i < q_path_capacity; ++i)
    {
      result->labels[i] = ZERO_INDEX;
    }
  return result;
}

void
free_q_path(Q_PATH* q_path)
{
  if (q_path != NULL)
    {
      if (q_path->labels != NULL)
        {
          free(q_path->labels);
        }
      free(q_path);
    }
}

void
free_q_path_void(void* q_path)
{
  free_q_path((Q_PATH*) q_path);
}

Q_GRAM *
create_q_gram(unsigned char pattern_id, Q_PATH* q_path)
{
  Q_GRAM* result;
  result = (Q_GRAM*) my_malloc(sizeof(Q_GRAM));
  result->pattern_id = pattern_id;
  result->q_path = q_path;
  return result;
}

void
free_q_gram_void(void* q_gram)
{
  free_q_gram((Q_GRAM*) q_gram);
}

void
free_q_gram(Q_GRAM* q_gram)
{
  if (q_gram != NULL)
    {
      if (q_gram->q_path != NULL)
        {
          free_q_path(q_gram->q_path);
        }
      free(q_gram);
    }
}

unsigned char compute_first(char operation) {
  switch (operation) {
    case 'a':
    case '1':
    case 'm':
      return 1;
      break;
    case '2':
      return 0;
      break;
    default:
      perror("Unexpected second kernel operation.");
      exit(-1);
  }
}

unsigned char compute_second(char operation) {
  switch (operation) {
    case 'a':
    case '2':
    case 'm':
      return 1;
      break;
    case '1':
      return 0;
      break;
    default:
      perror("Unexpected second kernel operation.");
      exit(-1);
  }
}

SPECTRUM_TREE * create_spectrum_tree(char *source, STRING_INDEX* string_index, unsigned char q_gram_length, char kernel_combine_operation, char bow_only_terminals, char custom_kernel_input_format)
{
  SPECTRUM_TREE *result;
  node_count_t node_index = 0;
  int pos = 0;
  result = NULL;

  if (source && source[0] != '\0')
    {
      /*printf("%s\n", source);*/
      result = (SPECTRUM_TREE *) my_malloc(sizeof(SPECTRUM_TREE));
      result->node_count = 0;
      result->max_depth = 0;

      if (custom_kernel_input_format == 't') {
        result->root = parsetree(source, &pos, &(result->node_count), 0, &(result->max_depth), string_index);
        result->nodes = (TREE**) my_malloc((result->node_count) * sizeof(TREE*));
        add_tree_nodes(result->root, result->nodes, &node_index);
        /* check if we have read as many nodes as expected */
        assert(node_index == result->node_count);
      }
      else if (custom_kernel_input_format == 'q') {
        result->root = NULL;
        result->nodes = NULL;
      }
      else {
        perror("Unexpected custom_kernel_input_format value.");
        exit(-1);
      }

      result->norm_square = UNINITIALIZED_KERNEL_VALUE_DBL;
      result->norm_square_spectrum = UNINITIALIZED_KERNEL_VALUE;

      if (compute_first(kernel_combine_operation)){
        if (custom_kernel_input_format == 't') {
          result->q_profile = create_q_profile(result, q_gram_length, string_index, q_gram_length - 1);
        }
        else if (custom_kernel_input_format == 'q') {
          /* each odd index in the q-gram is the dependency
           * so there are (q_gram_length - 1)/2 dependencies
           * each dependency may have up_bottom or bottom_up direction = 2 possibilities
           * so we need exactly (q_gram_length - 1)/2 bits to store all possible path patterns
           * and so many bits means  pow(2, so_many) possibilities */
          result->q_profile = read_q_profile(source, q_gram_length, pow(2, (q_gram_length - 1)/2));
        }
        else {
          perror("Unexpected custom_kernel_input_format value.");
          exit(-1);
        }
      }
      else {
        result->q_profile = NULL;
      }

      if (compute_second(kernel_combine_operation))
        {
          result->bag_of_words = create_bag_of_words(result->nodes, result->node_count, bow_only_terminals, string_index);
        }
      else
        {
          result->bag_of_words = NULL;
        }

      /*
       print_tree(result->root, string_index);
       printf("\n----- mode_count = %u; max_depth = %u \n", result->node_count, result->max_depth);
       */

    }

  return result;
}

struct hashtable*
create_bag_of_words(TREE** nodes, node_count_t node_count, char bow_only_terminals, STRING_INDEX* string_index)
{
  struct hashtable* result;
  TREE* tree;
  int i;
  unsigned char* val;
  result = create_hashtable(node_count, index_hash, index_equal, free);

  for (i = 0; i < node_count; ++i)
    {
      tree = nodes[i];
      if (bow_only_terminals == 'a' || (bow_only_terminals == 't' && tree->subtrees == NULL))
        {
          /* a terminal symbol */
          if (hashtable_search(result, &(tree->label)) == NULL)
            {
              /* not there yet */
              string_index_key_t* key = (string_index_key_t*) my_malloc(
                  sizeof(string_index_key_t));
              *key = tree->label;
              val = (unsigned char*) my_malloc(sizeof(unsigned char));
              *val = 1;
              hashtable_insert(result, key, val);
            }
        }
    }

  return result;
}

void
print_escaped_element(char* str)
{
  int i;
  i = 0;
  while (str[i] != '\0')
    {
      if (str[i] == ',')
        {
          putc('_', stdout);
        }
      else
        {
          putc(str[i], stdout);
        }
      i++;
    }
}
void
test_q_profile(Q_PROFILE* q_profile, unsigned char q,
    STRING_INDEX* string_index)
{
  int i, j, ii;
  struct hashtable_itr *itr;
  Q_PATH* q_path;
  unsigned char val;

  assert(q_profile != NULL);
  /*printf("[");*/
  for (i = 0; i < q - 1; ++i)
    {
      if (i != 0)
        {
          /*printf(" ,");*/
        }
      /*printf("{");*/

      if (hashtable_count(q_profile->q_patterns[i]) != 0)
        {
          itr = hashtable_iterator(q_profile->q_patterns[i]);
          j = 0;
          do
            {
              q_path = (Q_PATH*) hashtable_iterator_key(itr);
              val = *((unsigned char*) hashtable_iterator_value(itr));

              printf("insert into kv1 (id, k, cnt) values (%d, '", i);
              /*if (j++ != 0) {
               printf(" ,");
               printf("\"");
               }*/
              if (q_path != NULL)
                {
                  for (ii = 0; ii < q_path->length; ++ii)
                    {
                      printf("%d", q_path->labels[ii]);
                    }
                }
              /*
               printf("\"");
               printf("=>%u", val);
               */
              printf("', %u);\n", val);
            }
          while (hashtable_iterator_advance(itr));
          free(itr);
        }
      /*printf("}");*/
    }
  /*printf("]\n");*/
  exit(-1);
}

int norm_square_spectrum(SPECTRUM_TREE* spectrum_tree, unsigned char q, unsigned char spectrum_tree_kernel_count_mode)
{
  assert(spectrum_tree != NULL);
  if (spectrum_tree->norm_square_spectrum == UNINITIALIZED_KERNEL_VALUE)
    {
      spectrum_tree->norm_square_spectrum = count_common_q_grams(spectrum_tree, spectrum_tree, q, spectrum_tree_kernel_count_mode);
    }
  return spectrum_tree->norm_square_spectrum;
}

int norm_square_bow(SPECTRUM_TREE* spectrum_tree)
{
  return hashtable_count(spectrum_tree->bag_of_words);
}

int count_common_words(SPECTRUM_TREE* spectrum_tree1, SPECTRUM_TREE* spectrum_tree2, char bow_only_terminals)
{
  int result = 0;
  struct hashtable_itr *itr;
  struct hashtable* ht2;
  unsigned char* val1;
  Q_PATH* key1;
  unsigned char* val2;

  if (spectrum_tree1 == spectrum_tree2) {
    return hashtable_count(spectrum_tree1->bag_of_words);
  }

  /* take the one with less nodes */
  if (spectrum_tree1->node_count <= spectrum_tree2->node_count)
    {
      if (hashtable_count(spectrum_tree1->bag_of_words) > 0)
        {
          itr = hashtable_iterator(spectrum_tree1->bag_of_words);
          assert(itr != NULL);
          ht2 = spectrum_tree2->bag_of_words;
          assert(ht2 != NULL);
          do
            {
              val1 = hashtable_iterator_value(itr);
              if (val1 != NULL)
                {
                  key1 = hashtable_iterator_key(itr);
                  assert(key1 != NULL);
                  val2 = hashtable_search(ht2, key1);
                  if (val2 != NULL)
                    {
                      /*printf("result += %d * %d\n", (*val1), (*val2));*/
                      result += (*val1) * (*val2);
                      /*printf("result = %d\n", result);*/
                    }
                }
            }
          while (hashtable_iterator_advance(itr));
          free(itr);
        }
      return result;
    }

  else
    {
      /* swap */
      return count_common_words(spectrum_tree2, spectrum_tree1, bow_only_terminals);
    }

}
int
count_common_q_grams(SPECTRUM_TREE* spectrum_tree1,
    SPECTRUM_TREE* spectrum_tree2, unsigned char q, unsigned char spectrum_tree_kernel_count_mode)
{
  int result = 0;
  int i = 0;
  struct hashtable_itr *itr;
  struct hashtable* ht2;
  unsigned char* val1;
  Q_PATH* key1;
  unsigned char* val2;

  /* take the one with less nodes */
  if (spectrum_tree1->node_count <= spectrum_tree2->node_count)
    {
      for (i = 0; i < spectrum_tree1->q_profile->q_patterns_count; i++)
        {
          if (hashtable_count(spectrum_tree1->q_profile->q_patterns[i]) > 0)
            {
              itr
                  = hashtable_iterator(spectrum_tree1->q_profile->q_patterns[i]);
              assert(itr != NULL);
              ht2 = spectrum_tree2->q_profile->q_patterns[i];
              assert(ht2 != NULL);
              do
                {
                  val1 = hashtable_iterator_value(itr);
                  if (val1 != NULL)
                    {
                      key1 = hashtable_iterator_key(itr);
                      assert(key1 != NULL);
                      val2 = hashtable_search(ht2, key1);
                      if (val2 != NULL)
                        {
                          /*printf("result += %d * %d\n", (*val1), (*val2));*/
                          if (spectrum_tree_kernel_count_mode == 'm') {
                            result += (*val1) * (*val2);
                          }
                          else if(spectrum_tree_kernel_count_mode == 'u') {
                            result++;
                          }
                          else {
                            perror("Unexpected value of spectrum_tree_kernel_count_mode.");
                            exit(-1);
                          }
                          /*printf("result = %d\n", result);*/
                        }
                    }
                }
              while (hashtable_iterator_advance(itr));
              free(itr);
            }
        }

      return result;
    }
  else
    {
      /* swap */
      return count_common_q_grams(spectrum_tree2, spectrum_tree1, q, spectrum_tree_kernel_count_mode);
    }
}

unsigned int read_int(char* src, int* pos) {
  char c;
  char* tmp;
  int i;
  int start;
  int len;
  int result;
  start = *pos;

  /* at least one digit */
  c = src[*pos];
  assert(isdigit(c));
  *pos = *pos +1;

  while (1) {
    c = src[*pos];
    if (!isdigit(c)) {
      break;
    }
    *pos = *pos +1;
  }

  len = *pos - start;
  tmp = (char*) my_malloc((len + 1) * sizeof(char));
  for (i = 0; i < len; ++i) {
    tmp[i] = src[start + i];
  }
  tmp[i] = '\0';

  result = atoi(tmp);
  free(tmp);
  return result;
}

Q_GRAM* read_q_gram(char* src, int* pos, int q) {
  int labels_read = 0;
  int nr;
  Q_GRAM* result;
  char c;

  consumespaces(src, pos);

  c = src[*pos];
  if (c == '\0') {
    /* end of input */
    return NULL;
  }

  assert(c == '(');
  *pos = *pos +1;

  result = (Q_GRAM*) my_malloc(sizeof(Q_GRAM));
  nr = read_int(src, pos);
  assert(nr <= UCHAR_MAX);
  result->pattern_id = nr;

  result->q_path = (Q_PATH*) my_malloc(sizeof(Q_PATH));
  result->q_path->length = q;
  result->q_path->labels = (string_index_key_t*) my_malloc(q* sizeof(string_index_key_t));

  while (1) {
    c = src[*pos];
    if (c == ')') {
      /* end of q_gram */
      *pos = *pos +1;
      break;
    }

    assert(labels_read < q);

    consumespaces(src, pos);

    nr = read_int(src, pos);
    assert(nr <= string_index_max_capacity);
    result->q_path->labels[labels_read++] = nr;

  };
  assert(labels_read == q);

  return result;
}

Q_PROFILE * read_q_profile(char* src, unsigned char q_gram_length, unsigned int q_patterns_count) {
  Q_PROFILE *result;
  Q_GRAM* q_gram;
  int q_grams_read = 0;
  int nr;
  int i;
  char c;
  unsigned char* val;
  int* pos;

  result = (Q_PROFILE*) my_malloc(sizeof(Q_PROFILE));
  result->q_patterns_count = q_patterns_count;
  result->q_patterns = (struct hashtable**) my_malloc(q_patterns_count * sizeof(struct hashtable*));
  for (i = 0; i < q_patterns_count; i++)
    {
      result->q_patterns[i] = create_hashtable(
          pow(2, (q_gram_length+1)/2) * 50 , /* 50 is the typical sentence length */
          path_hash, path_equal, free_q_path_void
      );
    }

  pos = (int*) my_malloc(sizeof(int));
  *pos = 0;
  while (1) {
    consumespaces(src, pos);
    q_gram = read_q_gram(src, pos, q_gram_length);
    if (q_gram == NULL) {
      break;
    }
    q_grams_read++;

    c = src[*pos];
    assert(c == ':');
    *pos = *pos +1;
    nr = read_int(src, pos);
    assert(nr <= UCHAR_MAX);

    val = (unsigned char*)hashtable_search(result->q_patterns[q_gram->pattern_id], q_gram->q_path);
    if (val == NULL) {
      val = (unsigned char*)my_malloc(sizeof(unsigned char));
      *val = 0;
      hashtable_insert(result->q_patterns[q_gram->pattern_id], q_gram->q_path, val);
      free(q_gram);
    }
    else {
      free_q_gram(q_gram);
    }
    *val = *val + nr;
  }
  assert(q_grams_read > 0);
  free(pos);
  return result;
}

Q_PROFILE * create_q_profile(SPECTRUM_TREE* spectrum_tree, unsigned char q, STRING_INDEX* string_index, unsigned int q_patterns_count)
{
  Q_PROFILE *result;
  int i = 0;
  int j = 0;
  int ii = 0;
  int d = 0;
  string_index_key_t l;
  int k = 0;
  int q_maxd_min = 0;
  int m = 0;
  int** cnt;
  int* cnt_element_sizes;
  int** sft;
  int* sft_element_sizes;
  int new_size;
  struct hashtable*** f; /* two-dimensional array of hashtables
   with q_path* keys and unsigned char* values
   the value pointers will not be shared with other hashtables
   the value pointers should be freed on every remove
   and on hashtable_destroy
   */
  struct hashtable** label; /* one-dimensional array of hashtables
   with q_gram* keys and unsigned char* values
   the value pointers will not be shared with other hashtables
   the value pointers should be freed on every remove
   and on hashtable_destroy
   */
  unsigned char f_val;
  TREE* t;
  struct hashtable_itr *itr;
  Q_PATH* w;
  Q_PATH* wl;
  Q_PATH* q_path;
  Q_GRAM* q_gram;
  unsigned char* val;
  Q_GRAM* ptn;
  /*unsigned char ruby_test = 1;*/

  /*printf("create_q_profile q= %u\n", q);*/

  result = (Q_PROFILE*) my_malloc(sizeof(Q_PROFILE));
  result->q_patterns_count = q_patterns_count;
  result->q_patterns = (struct hashtable**) my_malloc(q_patterns_count * sizeof(struct hashtable*));
  for (i = 0; i < q_patterns_count; i++)
    {
      result->q_patterns[i] = create_hashtable(spectrum_tree->node_count,
          path_hash, path_equal, free_q_path_void);
    }

  cnt = (int**) my_malloc((spectrum_tree->max_depth + 1) * sizeof(int*));
  cnt_element_sizes = (int*) my_malloc((spectrum_tree->max_depth + 1)
      * sizeof(int));
  for (d = (spectrum_tree->max_depth); d >= 0; d--)
    {
      /*printf("ctn d = %d\n", d);*/
      cnt_element_sizes[d] = 0;
      cnt[d] = NULL;

      q_maxd_min = (q - 1 < spectrum_tree->max_depth ? q - 1
          : spectrum_tree->max_depth);
      /*printf("q_maxd_min = %d\n", q_maxd_min);*/
      for (k = 1; k <= q_maxd_min; k++)
        {
          m = d - q + 1 + (2 * k);
          if (k <= m && m <= spectrum_tree->max_depth)
            {
              new_size = cnt_element_sizes[d] + 2;
              /*printf("resizing cnt[%d] to %d, << %d, %d\n", d, new_size, m-1, k-1);*/
              cnt[d] = (int*) my_realloc(cnt[d], new_size * sizeof(int));
              cnt[d][cnt_element_sizes[d]] = m - 1;
              cnt[d][cnt_element_sizes[d] + 1] = k - 1;
              cnt_element_sizes[d] = new_size;
            }
        }
    }

  sft = (int**) my_malloc((spectrum_tree->max_depth) * sizeof(int*));
  sft_element_sizes
      = (int*) my_malloc((spectrum_tree->max_depth) * sizeof(int));
  sft_element_sizes[0] = 0;
  sft[0] = NULL;
  for (d = (spectrum_tree->max_depth - 1); d >= 1; d--)
    {
      sft_element_sizes[d] = 0;
      sft[d] = NULL;
      for (k = 1; k <= (q - 1); k++)
        {
          m = d + k;
          if (0 <= m && m <= spectrum_tree->max_depth)
            {
              new_size = sft_element_sizes[d] + 2;
              /* printf("resizing sft[%d] to %d, << %d, %d\n", d, new_size, m-1, k-1);*/
              sft[d] = (int*) my_realloc(sft[d], new_size * sizeof(int));
              sft[d][sft_element_sizes[d]] = m - 1;
              sft[d][sft_element_sizes[d] + 1] = k - 1;
              sft_element_sizes[d] = new_size;
            }
        }
    }

  /* @f = Array.new( @maxd ) { Array.new(@q-1) { Hash.new(0) } } */
  f = (struct hashtable***) my_malloc(spectrum_tree->max_depth
      * sizeof(struct hashtable**));
  for (i = 0; i < spectrum_tree->max_depth; i++)
    {
      f[i]
          = (struct hashtable**) my_malloc((q - 1) * sizeof(struct hashtable*));
      for (k = 0; k < q - 1; k++)
        {
          f[i][k] = create_hashtable(spectrum_tree->node_count, path_hash,
              path_equal, free_q_path_void);
        }
    }

  /* @label = Array.new( @maxd+1 ) { Hash.new(0) } # 1st element is a dummy */
  label = (struct hashtable**) my_malloc((spectrum_tree->max_depth + 1)
      * sizeof(struct hashtable*));
  /* 1st element is a dummy */
  for (i = 0; i < spectrum_tree->max_depth + 1; i++)
    {
      label[i] = create_hashtable(spectrum_tree->node_count, q_gram_hash,
          q_gram_equal, free_q_gram_void);
    }

  /*
   if (ruby_test) {
   for (i = 0; i < spectrum_tree->node_count; i++) {
   t = spectrum_tree->nodes[i];
   t->label = (t->label % 9) +1;
   }
   printf("ruby qgram.rb -q %d -t ", q);
   for (i = 0; i < spectrum_tree->node_count; i++) {
   //printf("i = %d from %d\n", i, spectrum_tree->node_count);
   t = spectrum_tree->nodes[i];
   d = t->depth;
   if (i != 0) {
   printf(",");
   }
   printf("%d",d);
   }
   printf(" -l ");
   for (i = 0; i < spectrum_tree->node_count; i++) {
   //printf("i = %d from %d\n", i, spectrum_tree->node_count);
   t = spectrum_tree->nodes[i];
   l = t->label;
   if (i != 0) {
   printf(",");
   }
   printf("%d", l % 10);
   //print_escaped_element(value_for_index(l, string_index));
   }
   printf(" >> ruby.out\n");
   }
   */
  /* main */
  /* @d.each_with_index do |d, i| */
  for (i = 0; i < spectrum_tree->node_count; i++)
    {
      t = spectrum_tree->nodes[i];
      d = t->depth;

      /*l = @l[i]*/
      l = t->label;
      /*printf("i = %d (%d), d = %d, l = %d\n", i, spectrum_tree->node_count, d, l);*/

      /* @cnt[d].each do |j,k| */
      for (ii = 0; ii < cnt_element_sizes[d];)
        {
          j = cnt[d][ii++];
          k = cnt[d][ii++];

          /*printf("  (cnt) j = %d, k = %d\n", j, k);*/

          /* @f[j][k].each_pair do |w, f| */
          if (hashtable_count(f[j][k]) > 0)
            {
              itr = hashtable_iterator(f[j][k]);
              assert(itr != NULL);
              do
                {
                  w = (Q_PATH*) hashtable_iterator_key(itr);
                  f_val = *((unsigned char *) hashtable_iterator_value(itr));
                  /*
                   printf("    (cnt) f = %d, w = ", f_val);
                   print_q_path(w, string_index);
                   printf("\n");
                   */

                  /* @label[d][ [k,w+l] ] += f */
                  q_gram = create_q_gram(k, concat_q_path(w, l, q));
                  val = (unsigned char*) hashtable_search(label[d], q_gram);
                  /*
                   printf("    (cnt) --- label[%d][", d);
                   print_q_gram(q_gram, string_index);
                   */
                  if (val == NULL)
                    {
                      val = (unsigned char*) my_malloc(sizeof(unsigned char));
                      *val = 0;
                      hashtable_insert(label[d], q_gram, val);
                    }
                  else
                    {
                      /* we have not inserted the q_gram - we need to destroy it */
                      free_q_gram(q_gram);
                    }
                  *val += f_val;
                  /*printf("] = %d\n", (*val));*/
                }
              while (hashtable_iterator_advance(itr));
              free(itr);
            }
        }
      /*printf("&&& i = %d (%d), d = %d, l = %d\n", i, spectrum_tree->node_count, d, l);*/

      /*end if d < @maxd*/
      if (d < spectrum_tree->max_depth)
        {
          /*@sft[d].each do |j,k|*/
          for (ii = 0; ii < sft_element_sizes[d];)
            {
              j = sft[d][ii++];
              k = sft[d][ii++];
              /*printf("  (sft) ++ j = %d, k = %d\n", j, k);*/

              /*@f[j][k].each_pair do |w,f|*/
              if (hashtable_count(f[j][k]) > 0)
                {
                  itr = hashtable_iterator(f[j][k]);
                  assert(itr != NULL);
                  /*printf("  (sft) iterating over f[%d][%d]\n", j, k);*/
                  do
                    {
                      w = (Q_PATH*) hashtable_iterator_key(itr);
                      f_val
                          = *((unsigned char *) hashtable_iterator_value(itr));

                      /*
                       printf("    (sft) ** f = %d, w = ", f_val);
                       print_q_path(w, string_index);
                       printf("\n");
                       */

                      /*@f[j][k+1][w+l] += f if k+1 < @q-1*/
                      if (k + 1 < q - 1)
                        {
                          /*printf("concat_q_path 1\n");*/
                          q_path = concat_q_path(w, l, q);

                          /*
                           printf("    (sft) f[%d][%d][", j, k+1);
                           print_q_path(q_path, string_index);
                           */

                          val = (unsigned char*) hashtable_search(f[j][k + 1],
                              q_path);
                          if (val == NULL)
                            {
                              val = (unsigned char*) my_malloc(
                                  sizeof(unsigned char));
                              *val = 0;
                              hashtable_insert(f[j][k + 1], q_path, val);
                            }
                          else
                            {
                              /* we have not inserted the q_path - we need to destroy it */
                              free_q_path(q_path);
                            }
                          *val += f_val;
                          /*
                           printf("] = %d\n", (*val));
                           printf("    (sft) f[%d][%d].size after update = %d\n", j, k+1, hashtable_count(f[j][k+1]));
                           */
                        }

                      /*@f[j][k].delete(w)*/
                    }
                  while (hashtable_iterator_remove(itr) /*remove does also advance */);
                  free(itr);
                }
            }
        }
      /*printf("!!! i = %d (%d), d = %d, l = %d\n", i, spectrum_tree->node_count, d, l);*/

      /*@f[d-1][0][l] += 1 if d > 0;*/
      if (d > 0)
        {

          q_path = concat_q_path(NULL, l, q);
          val = (unsigned char*) hashtable_search(f[d - 1][0], q_path);

          /*
           printf("(mid) f[%d][%d][", d-1, 0);
           print_q_path(q_path, string_index);
           */

          if (val == NULL)
            {
              val = (unsigned char*) my_malloc(sizeof(unsigned char));
              *val = 0;
              hashtable_insert(f[d - 1][0], q_path, val);
            }
          else
            {
              free_q_path(q_path);
            }
          *val += 1;
          /*
           printf("] = %d\n", (*val));
           printf("(mid) f[%d][%d].size after update = %d\n", d-1, 0, hashtable_count(f[d-1][0]));
           */
        }

      /*end if d < @maxd*/
      if (d < spectrum_tree->max_depth)
        {
          /*@label[d+1].each_pair do |ptn, f|*/
          /*printf("(lbl) label[%d].size = %d\n", d+1, hashtable_count(label[d+1]));*/
          if (hashtable_count(label[d + 1]) > 0)
            {
              itr = hashtable_iterator(label[d + 1]);
              assert(itr != NULL);
              do
                {
                  ptn = (Q_GRAM*) hashtable_iterator_key(itr);
                  f_val = *((unsigned char *) hashtable_iterator_value(itr));

                  /*k, w = ptn*/
                  k = ptn->pattern_id;
                  w = ptn->q_path;

                  /*wl = w + l*/
                  wl = concat_q_path(w, l, q);

                  /*
                   printf("(lbl) k = %d, w = ", k);
                   print_q_path(w, string_index);
                   printf(", wl = ");
                   print_q_path(wl, string_index);
                   printf("\n");
                   */

                  /*if wl.size == @q*/
                  if (wl->length == q)
                    {
                      /*@ptn[k][wl] += f*/
                      val = (unsigned char*) hashtable_search(
                          result->q_patterns[k], wl);
                      /*
                       printf("(lbl) ptn[%d][", k);
                       print_q_path(wl, string_index);
                       */
                      if (val == NULL)
                        {
                          val = (unsigned char*) my_malloc(
                              sizeof(unsigned char));
                          *val = 0;
                          hashtable_insert(result->q_patterns[k], wl, val);
                        }
                      else
                        {
                          free_q_path(wl);
                        }
                      *val += f_val;
                      /*printf("] = %d\n", (*val));*/
                    }
                  else
                    {
                      /*@label[d][ [k,wl[0,@q]] ] += f # leftmost @q lenth string*/
                      q_gram
                          = create_q_gram(k, q_subpath(0, q < wl->length ? q
                              : wl->length, wl, q) /* TODO: check if wl[0,@q] is important */);
                      free_q_path(wl);
                      val = (unsigned char*) hashtable_search(label[d], q_gram);
                      /*
                       printf("(lbl) ... label[%d][", d);
                       print_q_gram(q_gram, string_index);
                       */

                      if (val == NULL)
                        {
                          val = (unsigned char*) my_malloc(
                              sizeof(unsigned char));
                          *val = 0;
                          hashtable_insert(label[d], q_gram, val);
                        }
                      else
                        {
                          free_q_gram(q_gram);
                        }
                      *val += f_val;
                      /*printf("] = %d\n", (*val));*/
                    }
                  /*@label[d+1].delete(ptn)*/
                  val = (unsigned char*) hashtable_remove(label[d + 1], ptn); /* TODO: the returned value should probably be freed */
                  free(val);

                }
              while (hashtable_iterator_advance(itr));
              free(itr);
            }
        }
      /*printf("??? i = %d (%d), d = %d, l = %d\n", i, spectrum_tree->node_count, d, l);*/

    }

  /*printf("(fin) label[0].size = %d\n", hashtable_count(label[0]));*/

  /*@label[0].each_pair do |ptn, f|*/
  if (hashtable_count(label[0]) > 0)
    {
      itr = hashtable_iterator(label[0]);
      assert(itr != NULL);
      do
        {
          ptn = (Q_GRAM*) hashtable_iterator_key(itr);
          f_val = *((unsigned char *) hashtable_iterator_value(itr));

          /*k, w = ptn*/
          k = ptn->pattern_id;
          w = ptn->q_path;

          /*
           printf("  (fin) k = %d, w = ", k);
           print_q_path(w, string_index);
           printf("\n");
           */

          /*next unless k == @q-2 # Pattern k is from 0 to q-2 here*/
          if (k == q - 2)
            {

              /*
               printf("    (fin) ptn[%d][", k);
               print_q_path(w, string_index);
               */

              /*@ptn[k][w] += f*/
              val
                  = (unsigned char *) hashtable_search(result->q_patterns[k], w);
              if (val == NULL)
                {
                  val = (unsigned char*) my_malloc(sizeof(unsigned char));
                  *val = 0;
                  hashtable_insert(result->q_patterns[k], q_path_dup(w, q), val);
                }
              *val += f_val;

              /*printf("] = %d\n", (*val));*/
            }
        }
      while (hashtable_iterator_advance(itr));
      free(itr);
    }

  /* clean up */

  /* destroy label */
  for (i = 0; i < spectrum_tree->max_depth + 1; i++)
    {
      hashtable_destroy(label[i], 1 /* destroy values */);
    }
  free(label);

  /* destroy f */
  for (i = 0; i < spectrum_tree->max_depth; i++)
    {
      for (k = 0; k < q - 1; k++)
        {
          hashtable_destroy(f[i][k], 1 /* destroy values */);
        }
      free(f[i]);
    }
  free(f);

  /* destroy sft */
  for (i = 0; i < (spectrum_tree->max_depth); i++)
    {
      if (sft_element_sizes[i] > 0)
        {
          free(sft[i]);
        }
    }
  free(sft);
  free(sft_element_sizes);

  /* destroy cnt */
  for (i = 0; i < (spectrum_tree->max_depth + 1); i++)
    {
      if (cnt_element_sizes[i] > 0)
        {
          free(cnt[i]);
        }
    }
  free(cnt);
  free(cnt_element_sizes);

  /*test_q_profile(result, q, string_index);*/
  return result;
}

void
free_q_profile(Q_PROFILE* q_profile, unsigned char q)
{
  int i;
  if (q_profile != NULL)
    {
      if (q_profile->q_patterns != NULL)
        {
          for (i = 0; i < q_profile->q_patterns_count; i++)
            {
              hashtable_destroy(q_profile->q_patterns[i], 1);
            }
          free(q_profile->q_patterns);
        }
      free(q_profile);
    }
}

void
free_spectrum_tree(SPECTRUM_TREE *spectrum_tree, unsigned char q)
{
  if (spectrum_tree != NULL)
    {
      if (spectrum_tree->nodes != NULL)
        {
          free(spectrum_tree->nodes);
        }
      free_tree(spectrum_tree->root);
      free_q_profile(spectrum_tree->q_profile, q);
      if (spectrum_tree->bag_of_words != NULL)
        {
          hashtable_destroy(spectrum_tree->bag_of_words, 1);
        }
      free(spectrum_tree);
    }
}

unsigned int
q_gram_hash(void* q_gram)
{
  Q_GRAM* qgm;
  unsigned int hash;
  qgm = (Q_GRAM*) q_gram;
  hash = path_hash(qgm->q_path);
  return ((hash << 5) + hash) + qgm->pattern_id;
}

int index_equal(void* k1, void* k2)
{
  return *((string_index_key_t*) k1) == *((string_index_key_t*) k2);
}
unsigned int index_hash(void* index)
{
  return (unsigned int) *((string_index_key_t*) index);
}
unsigned int
path_hash(void* path)
{
  unsigned int hash = 5381;
  Q_PATH* q_path;
  string_index_key_t* label;
  string_index_key_t* label_end;

  q_path = (Q_PATH*) path;
  label = q_path->labels;
  label_end = label + q_path->length;

  while (label < label_end)
    {
      hash = ((hash << 5) + hash) + (*label); /* hash * 33 + c */
      label++;
    }
  return hash;
}

int
q_gram_equal(void* q_gram1, void* q_gram2)
{
  Q_GRAM* qgm1;
  Q_GRAM* qgm2;
  qgm1 = (Q_GRAM*) q_gram1;
  qgm2 = (Q_GRAM*) q_gram2;
  return (qgm1->pattern_id == qgm2->pattern_id) && path_equal(qgm1->q_path,
      qgm2->q_path);
}

int
path_equal(void* path1, void* path2)
{
  Q_PATH* q_path1;
  Q_PATH* q_path2;
  q_path1 = (Q_PATH*) path1;
  q_path2 = (Q_PATH*) path2;

  return (q_path1->length == q_path2->length && 0 == memcmp(q_path1->labels,
      q_path2->labels, q_path1->length * sizeof(string_index_key_t)));
}

