/* tree.h
 *
 * Written by Mark Johnson, 27th May 1997
 *         and Peter Palaga 2009-04-13
 *             create_q_profile() based on Tetsui Kuboyama's Ruby implementation
 *             see also
 *             Kuboyama, T., Kashima, H., Aoki-Kinoshita, K.F., Hirata, K., Yasuda, H. (2007):
 *             A Spectrum Tree Kernel,
 *             Journal of Japanese Society of Artificial Intelligence, Vol.22, No.2, 2007.
 */

#ifndef TREE_H
#define TREE_H
#define NDEBUG

#define tree_depth_max UCHAR_MAX
#define tree_depth_min 1
#define node_count_max USHRT_MAX
#define UNINITIALIZED_KERNEL_VALUE (-1)
#define UNINITIALIZED_KERNEL_VALUE_DBL (-1.0)


#include "string_index.h"
#include "hashtable/hashtable.h"

typedef unsigned char tree_depth_t;

typedef struct tree {
  string_index_key_t label;
  struct tree  *subtrees;
  struct tree  *sibling;
  tree_depth_t depth; /* depth of this node = the number of edges in the path from root; the depth of root = 0 */
} TREE;

typedef struct q_path {
  unsigned char length;
  string_index_key_t* labels;
} Q_PATH;

typedef struct q_gram {
  unsigned char pattern_id;
  Q_PATH* q_path;
} Q_GRAM;

typedef struct q_profile {
  unsigned int q_patterns_count;
  struct hashtable** q_patterns; /* array of hashtables, its size is q_patterns_count
                            with q_path* keys and unsigned char* values
                            the value pointers will not be shared with other hashtables
                            the value pointers should be freed on every remove
                            and on hashtable_destroy
                            */
} Q_PROFILE;

typedef unsigned short node_count_t;

typedef struct spectrum_tree {
  TREE* root;
  node_count_t node_count;
  tree_depth_t max_depth;
  TREE** nodes; /* array of nodes of this tree in postorder */
  Q_PROFILE* q_profile;
  double norm_square; /* = kernel(this, this) or UNINITIALIZED_KERNEL_VALUE */
  int norm_square_spectrum; /* = kernel(this, this) or UNINITIALIZED_KERNEL_VALUE */
  struct hashtable* bag_of_words;
} SPECTRUM_TREE;


int count_common_q_grams(SPECTRUM_TREE* spectrum_tree1, SPECTRUM_TREE* spectrum_tree2, unsigned char q, unsigned char spectrum_tree_kernel_count_mode);

void consumespaces(char* src, int* pos);
string_index_key_t parselabel(char* src, int* pos, unsigned char is_terminal, STRING_INDEX *string_index);
TREE *parsetree(char *src, int *pos, node_count_t *node_count, tree_depth_t depth, tree_depth_t* max_depth, STRING_INDEX *string_index);
void free_tree(TREE* t);
void print_tree(TREE* t, STRING_INDEX* string_index);
void add_tree_nodes(TREE* t, TREE** nodes, node_count_t* node_index);

SPECTRUM_TREE * create_spectrum_tree(char *source, STRING_INDEX* string_index, unsigned char q_gram_length, char kernel_combine_operation, char bow_only_terminals, char custom_kernel_input_format);
void free_spectrum_tree(SPECTRUM_TREE *spectrum_tree, unsigned char q);

Q_PROFILE * read_q_profile(char* src, unsigned char q_gram_length, unsigned int q_patterns_count);

Q_PROFILE * create_q_profile(SPECTRUM_TREE* spectrum_tree, unsigned char q, STRING_INDEX* string_index, unsigned int q_patterns_count);
void free_q_profile(Q_PROFILE* q_profile, unsigned char q);
int norm_square_spectrum(SPECTRUM_TREE* spectrum_tree, unsigned char q, unsigned char spectrum_tree_kernel_count_mode);



Q_PATH *concat_q_path(Q_PATH* q_path, string_index_key_t next_element, unsigned char q_gram_length);
void print_q_path(Q_PATH* q_path, STRING_INDEX* string_index);
Q_PATH *q_path_dup(Q_PATH* q_path, unsigned char q_gram_length);

Q_PATH *create_q_path(unsigned char q_path_capacity);
void free_q_path(Q_PATH* q_path);
void free_q_path_void(void* q_path);
Q_PATH * q_subpath(unsigned char start, unsigned char end, Q_PATH* q_path, unsigned char q_gram_length);
void print_q_gram(Q_GRAM* q_gram, STRING_INDEX* string_index);


Q_GRAM *create_q_gram(unsigned char pattern_id, Q_PATH* q_path);
void free_q_gram(Q_GRAM* q_gram);
void free_q_gram_void(void* q_gram);

unsigned int q_gram_hash(void* q_gram);
unsigned int path_hash(void* path);
int q_gram_equal(void* q_gram1, void* q_gram2);
int path_equal(void* path1, void* path2);


struct hashtable* create_bag_of_words(TREE** nodes, node_count_t node_count, char bow_only_terminals, STRING_INDEX* string_index);
int index_equal(void* k1, void* k2);
unsigned int index_hash(void* index);
int count_common_words(SPECTRUM_TREE* spectrum_tree1, SPECTRUM_TREE* spectrum_tree2, char bow_only_terminals);
int norm_square_bow(SPECTRUM_TREE* spectrum_tree);

unsigned char compute_first(char operation);
unsigned char compute_second(char operation);


#endif




