/************************************************************************/
/*                                                                      */
/*   kernel.h                                                           */
/*                                                                      */
/*   User defined kernel function. Feel free to plug in your own.       */
/*                                                                      */
/*   Copyright: Thorsten Joachims                                       */
/*   Date: 16.12.97                                                     */
/*                                                                      */
/************************************************************************/

#include "tree.h"


unsigned char initialized = 0;

CUSTOM_PARM* custom_params = NULL;

STRING_INDEX* string_index;


void init_kernel(KERNEL_PARM *kernel_parm);
double custom_kernel(KERNEL_PARM *kernel_parm, SVECTOR *a, SVECTOR *b);
CUSTOM_PARM* parse_custom_parm(char* custom_parm);
double count_common_substructures(SPECTRUM_TREE* spectrum_tree1, SPECTRUM_TREE* spectrum_tree2);
double norm_square(SPECTRUM_TREE* spectrum_tree);
void init_svector(SVECTOR *a);
void print_custom_parm_script();


/* KERNEL_PARM is defined in svm_common.h The field 'custom' is reserved for */
/* parameters of the user defined kernel. You can also access and use */
/* the parameters of the other kernels. Just replace the line
             return((double)(1.0));
   with your own kernel. */

  /* Example: The following computes the polynomial kernel. sprod_ss
              computes the inner product between two sparse vectors.

      return((CFLOAT)pow(kernel_parm->coef_lin*sprod_ss(a->words,b->words)
             +kernel_parm->coef_const,(double)kernel_parm->poly_degree));
  */

/* If you are implementing a kernel that is not based on a
   feature/value representation, you might want to make use of the
   field "userdefined" in SVECTOR. By default, this field will contain
   whatever string you put behind a # sign in the example file. So, if
   a line in your training file looks like

   -1 1:3 5:6 #abcdefg

   then the SVECTOR field "words" will contain the vector 1:3 5:6, and
   "userdefined" will contain the string "abcdefg". */

double custom_kernel(KERNEL_PARM *kernel_parm, SVECTOR *a, SVECTOR *b)
     /* plug in you favorite kernel */
{
  int cnt;
  double result;
  /*extern unsigned char custom_kernel_normalize;*/


  if (!initialized) {
    init_kernel(kernel_parm);
  }

  init_svector(a);
  init_svector(b);

  /*return (double)1.0;*/

  /*printf("---\n%s\n-\n%s\n", a->userdefined, b->userdefined);*/

  if (a->parse == NULL || b->parse == NULL) {
    result = 0.0;
  }
  else if (a == b) {
    if (custom_params->custom_kernel_normalize_whole) {
      result = 1.0;
    }
    else {
      result = norm_square(a->parse);
    }
  }
  else {
    /* not trivial - count */
    cnt = count_common_substructures(a->parse, b->parse);
    /*printf("common substr = %d\n", cnt);*/
    if (custom_params->custom_kernel_normalize_whole && cnt != 0){
/*
      printf("cnt = %d, ns(a) = %d, ns(b) = %d, sqrt(ns(a)*ns(b)) = %f\n",
          cnt,
          norm_square(a->parse, q_gram_length),
          norm_square(b->parse, q_gram_length),
          sqrt(norm_square(a->parse, q_gram_length) * norm_square(b->parse, q_gram_length))
      );
*/
      result = (
          (double)cnt)
          /
          ((double) sqrt(
              norm_square(a->parse)
              * norm_square(b->parse)
          )
      );
    }
    else{
      result = (double) cnt;
    }
  }
  /*printf("kernel = %f\n", result);*/
  return result;
}

void init_kernel(KERNEL_PARM *kernel_parm) {

  if (strlen(kernel_parm->custom) == 0) {
    perror("-u option must be set for spectrum kernel.");
    exit(1);
  }
  custom_params = parse_custom_parm(kernel_parm->custom);
  printf("update esk set kernel_script = '");
  print_custom_parm_script();
  printf("' where eskid = currval('esk_eskid_seq');\n");

  switch (custom_params->custom_kernel_input_format) {
    case 't':
      string_index = create_string_index(1024U, 512U);
      break;
    case 'q':
    case 'c':
      string_index = NULL;
      break;
    default:
      perror("Unexpected custom_kernel_input_format value.");
      exit(-1);
  }

  initialized = 1;
}

void init_svector(SVECTOR *a) {
  if (a->parse == NULL && a->userdefined != NULL && a->userdefined[0] != '\0') {
    a->parse = create_spectrum_tree(a->userdefined, string_index, custom_params);
  }
}

void print_custom_parm_script() {
  if (custom_params->custom_kernel_normalize_whole) {
    printf("N(");
  }

  if (compute_first(custom_params->kernels_combine_operation)) {
    if (custom_params->custom_kernel_normalize_parts) {
      printf("n(");
    }

    switch (custom_params->custom_kernel_input_format) {
      case 't':
        break;
      case 'q':
        printf("dep-");
      case 'c':
        printf("ctxt-");
        break;
      default:
        perror("Unexpected custom_kernel_input_format value.");
        exit(-1);
    }

    if (custom_params->spectrum_tree_kernel_count_mode == 'u') {
      printf("u-");
    }

    if (custom_params->custom_kernel_input_format == 'c') {
      printf("spectrum(%u,%c,%c,%c)",
          custom_params->q_gram_length,
          custom_params->q_gram_length_mode,
          custom_params->spectrum_tree_kernel_match_mode,
          custom_params->entity_context_input_format_entity_position
      );
    }
    else {
      printf("spectrum(%u)", custom_params->q_gram_length);

    }

    if (custom_params->custom_kernel_normalize_parts) {
      printf(")");
    }

  }

  if (compute_first(custom_params->kernels_combine_operation)
      && compute_second(custom_params->kernels_combine_operation)
      ) {
    if (custom_params->kernels_combine_operation == 'a') {
      printf(" + (");
    }
    else if (custom_params->kernels_combine_operation == 'm') {
      printf(" * (");
    }
    else if (custom_params->kernels_combine_operation == '1') {
    }
    else if (custom_params->kernels_combine_operation == '2') {
    }
    else {
      perror("Unexpected kernel combine operation.");
      exit(-1);
    }

  }

  if (compute_second(custom_params->kernels_combine_operation)) {

    if (compute_first(custom_params->second_kernel_factor_operation)) {
      printf("%f", custom_params->second_kernel_factor);
      if (custom_params->second_kernel_factor_operation == 'a') {
        printf(" + ");
      }
      else if (custom_params->second_kernel_factor_operation == 'm') {
        printf(" * ");
      }
      else {
        perror("Unexpected second kernel operation.");
        exit(-1);
      }
    }


    if (custom_params->custom_kernel_normalize_parts) {
      printf("n(");
    }

    printf("bow(%c)", custom_params->bow_only_terminals);

    if (custom_params->custom_kernel_normalize_parts) {
      printf(")");
    }

    /* end of op */
    printf(")");

  }


  if (custom_params->custom_kernel_normalize_whole) {
    printf(")");
  }

}

CUSTOM_PARM* parse_custom_parm(char* custom_parm)
{
  char* custom_parm_cpy;
  char* key;
  char* val;
  CUSTOM_PARM* result;

  result = (CUSTOM_PARM*) my_malloc(sizeof(CUSTOM_PARM));
  result->q_gram_length = 0;
  result->q_gram_length_mode = 'f';
  /* 'f' ... fixed q_gram_length
   * 'u' ... all up to q_gram_length
  */

  result->custom_kernel_normalize_whole = 0;
  result->custom_kernel_normalize_parts = 0;
  result->second_kernel_factor = 1.0;
  /* '1' ... the first only
   * '2' ... the second only
   * 'a' ... addition
   * 'm' ... multiplication */
  result->second_kernel_factor_operation = 'm';
  result->kernels_combine_operation = 'a';
  /* 'a' ... all
   * 't' ... only terminals
  */
  result->bow_only_terminals = 'a';
  result->custom_kernel_input_format = 't';
  /* 'q' ... q-grams
   * 't' ... PTB bracketing trees
   * 'c' ... entity context
   */
  result->spectrum_tree_kernel_count_mode = 'm';
  /* 'm' ... multiply the real counts
   * 'u' ... increment by 1 for each unique match
   */
  result->spectrum_tree_kernel_match_mode = 'e';
  /* 'e' ... exact
   * 't' ... tolerant - dependecy types must match, nodel labels need not
   */
  result->entity_context_input_format_entity_position = '0';
  /* '0' ... beginning
   * 'a' ... anywhere
   */




  custom_parm_cpy = (char*) my_malloc((strlen(custom_parm) + 1) * sizeof(char));
  strcpy(custom_parm_cpy, custom_parm);

  key = strtok(custom_parm_cpy, "=,");

  while (key != NULL)
    {
      val = strtok(NULL,"=,");
      assert(val != NULL);
      if (strcmp(key, "q") == 0)
        {
          result->q_gram_length = atoi(val);
        }
      else if (strcmp(key, "N") == 0)
        {
          result->custom_kernel_normalize_whole = atoi(val);
        }
      else if (strcmp(key, "n") == 0)
        {
          result->custom_kernel_normalize_parts = atoi(val);
        }
      else if (strcmp(key, "c") == 0)
        {
          result->second_kernel_factor = atof(val);
        }
      else if (strcmp(key, "O") == 0)
        {
          result->kernels_combine_operation = val[0];
        }
      else if (strcmp(key, "o") == 0)
        {
          result->second_kernel_factor_operation = val[0];
        }
      else if (strcmp(key, "b") == 0)
        {
          result->bow_only_terminals = val[0];
        }
      else if (strcmp(key, "f") == 0)
        {
          result->custom_kernel_input_format = val[0];
        }
      else if (strcmp(key, "m") == 0)
        {
          result->spectrum_tree_kernel_count_mode = val[0];
        }
      else if (strcmp(key, "t") == 0)
        {
          result->spectrum_tree_kernel_match_mode = val[0];
        }
      else if (strcmp(key, "e") == 0)
        {
          result->entity_context_input_format_entity_position = val[0];
        }
      else if (strcmp(key, "l") == 0)
        {
          result->q_gram_length_mode = val[0];
        }
      else
        {
          perror("Unexpected custom kernel option.");
          exit(-1);
        }

      key = strtok(NULL,"=,");
    }

  free(custom_parm_cpy);

  return result;
}


double combine(char operation, double operand1, double operand2) {

  if (operation == 'a') {
    return operand1 + operand2;
  }
  else if (operation == '1') {
    return operand1;
  }
  else if (operation == '2') {
    return operand2;
  }
  else if (operation == 'm') {
    return operand1 * operand2;
  }
  else {
    perror("Unexpected second kernel operation.");
    exit(-1);
  }
}

double count_common_substructures(SPECTRUM_TREE* spectrum_tree1, SPECTRUM_TREE* spectrum_tree2) {
  double k1 = 0.0;
  double k2 = 0.0;
  int cnt = 0;

  if (compute_first(custom_params->kernels_combine_operation)) {
    if (custom_params->custom_kernel_input_format == 'c') {
      cnt = score_context_paths(spectrum_tree1, spectrum_tree2, custom_params);
    }
    else {
      cnt = count_common_q_grams(spectrum_tree1, spectrum_tree2, custom_params);
    }
    /*printf("common qgrams = %d\n", cnt);*/
    if (custom_params->custom_kernel_normalize_parts && cnt != 0) {

      k1 = (
          (double)cnt)
          /
          ((double) sqrt(
              norm_square_spectrum(spectrum_tree1, custom_params)
              * norm_square_spectrum(spectrum_tree2, custom_params)
          )
      );
    }
    else {
      k1 = (double) cnt;
    }
  }

  if (compute_second(custom_params->kernels_combine_operation)) {
    cnt = count_common_words(spectrum_tree1, spectrum_tree2, custom_params->bow_only_terminals);
    /*printf("common words = %d\n", cnt);*/
    if (custom_params->custom_kernel_normalize_parts && cnt != 0) {
      k2 = (
          (double)cnt)
          /
          ((double) sqrt(
              norm_square_bow(spectrum_tree1)
              * norm_square_bow(spectrum_tree2)
          )
      );
    }
    else {
      k2 = (double) cnt;
    }

    if (compute_first(custom_params->second_kernel_factor_operation)) {
      k2 = combine(custom_params->second_kernel_factor_operation, custom_params->second_kernel_factor, k2);
    }
  }
  return combine(custom_params->kernels_combine_operation, k1, k2);
}

double norm_square(SPECTRUM_TREE* spectrum_tree)
{
  assert(spectrum_tree != NULL);
  if (spectrum_tree->norm_square == UNINITIALIZED_KERNEL_VALUE_DBL)
    {
      spectrum_tree->norm_square = count_common_substructures(spectrum_tree, spectrum_tree);
    }
  return spectrum_tree->norm_square;
}

