/* tree.c
 *
 *  Written by Mark Johnson, 27th May 1997
 */

#include <stdlib.h>
#include <assert.h>
#include <ctype.h>
#include "tree.h"
#include "svm_common.h"
#include "hashtable/hashtable_itr.h"

void
consumespaces(char* src, int* pos)
{
  int c;
  /* skip spaces */
  while (1)
    {
      c = src[*pos];
      if (!isspace(c))
        {
          break;
        }
      *pos = (*pos) + 1;
    }
}

string_index_key_t
parselabel(char* src, int* pos, unsigned char is_terminal,
    STRING_INDEX *string_index)
{
  int c, start_pos, len;

  start_pos = *pos;

  while (1)
    {
      c = src[*pos];
      if (c == '\0' || isspace(c) || c == '(' || c == ')')
        {
          break;
        }
      *pos = (*pos) + 1;
    };

  len = (*pos) - start_pos;
  assert(len> 0);

  consumespaces(src, pos);

  return index_for_value(src + start_pos, len, is_terminal, string_index);
}

TREE *
parsetree(char *src, int *pos, node_count_t *node_count, tree_depth_t depth,
    tree_depth_t* max_depth, STRING_INDEX *string_index)
{
  int c;
  TREE* t;
  TREE* p;

  c = src[*pos];
  assert(c != '\0'); /* file ended w/out closing ')' */

  switch (c)
    {

  case ')':
    *pos = (*pos) + 1;
    consumespaces(src, pos);
    return (NULL);
    break; /* empty tree */

  case '(': /* nonterminal */
    *pos = (*pos) + 1;
    consumespaces(src, pos);
    t = (TREE*) my_malloc(sizeof(TREE));
    t->subtrees = NULL;
    t->sibling = NULL;
    if ((*node_count) + 1 > node_count_max)
      {
        perror("node_count overflow.");
      }
    *node_count = (*node_count) + 1;
    t->label = parselabel(src, pos, 0, string_index);
    t->depth = depth;
    if ((*max_depth) < depth)
      {
        *max_depth = depth;
      }

    t->subtrees = p = parsetree(src, pos, node_count, depth + 1, max_depth,
        string_index);

    while (p)
      {
        p->sibling = parsetree(src, pos, node_count, depth + 1, max_depth,
            string_index);
        p = p->sibling;
      }
    return (t);
    break;

  default:
    /* terminal */
    t = (TREE*) my_malloc(sizeof(TREE));
    t->subtrees = NULL;
    t->sibling = NULL;
    if ((*node_count) + 1 > node_count_max)
      {
        perror("node_count overflow.");
      }
    *node_count = (*node_count) + 1;
    t->label = parselabel(src, pos, 1, string_index);
    t->depth = depth;
    if ((*max_depth) < depth)
      {
        *max_depth = depth;
      }

    consumespaces(src, pos);

    /* check the closing bracket */
    c = src[*pos];
    *pos = (*pos) + 1;
    assert(c==')');
    *pos = (*pos) - 1;
    return (t);
    }
}

void
free_tree(TREE* t)
{
  if (t != NULL)
    {
      free_tree(t->sibling);
      free_tree(t->subtrees);
      free(t);
    }
}

void
add_tree_nodes(TREE* t, TREE** nodes, node_count_t* node_index)
{
  if (t->subtrees != NULL)
    {
      TREE* subtree = t->subtrees;

      while (subtree != NULL)
        {
          add_tree_nodes(subtree, nodes, node_index);
          subtree = subtree->sibling;
        }
    }
  nodes[*node_index] = t;
  *node_index = (*node_index) + 1;
  /*printf("array_element = %u\n", *node_index);*/
}
void
print_tree(TREE* t, STRING_INDEX* string_index)
{
  char *label;
  label = value_for_index(t->label, string_index);

  if (t->subtrees)
    {
      TREE* p;

      printf("(%s{%d} ", label, t->depth);
      print_tree(p = t->subtrees, string_index);

      while ((p = p->sibling))
        {
          printf(" ");
          print_tree(p, string_index);
        }
      printf(")");
    }
  else
    {
      printf("%s{%d}", label, t->depth);
    }
}

Q_PATH *
concat_q_path(Q_PATH* q_path, string_index_key_t next_element,
    unsigned char q_gram_length)
{
  int i;
  Q_PATH *result;
  if (q_path == NULL)
    {
      result = create_q_path(q_gram_length);
      result->labels[0] = next_element;
      result->length = 1;
      return result;
    }
  else
    {
      if (q_path->length >= q_gram_length)
        {
          result = create_q_path(q_gram_length + 1);
        }
      else
        {
          result = create_q_path(q_path->length);
        }
      for (i = 0; i < q_path->length; ++i)
        {
          result->labels[i] = q_path->labels[i];
        }
      result->labels[i++] = next_element;
      result->length = i;
      return result;
    }
}

Q_PATH *
q_subpath(unsigned char start, unsigned char end, Q_PATH* q_path,
    unsigned char q_gram_length)
{
  Q_PATH* result;
  int i, n;
  assert(q_path != NULL);
  assert(start < end);
  /*assert(end <= q_path->length || end <= q_gram_length TODO: check if this is not too liberal );*/
  n = end - start;

  result = create_q_path(n > q_gram_length ? n : q_gram_length);
  result->length = n;
  for (i = 0; i < n; ++i)
    {
      result->labels[i] = q_path->labels[start + i];
    }
  return result;
}
void
print_q_gram(Q_GRAM* q_gram, STRING_INDEX* string_index)
{
  printf("%u|", q_gram->pattern_id);
  print_q_path(q_gram->q_path);
}
void
print_q_path(Q_PATH* q_path)
{
  int i;
  if (q_path != NULL)
    {
      printf("(");
      for (i = 0; i < q_path->length; ++i)
        {
          if (i > 0) {
            printf(" ");
          }
          printf("%u", q_path->labels[i]);
        }
      printf(")");
    }
}
Q_PATH *
q_path_dup(Q_PATH* q_path, unsigned char q_gram_length)
{
  Q_PATH* result;
  int i;
  assert(q_path != NULL);
  result = create_q_path(q_path->length > q_gram_length ? q_path->length
      : q_gram_length);
  result->length = q_path->length;
  for (i = 0; i < q_path->length; ++i)
    {
      result->labels[i] = q_path->labels[i];
    }
  return result;
}

Q_PATH *
create_q_path(unsigned char q_path_capacity)
{
  Q_PATH *result;
  int i;
  result = (Q_PATH*) my_malloc(sizeof(Q_PATH));
  result->length = 0;
  result->labels = (string_index_key_t*) my_malloc(q_path_capacity
      * sizeof(string_index_key_t));
  for (i = 0; i < q_path_capacity; ++i)
    {
      result->labels[i] = ZERO_INDEX;
    }
  return result;
}

void
free_q_path(Q_PATH* q_path)
{
  if (q_path != NULL)
    {
      if (q_path->labels != NULL)
        {
          free(q_path->labels);
        }
      free(q_path);
    }
}

void
free_q_path_void(void* q_path)
{
  free_q_path((Q_PATH*) q_path);
}

Q_GRAM *
create_q_gram(unsigned char pattern_id, Q_PATH* q_path)
{
  Q_GRAM* result;
  result = (Q_GRAM*) my_malloc(sizeof(Q_GRAM));
  result->pattern_id = pattern_id;
  result->q_path = q_path;
  return result;
}

void
free_q_gram_void(void* q_gram)
{
  free_q_gram((Q_GRAM*) q_gram);
}

void
free_q_gram(Q_GRAM* q_gram)
{
  if (q_gram != NULL)
    {
      if (q_gram->q_path != NULL)
        {
          free_q_path(q_gram->q_path);
        }
      free(q_gram);
    }
}

unsigned char
compute_first(char operation)
{
  switch (operation)
    {
  case 'a':
  case '1':
  case 'm':
    return 1;
    break;
  case '2':
    return 0;
    break;
  default:
    perror("Unexpected second kernel operation.");
    exit(-1);
    }
}

unsigned char
compute_second(char operation)
{
  switch (operation)
    {
  case 'a':
  case '2':
  case 'm':
    return 1;
    break;
  case '1':
    return 0;
    break;
  default:
    perror("Unexpected second kernel operation.");
    exit(-1);
    }
}

SPECTRUM_TREE*
create_spectrum_tree(char *source, STRING_INDEX* string_index,
    CUSTOM_PARM* custom_params)
{
  SPECTRUM_TREE *result;
  node_count_t node_index = 0;
  int pos = 0;
  result = NULL;

  if (source && source[0] != '\0')
    {
      /*printf("%s\n", source);*/
      result = (SPECTRUM_TREE *) my_malloc(sizeof(SPECTRUM_TREE));
      result->node_count = 0;
      result->max_depth = 0;
      result->example_id = NULL;

      switch (custom_params->custom_kernel_input_format)
        {
      case 't':
        result->root = parsetree(source, &pos, &(result->node_count), 0,
            &(result->max_depth), string_index);
        result->nodes
            = (TREE**) my_malloc((result->node_count) * sizeof(TREE*));
        add_tree_nodes(result->root, result->nodes, &node_index);
        /* check if we have read as many nodes as expected */
        assert(node_index == result->node_count);
        break;
      case 'q':
      case 'c':
        result->root = NULL;
        result->nodes = NULL;
        break;
      default:
        perror("Unexpected custom_kernel_input_format value.");
        exit(-1);
        }

      result->norm_square = UNINITIALIZED_KERNEL_VALUE_DBL;
      result->norm_square_spectrum = UNINITIALIZED_KERNEL_VALUE;

      if (compute_first(custom_params->kernels_combine_operation))
        {
          switch (custom_params->custom_kernel_input_format)
            {
          case 't':
            result->q_profile = create_q_profile(result,
                custom_params->q_gram_length, string_index,
                custom_params->q_gram_length - 1);
            result->context_example = NULL;
            break;
          case 'q':
            /* each odd index in the q-gram is the dependency
             * so there are (q_gram_length - 1)/2 dependencies
             * each dependency may have up_bottom or bottom_up direction = 2 possibilities
             * so we need exactly (q_gram_length - 1)/2 bits to store all possible path patterns
             * and so many bits means  pow(2, so_many) possibilities */
            result->q_profile = read_q_profile(source,
                custom_params->q_gram_length, pow(2,
                    (custom_params->q_gram_length - 1) / 2));
            result->context_example = NULL;
            break;
          case 'c':
            result->q_profile = NULL;
            result->context_example = read_context_example(source, &pos,
                custom_params);
            result->example_id = read_example_id(source, &pos, custom_params);
            break;
          default:
            perror("Unexpected custom_kernel_input_format value.");
            exit(-1);
            }
        }
      else
        {
          result->q_profile = NULL;
        }

      if (compute_second(custom_params->kernels_combine_operation))
        {
          result->bag_of_words = create_bag_of_words(result->nodes,
              result->node_count, custom_params->bow_only_terminals,
              string_index);
        }
      else
        {
          result->bag_of_words = NULL;
        }

      /*
       print_tree(result->root, string_index);
       printf("\n----- mode_count = %u; max_depth = %u \n", result->node_count, result->max_depth);
       */

    }

  return result;
}

struct hashtable*
create_bag_of_words(TREE** nodes, node_count_t node_count,
    char bow_only_terminals, STRING_INDEX* string_index)
{
  struct hashtable* result;
  TREE* tree;
  int i;
  unsigned char* val;
  result = create_hashtable(node_count, index_hash, index_equal, free);

  for (i = 0; i < node_count; ++i)
    {
      tree = nodes[i];
      if (bow_only_terminals == 'a' || (bow_only_terminals == 't'
          && tree->subtrees == NULL))
        {
          /* a terminal symbol */
          if (hashtable_search(result, &(tree->label)) == NULL)
            {
              /* not there yet */
              string_index_key_t* key = (string_index_key_t*) my_malloc(
                  sizeof(string_index_key_t));
              *key = tree->label;
              val = (unsigned char*) my_malloc(sizeof(unsigned char));
              *val = 1;
              hashtable_insert(result, key, val);
            }
        }
    }

  return result;
}

void
print_escaped_element(char* str)
{
  int i;
  i = 0;
  while (str[i] != '\0')
    {
      if (str[i] == ',')
        {
          putc('_', stdout);
        }
      else
        {
          putc(str[i], stdout);
        }
      i++;
    }
}
void
test_q_profile(Q_PROFILE* q_profile, unsigned char q,
    STRING_INDEX* string_index)
{
  int i, j, ii;
  struct hashtable_itr *itr;
  Q_PATH* q_path;
  unsigned char val;

  assert(q_profile != NULL);
  /*printf("[");*/
  for (i = 0; i < q - 1; ++i)
    {
      if (i != 0)
        {
          /*printf(" ,");*/
        }
      /*printf("{");*/

      if (hashtable_count(q_profile->q_patterns[i]) != 0)
        {
          itr = hashtable_iterator(q_profile->q_patterns[i]);
          j = 0;
          do
            {
              q_path = (Q_PATH*) hashtable_iterator_key(itr);
              val = *((unsigned char*) hashtable_iterator_value(itr));

              printf("insert into kv1 (id, k, cnt) values (%d, '", i);
              /*if (j++ != 0) {
               printf(" ,");
               printf("\"");
               }*/
              if (q_path != NULL)
                {
                  for (ii = 0; ii < q_path->length; ++ii)
                    {
                      printf("%d", q_path->labels[ii]);
                    }
                }
              /*
               printf("\"");
               printf("=>%u", val);
               */
              printf("', %u);\n", val);
            }
          while (hashtable_iterator_advance(itr));
          free(itr);
        }
      /*printf("}");*/
    }
  /*printf("]\n");*/
  exit(-1);
}

int
norm_square_spectrum(SPECTRUM_TREE* spectrum_tree, CUSTOM_PARM* custom_params)
{
  assert(spectrum_tree != NULL);
  if (spectrum_tree->norm_square_spectrum == UNINITIALIZED_KERNEL_VALUE)
    {
      if (custom_params->custom_kernel_input_format == 'c') {
        spectrum_tree->norm_square_spectrum = score_context_paths(spectrum_tree, spectrum_tree, custom_params);
      }
      else {
        spectrum_tree->norm_square_spectrum = count_common_q_grams(spectrum_tree,spectrum_tree, custom_params);
      }
    }
  return spectrum_tree->norm_square_spectrum;
}

int
norm_square_bow(SPECTRUM_TREE* spectrum_tree)
{
  return hashtable_count(spectrum_tree->bag_of_words);
}

int
count_common_words(SPECTRUM_TREE* spectrum_tree1,
    SPECTRUM_TREE* spectrum_tree2, char bow_only_terminals)
{
  int result = 0;
  struct hashtable_itr *itr;
  struct hashtable* ht2;
  unsigned char* val1;
  Q_PATH* key1;
  unsigned char* val2;

  if (spectrum_tree1 == spectrum_tree2)
    {
      return hashtable_count(spectrum_tree1->bag_of_words);
    }

  /* take the one with less nodes */
  if (spectrum_tree1->node_count <= spectrum_tree2->node_count)
    {
      if (hashtable_count(spectrum_tree1->bag_of_words) > 0)
        {
          itr = hashtable_iterator(spectrum_tree1->bag_of_words);
          assert(itr != NULL);
          ht2 = spectrum_tree2->bag_of_words;
          assert(ht2 != NULL);
          do
            {
              val1 = hashtable_iterator_value(itr);
              if (val1 != NULL)
                {
                  key1 = hashtable_iterator_key(itr);
                  assert(key1 != NULL);
                  val2 = hashtable_search(ht2, key1);
                  if (val2 != NULL)
                    {
                      /*printf("result += %d * %d\n", (*val1), (*val2));*/
                      result += (*val1) * (*val2);
                      /*printf("result = %d\n", result);*/
                    }
                }
            }
          while (hashtable_iterator_advance(itr));
          free(itr);
        }
      return result;
    }

  else
    {
      /* swap */
      return count_common_words(spectrum_tree2, spectrum_tree1,
          bow_only_terminals);
    }

}

int score_pathes(Q_PATH* path1, Q_PATH* path2, CUSTOM_PARM* custom_params)
{
  int result = 0;
  int i = 0;
  int eq = 0;

/*
  printf("score_pathes ");
  print_q_path(path1);
  printf(" ");
  print_q_path(path2);
*/

  switch (custom_params->spectrum_tree_kernel_match_mode)
    {
  case 'e':
    /* exact */
    assert(path1->length == path2->length);

    if (0 == memcmp(path1->labels, path2->labels, path1->length
        * sizeof(string_index_key_t)))
      {
        result = path1->length;
      }
    else
      {
        result = 0;
      }
    /*printf(" = %d\n", result);*/

    return result;
  case 't':
    /* tolerant */
    assert(path1->length == path2->length);

    for (i = 1; i < path1->length; ++i) {
      eq = (path1->labels[i] == path2->labels[i]);
      if (i % 2 == 0) {
        /* node label */
        if (eq) {
          result += 2;
        }
        else {
          /* tolerate */
        }
      }
      else {
        /* dependency type */
        if (eq) {
          result += 1;
        }
        else {
          /* do not tolerate */
          return 0;
        }
      }
    }

    return result;
  case 'l':
    /* levenshtein */
    assert(path1->length == path2->length);
    return levenshtein_similarity(path1->labels, path2->labels, path1->length);
  case 'd':
    /* levenshtein */
    assert(path1->length == path2->length);
    return levenshtein_distance(path1->labels, path1->length, path2->labels, path2->length);
  case 'i':
    /* intelligent levenshtein */
    assert(path1->length == path2->length);
    return i_levenshtein_similarity(path1->labels, path2->labels, path1->length, custom_params);
  default:
    perror("Unexpected spectrum_tree_kernel_match_mode.");
    exit(-1);
    }

}

/* Computes the score for two entity occurrences. */
int score_occurence_pair(CONTEXT_ENTITY_OCCURENCE* occ1,
    CONTEXT_ENTITY_OCCURENCE* occ2, CUSTOM_PARM* custom_params)
{
  Q_PATH* path1;
  Q_PATH* path2;
  int pi1, pi2;
  int pci;
  int pc1, pc2;
  int path_counts;
  int result = 0;

  path_counts = path_count(custom_params);
  for (pci = 0; pci < path_counts; ++pci)
    {
      pc1 = occ1->path_counts[pci];
      pc2 = occ2->path_counts[pci];

      for (pi1 = 0; pi1 < pc1; ++pi1)
        {
          path1 = occ1->pathes[pci][pi1];

          for (pi2 = 0; pi2 < pc2; ++pi2)
            {
              path2 = occ2->pathes[pci][pi2];

              result += score_pathes(path1, path2, custom_params);
            }
        }

    }

  return result;

}

/* occs1 and occs2 are arrays of entity occurrences
 * returns the best pair score. The pair's first
 * element comes from occs1 and the second pair's
 * element comes from occs2
 * */
int score_occurences(CONTEXT_ENTITY_OCCURENCE** occs1, int occ_cnt1,
    CONTEXT_ENTITY_OCCURENCE** occs2, int occ_cnt2, CUSTOM_PARM* custom_params)
{
  CONTEXT_ENTITY_OCCURENCE* occ1;
  CONTEXT_ENTITY_OCCURENCE* occ2;
  int oi1, oi2;

  int result = 0;
  int score = 0;

  for (oi1 = 0; oi1 < occ_cnt1; ++oi1)
    {
      occ1 = occs1[oi1];
      for (oi2 = 0; oi2 < occ_cnt2; ++oi2)
        {
          occ2 = occs2[oi2];
          score = score_occurence_pair(occ1, occ2, custom_params);
          if (score > result) {
            result = score;
          }
        }
    }
  /*printf("score_occurences = %d\n", result);*/
  return result;
}

int score_context_paths(SPECTRUM_TREE* spectrum_tree1, SPECTRUM_TREE* spectrum_tree2, CUSTOM_PARM* custom_params)
{
  CONTEXT_EXAMPLE* ce1;
  CONTEXT_EXAMPLE* ce2;
  int score1122 = 0;
  int score1221 = 0;

  ce1 = spectrum_tree1->context_example;
  ce2 = spectrum_tree2->context_example;

  score1122 = score_occurences(
      ce1->entity_occurences_1, ce1->entity_occurence_count_1,
      ce2->entity_occurences_1, ce2->entity_occurence_count_1,
      custom_params
  )
  + score_occurences(
      ce1->entity_occurences_2, ce1->entity_occurence_count_2,
      ce2->entity_occurences_2, ce2->entity_occurence_count_2,
      custom_params
  );
  /*printf("score1122 = %d\n", score1122);*/

  score1221 = score_occurences(
      ce1->entity_occurences_1, ce1->entity_occurence_count_1,
      ce2->entity_occurences_2, ce2->entity_occurence_count_2,
      custom_params
  )
  + score_occurences(
      ce1->entity_occurences_2, ce1->entity_occurence_count_2,
      ce2->entity_occurences_1, ce2->entity_occurence_count_1,
      custom_params
  );
  /*printf("score1122 = %d\n-----\n", score1221);*/

  if (score1122 > score1221) {
    return score1122;
  }
  else {
    return score1221;
  }

}

int
count_common_q_grams(SPECTRUM_TREE* spectrum_tree1,
    SPECTRUM_TREE* spectrum_tree2, CUSTOM_PARM* custom_params)
{
  int result = 0;
  int i = 0;
  struct hashtable_itr *itr;
  struct hashtable* ht2;
  unsigned char* val1;
  Q_PATH* key1;
  unsigned char* val2;

  /* take the one with less nodes */
  if (spectrum_tree1->node_count <= spectrum_tree2->node_count)
    {
      for (i = 0; i < spectrum_tree1->q_profile->q_patterns_count; i++)
        {
          if (hashtable_count(spectrum_tree1->q_profile->q_patterns[i]) > 0)
            {
              itr
                  = hashtable_iterator(spectrum_tree1->q_profile->q_patterns[i]);
              assert(itr != NULL);
              ht2 = spectrum_tree2->q_profile->q_patterns[i];
              assert(ht2 != NULL);
              do
                {
                  val1 = hashtable_iterator_value(itr);
                  if (val1 != NULL)
                    {
                      key1 = hashtable_iterator_key(itr);
                      assert(key1 != NULL);
                      val2 = hashtable_search(ht2, key1);
                      if (val2 != NULL)
                        {
                          /*printf("result += %d * %d\n", (*val1), (*val2));*/
                          if (custom_params->spectrum_tree_kernel_count_mode
                              == 'm')
                            {
                              result += (*val1) * (*val2);
                            }
                          else if (custom_params->spectrum_tree_kernel_count_mode
                              == 'u')
                            {
                              result++;
                            }
                          else
                            {
                              perror(
                                  "Unexpected value of spectrum_tree_kernel_count_mode.");
                              exit(-1);
                            }
                          /*printf("result = %d\n", result);*/
                        }
                    }
                }
              while (hashtable_iterator_advance(itr));
              free(itr);
            }
        }

      return result;
    }
  else
    {
      /* swap */
      return count_common_q_grams(spectrum_tree2, spectrum_tree1, custom_params);
    }
}

unsigned int
read_int(char* src, int* pos)
{
  char c;
  char* tmp;
  int i;
  int start;
  int len;
  int result;
  start = *pos;

  /* at least one digit */
  c = src[*pos];
  assert(isdigit(c));
  *pos = *pos + 1;

  while (1)
    {
      c = src[*pos];
      if (!isdigit(c))
        {
          break;
        }
      *pos = *pos + 1;
    }

  len = *pos - start;
  tmp = (char*) my_malloc((len + 1) * sizeof(char));
  for (i = 0; i < len; ++i)
    {
      tmp[i] = src[start + i];
    }
  tmp[i] = '\0';

  result = atoi(tmp);
  free(tmp);
  return result;
}

Q_GRAM*
read_q_gram(char* src, int* pos, int q)
{
  int labels_read = 0;
  int nr;
  Q_GRAM* result;
  char c;

  consumespaces(src, pos);

  c = src[*pos];
  if (c == '\0')
    {
      /* end of input */
      return NULL;
    }

  assert(c == '(');
  *pos = *pos + 1;

  result = (Q_GRAM*) my_malloc(sizeof(Q_GRAM));
  nr = read_int(src, pos);
  assert(nr <= UCHAR_MAX);
  result->pattern_id = nr;

  result->q_path = (Q_PATH*) my_malloc(sizeof(Q_PATH));
  result->q_path->length = q;
  result->q_path->labels = (string_index_key_t*) my_malloc(q
      * sizeof(string_index_key_t));

  while (1)
    {
      c = src[*pos];
      if (c == ')')
        {
          /* end of q_gram */
          *pos = *pos + 1;
          break;
        }

      assert(labels_read < q);

      consumespaces(src, pos);

      nr = read_int(src, pos);
      assert(nr <= string_index_max_capacity);
      result->q_path->labels[labels_read++] = nr;

    };
  assert(labels_read == q);

  return result;
}

Q_PATH*
read_q_path(char* src, int* pos, CUSTOM_PARM* custom_params)
{
  int nr;
  Q_PATH* result;
  char c;

  consumespaces(src, pos);

  c = src[*pos];
  if (c == '\0')
    {
      /* end of input */
      return NULL;
    }
  else if (c != '(')
    {
      /* not a q_path */
      return NULL;
    }

  assert(c == '(');
  *pos = *pos + 1;

  result = (Q_PATH*) my_malloc(sizeof(Q_PATH));

  result->length = 0;
  result->labels = NULL;

  while (1)
    {
      consumespaces(src, pos);
      c = src[*pos];
      if (c == ')')
        {
          /* end of q_gram */
          break;
        }

      nr = read_int(src, pos);
      assert(nr <= string_index_max_capacity);
      result->length++;

      result->labels = (string_index_key_t*) my_realloc(result->labels,
          result->length * sizeof(string_index_key_t));
      result->labels[result->length - 1] = nr;

    }
  assert(c == ')');
  *pos = *pos + 1;

  return result;
}
int path_count(CUSTOM_PARM* custom_params)
{
  switch (custom_params->q_gram_length_mode)
    {
  case 'f':
    return 1;
  case 'u':
    return (custom_params->q_gram_length - 1) / 2;
  default:
    perror("Unexpected custom_params->q_gram_length_mode");
    exit(-1);
    }
}
int path_index(int q_path_length, CUSTOM_PARM* custom_params)
{
  switch (custom_params->q_gram_length_mode)
    {
  case 'f':
    return 0;
  case 'u':
    return ((q_path_length - 3) / 2);
  default:
    perror("Unexpected custom_params->q_gram_length_mode");
    exit(-1);
    }
}

CONTEXT_ENTITY_OCCURENCE*
read_entity_occurence(char* src, int* pos, CUSTOM_PARM* custom_params)
{
  char c;
  int cnt = 0;
  int i;
  CONTEXT_ENTITY_OCCURENCE* result;
  Q_PATH* q_path;

  consumespaces(src, pos);
  c = src[*pos];

  if (c == '\0')
    {
      /* end of input */
      return NULL;
    }
  else if (c != '(')
    {
      /* not an CONTEXT_ENTITY_OCCURENCE */
      return NULL;
    }

  assert(c == '(');
  *pos = *pos + 1;

  cnt = path_count(custom_params);

  result = (CONTEXT_ENTITY_OCCURENCE*) my_malloc(
      sizeof(CONTEXT_ENTITY_OCCURENCE));
  result->path_counts = (unsigned short*) my_malloc(cnt
      * sizeof(unsigned short));
  result->pathes = (Q_PATH***) my_malloc(cnt * sizeof(Q_PATH**));
  for (i = 0; i < cnt; ++i)
    {
      result->path_counts[i] = 0;
      result->pathes[i] = NULL;
    }

  while (1)
    {
      q_path = read_q_path(src, pos, custom_params);
      if (q_path == NULL)
        {
          break;
        }

      i = path_index(q_path->length, custom_params);

      assert(i < cnt);
      result->path_counts[i]++;
      result->pathes[i] = (Q_PATH**) my_realloc(result->pathes[i],
          result->path_counts[i] * sizeof(Q_PATH*));
      result->pathes[i][result->path_counts[i] - 1] = q_path;

    }

  c = src[*pos];
  assert(c == ')');
  *pos = *pos + 1;

  return result;
}

char*
read_example_id(char* src, int* pos, CUSTOM_PARM* custom_params)
{
  char c;
  char* result;
  int i;
  switch (custom_params->custom_kernel_input_format)
    {
  case 't':
  case 'q':
    return NULL;
    break;
  case 'c':

    c = src[*pos];
    assert(c == '#');
    *pos = *pos + 1;

    i = *pos;

    while (src[i] != '\0')
      {
        i++;
      }

    result = (char*) my_malloc((i - *pos + 1) * sizeof(char));
    i = *pos;
    while (src[*pos] != '\0')
      {
        result[*pos - i] = src[*pos];
        *pos = *pos + 1;
      }
    result[*pos - i] = '\0';

    return result;

    break;
  default:
    perror("Unexpected custom_kernel_input_format value.");
    exit(-1);
    }
}
CONTEXT_EXAMPLE*
read_context_example(char* src, int* pos, CUSTOM_PARM* custom_params)
{
  CONTEXT_EXAMPLE *result;
  CONTEXT_ENTITY_OCCURENCE* occurence;
  char c;

  consumespaces(src, pos);
  c = src[*pos];

  if (c == '\0')
    {
      /* end of input */
      return NULL;
    }
  else if (c != '(')
    {
      /* not an CONTEXT_ENTITY_OCCURENCE */
      return NULL;
    }

  assert(c == '(');
  *pos = *pos + 1;

  result = (CONTEXT_EXAMPLE*) my_malloc(sizeof(CONTEXT_EXAMPLE));
  result->entity_occurence_count_1 = 0;
  result->entity_occurence_count_2 = 0;
  result->entity_occurences_1 = NULL;
  result->entity_occurences_2 = NULL;

  while (1)
    {
      occurence = read_entity_occurence(src, pos, custom_params);
      if (occurence == NULL)
        {
          break;
        }
      result->entity_occurence_count_1++;
      result->entity_occurences_1 = (CONTEXT_ENTITY_OCCURENCE**) my_realloc(
          result->entity_occurences_1, result->entity_occurence_count_1
              * sizeof(CONTEXT_ENTITY_OCCURENCE*));

      result->entity_occurences_1[result->entity_occurence_count_1 - 1]
          = occurence;
    }
  c = src[*pos];
  assert(c == ')');
  *pos = *pos + 1;

  consumespaces(src, pos);
  c = src[*pos];
  assert(c == '(');
  *pos = *pos + 1;
  while (1)
    {
      occurence = read_entity_occurence(src, pos, custom_params);
      if (occurence == NULL)
        {
          break;
        }
      result->entity_occurence_count_2++;
      result->entity_occurences_2 = (CONTEXT_ENTITY_OCCURENCE**) my_realloc(
          result->entity_occurences_2, result->entity_occurence_count_2
              * sizeof(CONTEXT_ENTITY_OCCURENCE*));

      result->entity_occurences_2[result->entity_occurence_count_2 - 1]
          = occurence;
    }
  c = src[*pos];
  assert(c == ')');
  *pos = *pos + 1;

  return result;
}

Q_PROFILE*
read_q_profile(char* src, unsigned char q_gram_length,
    unsigned int q_patterns_count)
{
  Q_PROFILE* result;
  Q_GRAM* q_gram;
  int q_grams_read = 0;
  int nr;
  int i;
  char c;
  unsigned char* val;
  int* pos;

  result = (Q_PROFILE*) my_malloc(sizeof(Q_PROFILE));
  result->q_patterns_count = q_patterns_count;
  result->q_patterns = (struct hashtable**) my_malloc(q_patterns_count
      * sizeof(struct hashtable*));
  for (i = 0; i < q_patterns_count; i++)
    {
      result->q_patterns[i] = create_hashtable(pow(2, (q_gram_length + 1) / 2)
          * 50, /* 50 is the typical sentence length */
      path_hash, path_equal, free_q_path_void);
    }

  pos = (int*) my_malloc(sizeof(int));
  *pos = 0;
  while (1)
    {
      consumespaces(src, pos);
      q_gram = read_q_gram(src, pos, q_gram_length);
      if (q_gram == NULL)
        {
          break;
        }
      q_grams_read++;

      c = src[*pos];
      assert(c == ':');
      *pos = *pos + 1;
      nr = read_int(src, pos);
      assert(nr <= UCHAR_MAX);

      val = (unsigned char*) hashtable_search(
          result->q_patterns[q_gram->pattern_id], q_gram->q_path);
      if (val == NULL)
        {
          val = (unsigned char*) my_malloc(sizeof(unsigned char));
          *val = 0;
          hashtable_insert(result->q_patterns[q_gram->pattern_id],
              q_gram->q_path, val);
          free(q_gram);
        }
      else
        {
          free_q_gram(q_gram);
        }
      *val = *val + nr;
    }
  assert(q_grams_read> 0);
  free(pos);
  return result;
}

Q_PROFILE *
create_q_profile(SPECTRUM_TREE* spectrum_tree, unsigned char q,
    STRING_INDEX* string_index, unsigned int q_patterns_count)
{
  Q_PROFILE *result;
  int i = 0;
  int j = 0;
  int ii = 0;
  int d = 0;
  string_index_key_t l;
  int k = 0;
  int q_maxd_min = 0;
  int m = 0;
  int** cnt;
  int* cnt_element_sizes;
  int** sft;
  int* sft_element_sizes;
  int new_size;
  struct hashtable*** f; /* two-dimensional array of hashtables
   with q_path* keys and unsigned char* values
   the value pointers will not be shared with other hashtables
   the value pointers should be freed on every remove
   and on hashtable_destroy
   */
  struct hashtable** label; /* one-dimensional array of hashtables
   with q_gram* keys and unsigned char* values
   the value pointers will not be shared with other hashtables
   the value pointers should be freed on every remove
   and on hashtable_destroy
   */
  unsigned char f_val;
  TREE* t;
  struct hashtable_itr *itr;
  Q_PATH* w;
  Q_PATH* wl;
  Q_PATH* q_path;
  Q_GRAM* q_gram;
  unsigned char* val;
  Q_GRAM* ptn;
  /*unsigned char ruby_test = 1;*/

  /*printf("create_q_profile q= %u\n", q);*/

  result = (Q_PROFILE*) my_malloc(sizeof(Q_PROFILE));
  result->q_patterns_count = q_patterns_count;
  result->q_patterns = (struct hashtable**) my_malloc(q_patterns_count
      * sizeof(struct hashtable*));
  for (i = 0; i < q_patterns_count; i++)
    {
      result->q_patterns[i] = create_hashtable(spectrum_tree->node_count,
          path_hash, path_equal, free_q_path_void);
    }

  cnt = (int**) my_malloc((spectrum_tree->max_depth + 1) * sizeof(int*));
  cnt_element_sizes = (int*) my_malloc((spectrum_tree->max_depth + 1)
      * sizeof(int));
  for (d = (spectrum_tree->max_depth); d >= 0; d--)
    {
      /*printf("ctn d = %d\n", d);*/
      cnt_element_sizes[d] = 0;
      cnt[d] = NULL;

      q_maxd_min = (q - 1 < spectrum_tree->max_depth ? q - 1
          : spectrum_tree->max_depth);
      /*printf("q_maxd_min = %d\n", q_maxd_min);*/
      for (k = 1; k <= q_maxd_min; k++)
        {
          m = d - q + 1 + (2 * k);
          if (k <= m && m <= spectrum_tree->max_depth)
            {
              new_size = cnt_element_sizes[d] + 2;
              /*printf("resizing cnt[%d] to %d, << %d, %d\n", d, new_size, m-1, k-1);*/
              cnt[d] = (int*) my_realloc(cnt[d], new_size * sizeof(int));
              cnt[d][cnt_element_sizes[d]] = m - 1;
              cnt[d][cnt_element_sizes[d] + 1] = k - 1;
              cnt_element_sizes[d] = new_size;
            }
        }
    }

  sft = (int**) my_malloc((spectrum_tree->max_depth) * sizeof(int*));
  sft_element_sizes
      = (int*) my_malloc((spectrum_tree->max_depth) * sizeof(int));
  sft_element_sizes[0] = 0;
  sft[0] = NULL;
  for (d = (spectrum_tree->max_depth - 1); d >= 1; d--)
    {
      sft_element_sizes[d] = 0;
      sft[d] = NULL;
      for (k = 1; k <= (q - 1); k++)
        {
          m = d + k;
          if (0 <= m && m <= spectrum_tree->max_depth)
            {
              new_size = sft_element_sizes[d] + 2;
              /* printf("resizing sft[%d] to %d, << %d, %d\n", d, new_size, m-1, k-1);*/
              sft[d] = (int*) my_realloc(sft[d], new_size * sizeof(int));
              sft[d][sft_element_sizes[d]] = m - 1;
              sft[d][sft_element_sizes[d] + 1] = k - 1;
              sft_element_sizes[d] = new_size;
            }
        }
    }

  /* @f = Array.new( @maxd ) { Array.new(@q-1) { Hash.new(0) } } */
  f = (struct hashtable***) my_malloc(spectrum_tree->max_depth
      * sizeof(struct hashtable**));
  for (i = 0; i < spectrum_tree->max_depth; i++)
    {
      f[i]
          = (struct hashtable**) my_malloc((q - 1) * sizeof(struct hashtable*));
      for (k = 0; k < q - 1; k++)
        {
          f[i][k] = create_hashtable(spectrum_tree->node_count, path_hash,
              path_equal, free_q_path_void);
        }
    }

  /* @label = Array.new( @maxd+1 ) { Hash.new(0) } # 1st element is a dummy */
  label = (struct hashtable**) my_malloc((spectrum_tree->max_depth + 1)
      * sizeof(struct hashtable*));
  /* 1st element is a dummy */
  for (i = 0; i < spectrum_tree->max_depth + 1; i++)
    {
      label[i] = create_hashtable(spectrum_tree->node_count, q_gram_hash,
          q_gram_equal, free_q_gram_void);
    }

  /*
   if (ruby_test) {
   for (i = 0; i < spectrum_tree->node_count; i++) {
   t = spectrum_tree->nodes[i];
   t->label = (t->label % 9) +1;
   }
   printf("ruby qgram.rb -q %d -t ", q);
   for (i = 0; i < spectrum_tree->node_count; i++) {
   //printf("i = %d from %d\n", i, spectrum_tree->node_count);
   t = spectrum_tree->nodes[i];
   d = t->depth;
   if (i != 0) {
   printf(",");
   }
   printf("%d",d);
   }
   printf(" -l ");
   for (i = 0; i < spectrum_tree->node_count; i++) {
   //printf("i = %d from %d\n", i, spectrum_tree->node_count);
   t = spectrum_tree->nodes[i];
   l = t->label;
   if (i != 0) {
   printf(",");
   }
   printf("%d", l % 10);
   //print_escaped_element(value_for_index(l, string_index));
   }
   printf(" >> ruby.out\n");
   }
   */
  /* main */
  /* @d.each_with_index do |d, i| */
  for (i = 0; i < spectrum_tree->node_count; i++)
    {
      t = spectrum_tree->nodes[i];
      d = t->depth;

      /*l = @l[i]*/
      l = t->label;
      /*printf("i = %d (%d), d = %d, l = %d\n", i, spectrum_tree->node_count, d, l);*/

      /* @cnt[d].each do |j,k| */
      for (ii = 0; ii < cnt_element_sizes[d];)
        {
          j = cnt[d][ii++];
          k = cnt[d][ii++];

          /*printf("  (cnt) j = %d, k = %d\n", j, k);*/

          /* @f[j][k].each_pair do |w, f| */
          if (hashtable_count(f[j][k]) > 0)
            {
              itr = hashtable_iterator(f[j][k]);
              assert(itr != NULL);
              do
                {
                  w = (Q_PATH*) hashtable_iterator_key(itr);
                  f_val = *((unsigned char *) hashtable_iterator_value(itr));
                  /*
                   printf("    (cnt) f = %d, w = ", f_val);
                   print_q_path(w, string_index);
                   printf("\n");
                   */

                  /* @label[d][ [k,w+l] ] += f */
                  q_gram = create_q_gram(k, concat_q_path(w, l, q));
                  val = (unsigned char*) hashtable_search(label[d], q_gram);
                  /*
                   printf("    (cnt) --- label[%d][", d);
                   print_q_gram(q_gram, string_index);
                   */
                  if (val == NULL)
                    {
                      val = (unsigned char*) my_malloc(sizeof(unsigned char));
                      *val = 0;
                      hashtable_insert(label[d], q_gram, val);
                    }
                  else
                    {
                      /* we have not inserted the q_gram - we need to destroy it */
                      free_q_gram(q_gram);
                    }
                  *val += f_val;
                  /*printf("] = %d\n", (*val));*/
                }
              while (hashtable_iterator_advance(itr));
              free(itr);
            }
        }
      /*printf("&&& i = %d (%d), d = %d, l = %d\n", i, spectrum_tree->node_count, d, l);*/

      /*end if d < @maxd*/
      if (d < spectrum_tree->max_depth)
        {
          /*@sft[d].each do |j,k|*/
          for (ii = 0; ii < sft_element_sizes[d];)
            {
              j = sft[d][ii++];
              k = sft[d][ii++];
              /*printf("  (sft) ++ j = %d, k = %d\n", j, k);*/

              /*@f[j][k].each_pair do |w,f|*/
              if (hashtable_count(f[j][k]) > 0)
                {
                  itr = hashtable_iterator(f[j][k]);
                  assert(itr != NULL);
                  /*printf("  (sft) iterating over f[%d][%d]\n", j, k);*/
                  do
                    {
                      w = (Q_PATH*) hashtable_iterator_key(itr);
                      f_val
                          = *((unsigned char *) hashtable_iterator_value(itr));

                      /*
                       printf("    (sft) ** f = %d, w = ", f_val);
                       print_q_path(w, string_index);
                       printf("\n");
                       */

                      /*@f[j][k+1][w+l] += f if k+1 < @q-1*/
                      if (k + 1 < q - 1)
                        {
                          /*printf("concat_q_path 1\n");*/
                          q_path = concat_q_path(w, l, q);

                          /*
                           printf("    (sft) f[%d][%d][", j, k+1);
                           print_q_path(q_path, string_index);
                           */

                          val = (unsigned char*) hashtable_search(f[j][k + 1],
                              q_path);
                          if (val == NULL)
                            {
                              val = (unsigned char*) my_malloc(
                                  sizeof(unsigned char));
                              *val = 0;
                              hashtable_insert(f[j][k + 1], q_path, val);
                            }
                          else
                            {
                              /* we have not inserted the q_path - we need to destroy it */
                              free_q_path(q_path);
                            }
                          *val += f_val;
                          /*
                           printf("] = %d\n", (*val));
                           printf("    (sft) f[%d][%d].size after update = %d\n", j, k+1, hashtable_count(f[j][k+1]));
                           */
                        }

                      /*@f[j][k].delete(w)*/
                    }
                  while (hashtable_iterator_remove(itr) /*remove does also advance */);
                  free(itr);
                }
            }
        }
      /*printf("!!! i = %d (%d), d = %d, l = %d\n", i, spectrum_tree->node_count, d, l);*/

      /*@f[d-1][0][l] += 1 if d > 0;*/
      if (d > 0)
        {

          q_path = concat_q_path(NULL, l, q);
          val = (unsigned char*) hashtable_search(f[d - 1][0], q_path);

          /*
           printf("(mid) f[%d][%d][", d-1, 0);
           print_q_path(q_path, string_index);
           */

          if (val == NULL)
            {
              val = (unsigned char*) my_malloc(sizeof(unsigned char));
              *val = 0;
              hashtable_insert(f[d - 1][0], q_path, val);
            }
          else
            {
              free_q_path(q_path);
            }
          *val += 1;
          /*
           printf("] = %d\n", (*val));
           printf("(mid) f[%d][%d].size after update = %d\n", d-1, 0, hashtable_count(f[d-1][0]));
           */
        }

      /*end if d < @maxd*/
      if (d < spectrum_tree->max_depth)
        {
          /*@label[d+1].each_pair do |ptn, f|*/
          /*printf("(lbl) label[%d].size = %d\n", d+1, hashtable_count(label[d+1]));*/
          if (hashtable_count(label[d + 1]) > 0)
            {
              itr = hashtable_iterator(label[d + 1]);
              assert(itr != NULL);
              do
                {
                  ptn = (Q_GRAM*) hashtable_iterator_key(itr);
                  f_val = *((unsigned char *) hashtable_iterator_value(itr));

                  /*k, w = ptn*/
                  k = ptn->pattern_id;
                  w = ptn->q_path;

                  /*wl = w + l*/
                  wl = concat_q_path(w, l, q);

                  /*
                   printf("(lbl) k = %d, w = ", k);
                   print_q_path(w, string_index);
                   printf(", wl = ");
                   print_q_path(wl, string_index);
                   printf("\n");
                   */

                  /*if wl.size == @q*/
                  if (wl->length == q)
                    {
                      /*@ptn[k][wl] += f*/
                      val = (unsigned char*) hashtable_search(
                          result->q_patterns[k], wl);
                      /*
                       printf("(lbl) ptn[%d][", k);
                       print_q_path(wl, string_index);
                       */
                      if (val == NULL)
                        {
                          val = (unsigned char*) my_malloc(
                              sizeof(unsigned char));
                          *val = 0;
                          hashtable_insert(result->q_patterns[k], wl, val);
                        }
                      else
                        {
                          free_q_path(wl);
                        }
                      *val += f_val;
                      /*printf("] = %d\n", (*val));*/
                    }
                  else
                    {
                      /*@label[d][ [k,wl[0,@q]] ] += f # leftmost @q lenth string*/
                      q_gram
                          = create_q_gram(k, q_subpath(0, q < wl->length ? q
                              : wl->length, wl, q) /* TODO: check if wl[0,@q] is important */);
                      free_q_path(wl);
                      val = (unsigned char*) hashtable_search(label[d], q_gram);
                      /*
                       printf("(lbl) ... label[%d][", d);
                       print_q_gram(q_gram, string_index);
                       */

                      if (val == NULL)
                        {
                          val = (unsigned char*) my_malloc(
                              sizeof(unsigned char));
                          *val = 0;
                          hashtable_insert(label[d], q_gram, val);
                        }
                      else
                        {
                          free_q_gram(q_gram);
                        }
                      *val += f_val;
                      /*printf("] = %d\n", (*val));*/
                    }
                  /*@label[d+1].delete(ptn)*/
                  val = (unsigned char*) hashtable_remove(label[d + 1], ptn); /* TODO: the returned value should probably be freed */
                  free(val);

                }
              while (hashtable_iterator_advance(itr));
              free(itr);
            }
        }
      /*printf("??? i = %d (%d), d = %d, l = %d\n", i, spectrum_tree->node_count, d, l);*/

    }

  /*printf("(fin) label[0].size = %d\n", hashtable_count(label[0]));*/

  /*@label[0].each_pair do |ptn, f|*/
  if (hashtable_count(label[0]) > 0)
    {
      itr = hashtable_iterator(label[0]);
      assert(itr != NULL);
      do
        {
          ptn = (Q_GRAM*) hashtable_iterator_key(itr);
          f_val = *((unsigned char *) hashtable_iterator_value(itr));

          /*k, w = ptn*/
          k = ptn->pattern_id;
          w = ptn->q_path;

          /*
           printf("  (fin) k = %d, w = ", k);
           print_q_path(w, string_index);
           printf("\n");
           */

          /*next unless k == @q-2 # Pattern k is from 0 to q-2 here*/
          if (k == q - 2)
            {

              /*
               printf("    (fin) ptn[%d][", k);
               print_q_path(w, string_index);
               */

              /*@ptn[k][w] += f*/
              val
                  = (unsigned char *) hashtable_search(result->q_patterns[k], w);
              if (val == NULL)
                {
                  val = (unsigned char*) my_malloc(sizeof(unsigned char));
                  *val = 0;
                  hashtable_insert(result->q_patterns[k], q_path_dup(w, q), val);
                }
              *val += f_val;

              /*printf("] = %d\n", (*val));*/
            }
        }
      while (hashtable_iterator_advance(itr));
      free(itr);
    }

  /* clean up */

  /* destroy label */
  for (i = 0; i < spectrum_tree->max_depth + 1; i++)
    {
      hashtable_destroy(label[i], 1 /* destroy values */);
    }
  free(label);

  /* destroy f */
  for (i = 0; i < spectrum_tree->max_depth; i++)
    {
      for (k = 0; k < q - 1; k++)
        {
          hashtable_destroy(f[i][k], 1 /* destroy values */);
        }
      free(f[i]);
    }
  free(f);

  /* destroy sft */
  for (i = 0; i < (spectrum_tree->max_depth); i++)
    {
      if (sft_element_sizes[i] > 0)
        {
          free(sft[i]);
        }
    }
  free(sft);
  free(sft_element_sizes);

  /* destroy cnt */
  for (i = 0; i < (spectrum_tree->max_depth + 1); i++)
    {
      if (cnt_element_sizes[i] > 0)
        {
          free(cnt[i]);
        }
    }
  free(cnt);
  free(cnt_element_sizes);

  /*test_q_profile(result, q, string_index);*/
  return result;
}

void
free_q_profile(Q_PROFILE* q_profile, unsigned char q)
{
  int i;
  if (q_profile != NULL)
    {
      if (q_profile->q_patterns != NULL)
        {
          for (i = 0; i < q_profile->q_patterns_count; i++)
            {
              hashtable_destroy(q_profile->q_patterns[i], 1);
            }
          free(q_profile->q_patterns);
        }
      free(q_profile);
    }
}

void
free_spectrum_tree(SPECTRUM_TREE *spectrum_tree, unsigned char q)
{
  if (spectrum_tree != NULL)
    {
      if (spectrum_tree->nodes != NULL)
        {
          free(spectrum_tree->nodes);
        }
      free_tree(spectrum_tree->root);
      free_q_profile(spectrum_tree->q_profile, q);
      if (spectrum_tree->bag_of_words != NULL)
        {
          hashtable_destroy(spectrum_tree->bag_of_words, 1);
        }
      free(spectrum_tree);
    }
}

unsigned int
q_gram_hash(void* q_gram)
{
  Q_GRAM* qgm;
  unsigned int hash;
  qgm = (Q_GRAM*) q_gram;
  hash = path_hash(qgm->q_path);
  return ((hash << 5) + hash) + qgm->pattern_id;
}

int
index_equal(void* k1, void* k2)
{
  return *((string_index_key_t*) k1) == *((string_index_key_t*) k2);
}
unsigned int
index_hash(void* index)
{
  return (unsigned int) *((string_index_key_t*) index);
}
unsigned int
path_hash(void* path)
{
  unsigned int hash = 5381;
  Q_PATH* q_path;
  string_index_key_t* label;
  string_index_key_t* label_end;

  q_path = (Q_PATH*) path;
  label = q_path->labels;
  label_end = label + q_path->length;

  while (label < label_end)
    {
      hash = ((hash << 5) + hash) + (*label); /* hash * 33 + c */
      label++;
    }
  return hash;
}

int
q_gram_equal(void* q_gram1, void* q_gram2)
{
  Q_GRAM* qgm1;
  Q_GRAM* qgm2;
  qgm1 = (Q_GRAM*) q_gram1;
  qgm2 = (Q_GRAM*) q_gram2;
  return (qgm1->pattern_id == qgm2->pattern_id) && path_equal(qgm1->q_path,
      qgm2->q_path);
}

int
path_equal(void* path1, void* path2)
{
  Q_PATH* q_path1;
  Q_PATH* q_path2;
  q_path1 = (Q_PATH*) path1;
  q_path2 = (Q_PATH*) path2;

  return (q_path1->length == q_path2->length && 0 == memcmp(q_path1->labels,
      q_path2->labels, q_path1->length * sizeof(string_index_key_t)));
}

string_index_key_t matrix_get(int i, int j, int i_len, string_index_key_t* d) {
  return d[(j*i_len)+i];
}
void matrix_set(int i, int j, string_index_key_t val, int i_len, string_index_key_t* d) {
  d[(j*i_len)+i] = val;
}

int levenshtein_similarity(string_index_key_t* s, string_index_key_t* t, int length) {
  return length - levenshtein_distance(s, length, t, length);
}
int i_levenshtein_similarity(string_index_key_t* s, string_index_key_t* t, int length, CUSTOM_PARM* custom_params) {
  return (((length -1)/2)* (I_DEPENDENCY_MISMATCH_COST + I_ENTITY_MISMATCH_COST)) - i_levenshtein_distance(s, length, t, length, custom_params);
}

int i_replace_cost(
    string_index_key_t dependency_type1,
    string_index_key_t node_label1,
    string_index_key_t dependency_type2,
    string_index_key_t node_label2,
    CUSTOM_PARM* custom_params) {
  int result = 0;
  if (dependency_type1 != dependency_type2) {
    result += I_DEPENDENCY_MISMATCH_COST;
  }
  if (node_label1 != node_label2) {
    if (node_label1 == OTHER_ENTITY_LABEL
        || node_label2 == OTHER_ENTITY_LABEL
        || node_label1 == THIS_OR_OTHER_ENTITY_LABEL
        || node_label2 == THIS_OR_OTHER_ENTITY_LABEL
    ) {
      /* this is very expensive */
      result += I_ENTITY_MISMATCH_COST;
    }
    else {
      result += I_NODE_LABEL_MISMATCH_COST;
    }
  }
  return result;
}

int i_ins_del_cost(
    string_index_key_t node_label,
    CUSTOM_PARM* custom_params) {
  int result = I_DEPENDENCY_MISMATCH_COST;
  if (node_label == OTHER_ENTITY_LABEL
      || node_label == THIS_OR_OTHER_ENTITY_LABEL
  ) {
    /* this is very expensive */
    result += I_ENTITY_MISMATCH_COST;
  }
  else {
    result += I_NODE_LABEL_MISMATCH_COST;
  }
  return result;
}

int i_levenshtein_distance(string_index_key_t* s, int m, string_index_key_t* t, int n, CUSTOM_PARM* custom_params) {
  string_index_key_t** d;
  int i,j, m_plus_1, n_plus_1, i_type, j_type;
  m_plus_1 = ((m-1)/2) +1;
  n_plus_1 = ((n-1)/2) +1;
  /*d is a table with m+1 rows and n+1 columns*/
  d = (string_index_key_t**) my_malloc(m_plus_1 * sizeof(string_index_key_t*));


  for (i = 0; i < m_plus_1; ++i) {
    d[i] = (string_index_key_t*) my_malloc(n_plus_1 * sizeof(string_index_key_t));
    d[i][0] = i;
  }

  for (j = 0; j < n_plus_1; ++j) {
    d[0][j] = j;
  }

  for (i = 1; i < m_plus_1; ++i) {
    for (j = 1; j < n_plus_1; ++j) {

      i_type = ((i-1)*2) + 1;
      j_type = ((j-1)*2) + 1;

      d[i][j] =
          min(
              d[i-1][j] + i_ins_del_cost(s[i_type+1], custom_params), /*deletion*/
              d[i][j-1] + i_ins_del_cost(s[j_type+1], custom_params), /*insertion*/
              d[i-1][j-1] + i_replace_cost(s[i_type], s[i_type+1], t[j_type], t[j_type+1], custom_params)
          );
    }
  }

  i = d[m_plus_1 - 1][n_plus_1 - 1];
  for (i = 0; i < m_plus_1; ++i) {
    free(d[i]);
  }
  free(d);
  return i;
}


int levenshtein_distance(string_index_key_t* s, int m, string_index_key_t* t, int n) {
  string_index_key_t* d;
  string_index_key_t cost;
  int i,j, m_plus_1;
  m_plus_1 = m+1;
  /*d is a table with m+1 rows and n+1 columns*/
  d = (string_index_key_t*) my_malloc((m_plus_1) * (n +1) * sizeof(string_index_key_t));

  for (i = 0; i <= m; ++i) {
    matrix_set(i, 0, i, m_plus_1, d);
  }

  for (j = 0; j <= n; ++j) {
    matrix_set(0, j, j, m_plus_1, d);
  }

  for (i = 1; i <= m; ++i) {
    for (j = 1; j <= n; ++j) {
      if (s[i -1] == t[j -1]) {
        cost = 0;
      }
      else {
        cost = 1;
      }

      matrix_set(i, j,
          min(
              matrix_get(i-1, j, m_plus_1, d) + 1,
              matrix_get(i, j-1, m_plus_1, d) +1,
              matrix_get(i-1, j-1, m_plus_1, d) +cost
          ),
          m_plus_1, d);
    }
  }

  i = matrix_get(m, n, m_plus_1, d);
  free(d);
  return i;
}

string_index_key_t min(string_index_key_t a, string_index_key_t b, string_index_key_t c) {
  if (a < b && a < c) {
    return a;
  }
  else if (b < a && b < c) {
    return b;
  }
  else {
    return c;
  }
}

