/* tree.h
 *
 * Written by Mark Johnson, 27th May 1997
 */

#ifndef TREE_H
#define TREE_H
#define NDEBUG

#define tree_depth_max UCHAR_MAX
#define tree_depth_min 1
#define node_count_max USHRT_MAX
#define UNINITIALIZED_KERNEL_VALUE (-1)
#define UNINITIALIZED_KERNEL_VALUE_DBL (-1.0)

#define THIS_ENTITY_LABEL 0
#define OTHER_ENTITY_LABEL 1
#define THIS_OR_OTHER_ENTITY_LABEL 2

#define I_DEPENDENCY_MISMATCH_COST 2
#define I_ENTITY_MISMATCH_COST 3
#define I_NODE_LABEL_MISMATCH_COST 1


#include "string_index.h"
#include "hashtable/hashtable.h"

typedef unsigned char tree_depth_t;
typedef struct custom_parm {
  unsigned char q_gram_length;
  unsigned char q_gram_length_mode;
  /* 'f' ... fixed q_gram_length
   * 'u' ... all up to q_gram_length
  */

  unsigned char custom_kernel_normalize_whole;
  unsigned char custom_kernel_normalize_parts;
  float second_kernel_factor;
  /* '1' ... the first only
   * '2' ... the second only
   * 'a' ... addition
   * 'm' ... multiplication */
  char second_kernel_factor_operation;
  char kernels_combine_operation;
  /* 'a' ... all
   * 't' ... only terminals
  */
  char bow_only_terminals;
  char custom_kernel_input_format;
  /* 'q' ... q-grams
   * 't' ... PTB bracketing trees
   * 'c' ... entity context
   */
  char spectrum_tree_kernel_count_mode;
  /* 'm' ... multiply the real counts
   * 'u' ... increment by 1 for each unique match
   */
  char spectrum_tree_kernel_match_mode;
  /* 'e' ... exact
   * 't' ... tolerant - dependecy types must match, nodel labels need not
   * 'l' ... (naive) levenshtein similarity
   * 'i' ... (intelligent) levenshtein similarity
   */
  char entity_context_input_format_entity_position;
  /* '0' ... beginning
   * 'a' ... anywhere
   * 'o' ... outside
   */
} CUSTOM_PARM;

typedef struct tree {
  string_index_key_t label;
  struct tree  *subtrees;
  struct tree  *sibling;
  tree_depth_t depth; /* depth of this node = the number of edges in the path from root; the depth of root = 0 */
} TREE;

typedef struct q_path {
  unsigned char length;
  string_index_key_t* labels;
} Q_PATH;

typedef struct q_gram {
  unsigned char pattern_id;
  Q_PATH* q_path;
} Q_GRAM;

typedef struct q_profile {
  unsigned int q_patterns_count;
  struct hashtable** q_patterns; /* array of hashtables, its size is q_patterns_count
                            with q_path* keys and unsigned char* values
                            the value pointers will not be shared with other hashtables
                            the value pointers should be freed on every remove
                            and on hashtable_destroy
                            */
} Q_PROFILE;

typedef struct context_entity_occurence {
  unsigned short* path_counts;
  Q_PATH*** pathes; /* 2-dimensional array of q_pathes
                       the first dimension size is
                       1 if custom_params->q_gram_length_mode == 'f'
                       (custom_params->q_gram_length - 1)/2 if custom_params->q_gram_length_mode == 'u'
                       the second dimension sizes are stored in path_counts array */
} CONTEXT_ENTITY_OCCURENCE;

typedef struct context_example {
  unsigned char entity_occurence_count_1;
  CONTEXT_ENTITY_OCCURENCE** entity_occurences_1;
  unsigned char entity_occurence_count_2;
  CONTEXT_ENTITY_OCCURENCE** entity_occurences_2;
} CONTEXT_EXAMPLE;



typedef unsigned short node_count_t;

typedef struct spectrum_tree {
  TREE* root;
  node_count_t node_count;
  tree_depth_t max_depth;
  TREE** nodes; /* array of nodes of this tree in postorder */
  Q_PROFILE* q_profile;
  double norm_square; /* = kernel(this, this) or UNINITIALIZED_KERNEL_VALUE */
  int norm_square_spectrum; /* = kernel(this, this) or UNINITIALIZED_KERNEL_VALUE */
  struct hashtable* bag_of_words;
  CONTEXT_EXAMPLE* context_example;
  char* example_id;
} SPECTRUM_TREE;


int count_common_q_grams(SPECTRUM_TREE* spectrum_tree1, SPECTRUM_TREE* spectrum_tree2, CUSTOM_PARM* custom_params);
char* read_example_id(char* src, int* pos, CUSTOM_PARM* custom_params);
int path_count(CUSTOM_PARM* custom_params);
int path_index(int q_path_length, CUSTOM_PARM* custom_params);

void consumespaces(char* src, int* pos);
string_index_key_t parselabel(char* src, int* pos, unsigned char is_terminal, STRING_INDEX *string_index);
TREE *parsetree(char *src, int *pos, node_count_t *node_count, tree_depth_t depth, tree_depth_t* max_depth, STRING_INDEX *string_index);
void free_tree(TREE* t);
void print_tree(TREE* t, STRING_INDEX* string_index);
void add_tree_nodes(TREE* t, TREE** nodes, node_count_t* node_index);

SPECTRUM_TREE * create_spectrum_tree(char *source, STRING_INDEX* string_index, CUSTOM_PARM* custom_params);
void free_spectrum_tree(SPECTRUM_TREE *spectrum_tree, unsigned char q);

CONTEXT_EXAMPLE* read_context_example(char* src, int* pos, CUSTOM_PARM* custom_params);
Q_PROFILE* read_q_profile(char* src, unsigned char q_gram_length, unsigned int q_patterns_count);

Q_PROFILE * create_q_profile(SPECTRUM_TREE* spectrum_tree, unsigned char q, STRING_INDEX* string_index, unsigned int q_patterns_count);
void free_q_profile(Q_PROFILE* q_profile, unsigned char q);
int norm_square_spectrum(SPECTRUM_TREE* spectrum_tree, CUSTOM_PARM* custom_params);
int score_context_paths(SPECTRUM_TREE* spectrum_tree1, SPECTRUM_TREE* spectrum_tree2, CUSTOM_PARM* custom_params);



Q_PATH *concat_q_path(Q_PATH* q_path, string_index_key_t next_element, unsigned char q_gram_length);
void print_q_path(Q_PATH* q_path);
Q_PATH *q_path_dup(Q_PATH* q_path, unsigned char q_gram_length);

Q_PATH *create_q_path(unsigned char q_path_capacity);
void free_q_path(Q_PATH* q_path);
void free_q_path_void(void* q_path);
Q_PATH * q_subpath(unsigned char start, unsigned char end, Q_PATH* q_path, unsigned char q_gram_length);
void print_q_gram(Q_GRAM* q_gram, STRING_INDEX* string_index);


Q_GRAM *create_q_gram(unsigned char pattern_id, Q_PATH* q_path);
void free_q_gram(Q_GRAM* q_gram);
void free_q_gram_void(void* q_gram);

unsigned int q_gram_hash(void* q_gram);
unsigned int path_hash(void* path);
int q_gram_equal(void* q_gram1, void* q_gram2);
int path_equal(void* path1, void* path2);


struct hashtable* create_bag_of_words(TREE** nodes, node_count_t node_count, char bow_only_terminals, STRING_INDEX* string_index);
int index_equal(void* k1, void* k2);
unsigned int index_hash(void* index);
int count_common_words(SPECTRUM_TREE* spectrum_tree1, SPECTRUM_TREE* spectrum_tree2, char bow_only_terminals);
int norm_square_bow(SPECTRUM_TREE* spectrum_tree);

unsigned char compute_first(char operation);
unsigned char compute_second(char operation);


int levenshtein_distance(string_index_key_t* s, int m, string_index_key_t* t, int n);
int levenshtein_similarity(string_index_key_t* s, string_index_key_t* t, int length);
string_index_key_t matrix_get(int i, int j, int i_len, string_index_key_t* d);
void matrix_set(int i, int j, string_index_key_t val, int i_len, string_index_key_t* d);
string_index_key_t min(string_index_key_t a, string_index_key_t b, string_index_key_t c);


int i_levenshtein_distance(string_index_key_t* s, int m, string_index_key_t* t, int n, CUSTOM_PARM* custom_params);
int i_ins_del_cost(string_index_key_t node_label, CUSTOM_PARM* custom_params);
int i_replace_cost(
    string_index_key_t dependency_type1,
    string_index_key_t node_label1,
    string_index_key_t dependency_type2,
    string_index_key_t node_label2,
    CUSTOM_PARM* custom_params);
int i_levenshtein_similarity(string_index_key_t* s, string_index_key_t* t, int length, CUSTOM_PARM* custom_params);

#endif




