/* kernel.h
 *
 * Written by Peter Palaga, 2009-04-13
 *
 */


#include "tree.h"

unsigned char initialized = 0;

CUSTOM_PARM* custom_params = NULL;

void init_kernel(KERNEL_PARM *kernel_parm);
double custom_kernel(KERNEL_PARM *kernel_parm, SVECTOR *a, SVECTOR *b);
CUSTOM_PARM* parse_custom_parm(char* custom_parm);
void init_svector(SVECTOR *a);
void print_custom_parm_script();

/* KERNEL_PARM is defined in svm_common.h The field 'custom' is reserved for */
/* parameters of the user defined kernel. You can also access and use */
/* the parameters of the other kernels. Just replace the line
 return((double)(1.0));
 with your own kernel. */

/* Example: The following computes the polynomial kernel. sprod_ss
 computes the inner product between two sparse vectors.

 return((CFLOAT)pow(kernel_parm->coef_lin*sprod_ss(a->words,b->words)
 +kernel_parm->coef_const,(double)kernel_parm->poly_degree));
 */

/* If you are implementing a kernel that is not based on a
 feature/value representation, you might want to make use of the
 field "userdefined" in SVECTOR. By default, this field will contain
 whatever string you put behind a # sign in the example file. So, if
 a line in your training file looks like

 -1 1:3 5:6 #abcdefg

 then the SVECTOR field "words" will contain the vector 1:3 5:6, and
 "userdefined" will contain the string "abcdefg". */

double
custom_kernel(KERNEL_PARM *kernel_parm, SVECTOR *a, SVECTOR *b)
/* plug in you favorite kernel */
{
  int cnt;
  double result;

  if (!initialized)
    {
      init_kernel(kernel_parm);
    }

  init_svector(a);
  init_svector(b);

  /*return (double)1.0;*/

  /*printf("---\n%s\n-\n%s\n", a->userdefined, b->userdefined);*/

  if (a->parse == NULL || b->parse == NULL)
    {
      /* first trivial case */
      result = 0.0;
    }
  else if (a == b)
    {
      /* second (semi-)trivial case */
      if (custom_params->custom_kernel_normalize)
        {
          result = 1.0;
        }
      else
        {
          result = norm_square_spectrum(a->parse, custom_params);
        }
    }
  else
    {
      /* non trivial case - count the common substructures */
      cnt = count_common_q_grams(a->parse, b->parse, custom_params);
      /*printf("common substr = %d\n", cnt);*/
      if (custom_params->custom_kernel_normalize && cnt != 0)
        {
          /*
           printf("cnt = %d, ns(a) = %d, ns(b) = %d, sqrt(ns(a)*ns(b)) = %f\n",
           cnt,
           norm_square(a->parse, q_gram_length),
           norm_square(b->parse, q_gram_length),
           sqrt(norm_square(a->parse, q_gram_length) * norm_square(b->parse, q_gram_length))
           );
           */
          result = ((double) cnt) / ((double) sqrt(norm_square_spectrum(a->parse, custom_params) * norm_square_spectrum(b->parse, custom_params)));
        }
      else
        {
          result = (double) cnt;
        }
    }
  /*printf("kernel = %f\n", result);*/
  return result;
}

void
init_kernel(KERNEL_PARM *kernel_parm)
{

  if (strlen(kernel_parm->custom) == 0)
    {
      perror("-u option must be set for spectrum kernel.");
      exit(1);
    }
  custom_params = parse_custom_parm(kernel_parm->custom);
  if ( verbosity > 1 )
  {
	printf("update %s set kernel_script = '", kernel_parm->dbname);
    print_custom_parm_script();
    printf("' where %sid = currval('%s_%sid_seq');\n", kernel_parm->dbname, kernel_parm->dbname, kernel_parm->dbname);
  }
  initialized = 1;
}

void
init_svector(SVECTOR *a)
{
  if (a->parse == NULL && a->userdefined != NULL && a->userdefined[0] != '\0')
    {
      a->parse = create_spectrum_tree(a->userdefined, custom_params);
    }
}

void
print_custom_parm_script()
{
  if (custom_params->custom_kernel_normalize)
    {
      printf("N(");
    }

  switch (custom_params->custom_kernel_input_format)
    {
  case 'b':
    printf("k-band(%uto%u k%u match=%c)",
        custom_params->q_gram_length_min, custom_params->q_gram_length_max,
        custom_params->k_band, custom_params->spectrum_tree_kernel_match_mode);
    break;
  default:
    perror("Unexpected custom_kernel_input_format value.");
    exit(-1);
    }
  if (custom_params->custom_kernel_normalize)
    {
      printf(")");
    }
}

CUSTOM_PARM*
parse_custom_parm(char* custom_parm)
{
  char* custom_parm_cpy;
  char* key;
  char* val;
  CUSTOM_PARM* result;

  result = (CUSTOM_PARM*) my_malloc(sizeof(CUSTOM_PARM));
  result->custom_kernel_input_format = 'b';
  result->custom_kernel_normalize = 0;

  result->q_gram_length_min = 1;
  result->q_gram_length_max = 1;
  result->k_band = 0;

  result->spectrum_tree_kernel_match_mode = 'e';

  result->token_label_score = DEFAULT_TOKEN_LABEL_SCORE;
  result->token_label_mismatch_score = DEFAULT_TOKEN_LABEL_MISMATCH_SCORE;
  result->entity_label_score = DEFAULT_TOKEN_LABEL_SCORE;
  result->entity_label_mismatch_score = DEFAULT_ENTITY_LABEL_MISMATCH_SCORE;
  result->dependency_type_score = DEFAULT_DEPENDENCY_TYPE_SCORE;
  result->dependency_type_mismatch_score = DEFAULT_DEPENDENCY_TYPE_MISMATCH_SCORE;

  custom_parm_cpy = (char*) my_malloc((strlen(custom_parm) + 1) * sizeof(char));
  strcpy(custom_parm_cpy, custom_parm);

  key = strtok(custom_parm_cpy, "=,");

  while (key != NULL)
    {
      val = strtok(NULL,"=,");
      assert(val != NULL);
      if (strcmp(key, "f") == 0)
        {
          result->custom_kernel_input_format = val[0];
        }
      else if (strcmp(key, "N") == 0)
        {
          result->custom_kernel_normalize = atoi(val);
        }
      else if (strcmp(key, "a") == 0)
        {
          result->q_gram_length_max = atoi(val);
        }
      else if (strcmp(key, "i") == 0)
        {
          result->q_gram_length_min = atoi(val);
        }
      else if (strcmp(key, "k") == 0)
        {
          result->k_band = atoi(val);
        }
      else if (strcmp(key, "t") == 0)
        {
          result->spectrum_tree_kernel_match_mode = val[0];
        }
      else if (strcmp(key, "l") == 0)
        {
          result->token_label_score = atoi(val);
        }
      else if (strcmp(key, "L") == 0)
        {
          result->token_label_mismatch_score = atoi(val);
        }
      else if (strcmp(key, "e") == 0)
        {
          result->entity_label_score = atoi(val);
        }
      else if (strcmp(key, "E") == 0)
        {
          result->entity_label_mismatch_score = atoi(val);
        }
      else if (strcmp(key, "d") == 0)
        {
          result->dependency_type_score = atoi(val);
        }
      else if (strcmp(key, "D") == 0)
        {
          result->dependency_type_mismatch_score = atoi(val);
        }
      else
        {
          perror("Unexpected custom kernel option.");
          exit(-1);
        }

      key = strtok(NULL,"=,");
    }

  free(custom_parm_cpy);

  return result;
}
