/* tree.c
 *
 *  Written by Mark Johnson, 27th May 1997
 *         and Peter Palaga 2009-04-13
 */

#include <stdlib.h>
#include <assert.h>
#include <ctype.h>
#include "tree.h"
#include "svm_common.h"
#include "hashtable/hashtable_itr.h"

void
consumespaces(char* src, int* pos)
{
  int c;
  /* skip spaces */
  while (1)
    {
      c = src[*pos];
      if (!isspace(c))
        {
          break;
        }
      *pos = (*pos) + 1;
    }
}

string_index_key_t
parselabel(char* src, int* pos, unsigned char is_terminal,
    STRING_INDEX *string_index)
{
  int c, start_pos, len;

  start_pos = *pos;

  while (1)
    {
      c = src[*pos];
      if (c == '\0' || isspace(c) || c == '(' || c == ')')
        {
          break;
        }
      *pos = (*pos) + 1;
    };

  len = (*pos) - start_pos;
  assert(len> 0);

  consumespaces(src, pos);

  return index_for_value(src + start_pos, len, is_terminal, string_index);
}

Q_PATH *
concat_q_path(Q_PATH* q_path, string_index_key_t next_element,
    unsigned char q_gram_length)
{
  int i;
  Q_PATH *result;
  if (q_path == NULL)
    {
      result = create_q_path(q_gram_length);
      result->labels[0] = next_element;
      result->length = 1;
      return result;
    }
  else
    {
      if (q_path->length >= q_gram_length)
        {
          result = create_q_path(q_gram_length + 1);
        }
      else
        {
          result = create_q_path(q_path->length);
        }
      for (i = 0; i < q_path->length; ++i)
        {
          result->labels[i] = q_path->labels[i];
        }
      result->labels[i++] = next_element;
      result->length = i;
      return result;
    }
}

Q_PATH *
q_subpath(unsigned char start, unsigned char end, Q_PATH* q_path,
    unsigned char q_gram_length)
{
  Q_PATH* result;
  int i, n;
  assert(q_path != NULL);
  assert(start < end);
  /*assert(end <= q_path->length || end <= q_gram_length TODO: check if this is not too liberal );*/
  n = end - start;

  result = create_q_path(n > q_gram_length ? n : q_gram_length);
  result->length = n;
  for (i = 0; i < n; ++i)
    {
      result->labels[i] = q_path->labels[start + i];
    }
  return result;
}

void
print_q_path(Q_PATH* q_path)
{
  int i;
  if (q_path != NULL)
    {
      printf("(");
      for (i = 0; i < q_path->length; ++i)
        {
          if (i > 0) {
            printf(" ");
          }
          printf("%u", q_path->labels[i]);
        }
      printf(")");
    }
}
Q_PATH *
q_path_dup(Q_PATH* q_path, unsigned char q_gram_length)
{
  Q_PATH* result;
  int i;
  assert(q_path != NULL);
  result = create_q_path(q_path->length > q_gram_length ? q_path->length
      : q_gram_length);
  result->length = q_path->length;
  for (i = 0; i < q_path->length; ++i)
    {
      result->labels[i] = q_path->labels[i];
    }
  return result;
}

Q_PATH *
create_q_path(unsigned char q_path_capacity)
{
  Q_PATH *result;
  int i;
  result = (Q_PATH*) my_malloc(sizeof(Q_PATH));
  result->length = 0;
  result->labels = (string_index_key_t*) my_malloc(q_path_capacity
      * sizeof(string_index_key_t));
  for (i = 0; i < q_path_capacity; ++i)
    {
      result->labels[i] = ZERO_INDEX;
    }
  return result;
}

void
free_q_path(Q_PATH* q_path)
{
  if (q_path != NULL)
    {
      if (q_path->labels != NULL)
        {
          free(q_path->labels);
        }
      free(q_path);
    }
}

void
free_q_path_void(void* q_path)
{
  free_q_path((Q_PATH*) q_path);
}


SPECTRUM_TREE*
create_spectrum_tree(char *source, CUSTOM_PARM* custom_params)
{
  SPECTRUM_TREE *result;
  int pos = 0;
  result = NULL;

  if (source && source[0] != '\0')
    {
      /*printf("%s\n", source);*/
      result = (SPECTRUM_TREE *) my_malloc(sizeof(SPECTRUM_TREE));
      result->example_id = NULL;
      result->norm_square_spectrum = UNINITIALIZED_KERNEL_VALUE;

      switch (custom_params->custom_kernel_input_format)
        {
      case 'b':
        result->path_profile = read_path_profile(source, &pos, custom_params);
        break;
      default:
        perror("Unexpected custom_kernel_input_format value.");
        exit(-1);
        }
    }

  return result;
}

void
print_escaped_element(char* str)
{
  int i;
  i = 0;
  while (str[i] != '\0')
    {
      if (str[i] == ',')
        {
          putc('_', stdout);
        }
      else
        {
          putc(str[i], stdout);
        }
      i++;
    }
}

int
norm_square_spectrum(SPECTRUM_TREE* spectrum_tree, CUSTOM_PARM* custom_params)
{
  assert(spectrum_tree != NULL);
  if (spectrum_tree->norm_square_spectrum == UNINITIALIZED_KERNEL_VALUE)
    {
      spectrum_tree->norm_square_spectrum = count_common_q_grams(spectrum_tree,spectrum_tree, custom_params);
    }
  return spectrum_tree->norm_square_spectrum;
}

int score_pathes(Q_PATH* path1, Q_PATH* path2, CUSTOM_PARM* custom_params)
{
  int result = 0;

/*
  printf("score_pathes ");
  print_q_path(path1);
  printf(" ");
  print_q_path(path2);
*/

  switch (custom_params->spectrum_tree_kernel_match_mode)
    {
  case 'e':
    /* exact */
    assert(path1->length == path2->length);

    if (0 == memcmp(path1->labels, path2->labels, path1->length
        * sizeof(string_index_key_t)))
      {
        result = 1;
      }
    else
      {
        result = 0;
      }
    /*printf(" = %d\n", result);*/

    return result;
  case 't':
    /* tolerant */
    /*fflush(stdout);
    result = score_tolerant(path1, path2, custom_params);
    printf(" = %d\n", result);
    return result;*/
    return score_tolerant(path1, path2, custom_params);
  case 'l':
    /* levenshtein */
    assert(path1->length == path2->length);
    return levenshtein_similarity(path1->labels, path2->labels, path1->length);
  case 'd':
    /* levenshtein */
    assert(path1->length == path2->length);
    return levenshtein_distance(path1->labels, path1->length, path2->labels, path2->length);
  case 'i':
    /* intelligent levenshtein */
    assert(path1->length == path2->length);
    return i_levenshtein_similarity(path1->labels, path2->labels, path1->length, custom_params);
  default:
    perror("Unexpected spectrum_tree_kernel_match_mode.");
    exit(-1);
    }

}

int score_tolerant_values(string_index_key_t val1, string_index_key_t val2, int i, CUSTOM_PARM* custom_params) {
  int eq = 0;
  int is_entity = 0;
  eq = (val1 == val2);
  if (i % 2 == 0) {
    /* node label */
    is_entity = val1 <= TOKEN_LABEL_ENTITY_12 && val1 >= TOKEN_LABEL_ENTITY_1;
    if (eq) {
      if (is_entity) {
        return custom_params->entity_label_score;
      }
      else {
        return custom_params->token_label_score;
      }
    }
    else {
      if (is_entity) {
        return custom_params->entity_label_mismatch_score;
      }
      else {
        return custom_params->token_label_mismatch_score;
      }
    }
  }
  else {
    /* dependency type */
    if (eq) {
      return custom_params->dependency_type_score;
    }
    else {
      return custom_params->dependency_type_mismatch_score;
    }
  }

}
int score_tolerant(Q_PATH* path1, Q_PATH* path2, CUSTOM_PARM* custom_params) {
  int result = 0;
  /*int result_inv = 0;*/
  int i = 0;
  int score = 0;
  /*int score_inv = 0;*/
  assert(path1->length == path2->length);

  for (i = 0; i < path1->length; ++i) {
    score = score_tolerant_values(path1->labels[i], path2->labels[i], i, custom_params);
    if (score == CANCELATION_SCORE) {
      result = 0;
      break;
    }
    else {
      result += score;
    }
  }

  return result;

/*
  for (i = 0; i < path1->length; ++i) {
    score_inv = score_tolerant_values(path1->labels[i], path2->labels[path2->length -1 -i], i, custom_params);
    if (score_inv == CANCELATION_SCORE) {
      result_inv = 0;
      break;
    }
    else {
      result_inv += score_inv;
    }
  }

  return result > result_inv ? result : result_inv;
*/
}

int
count_common_q_grams(SPECTRUM_TREE* spectrum_tree1,
    SPECTRUM_TREE* spectrum_tree2, CUSTOM_PARM* custom_params)
{
  int result = 0;
  int i = 0;
  int q_paths_count;

  int cnt1;
  int cnt2;
  int i1;
  int i2;

  PATH_PROFILE* pp1;
  PATH_PROFILE* pp2;
  Q_PATH** paths1;
  Q_PATH** paths2;

  pp1 = spectrum_tree1->path_profile;
  pp2 = spectrum_tree2->path_profile;

  if (pp1 != NULL && pp2 != NULL) {
    q_paths_count = path_count(custom_params);
    /*printf("q_paths_count = %u\n", q_paths_count);*/
    for (i = 0; i < q_paths_count; ++i) {
      cnt1 = pp1->path_counts[i];
      cnt2 = pp2->path_counts[i];
      /*printf("cnt1 = %u\n", cnt1);
      printf("cnt2 = %u\n", cnt2);*/
      paths1 = pp1->pathes[i];
      paths2 = pp2->pathes[i];

      for (i1 = 0; i1 < cnt1; ++i1) {
        for (i2 = 0; i2 < cnt2; ++i2) {
          result += score_pathes(paths1[i1], paths2[i2], custom_params);
        }
      }
    }
  }

  return result;

}

unsigned int
read_int(char* src, int* pos)
{
  char c;
  char* tmp;
  int i;
  int start;
  int len;
  int result;
  start = *pos;

  /* at least one digit */
  c = src[*pos];
  assert(isdigit(c));
  *pos = *pos + 1;

  while (1)
    {
      c = src[*pos];
      if (!isdigit(c))
        {
          break;
        }
      *pos = *pos + 1;
    }

  len = *pos - start;
  tmp = (char*) my_malloc((len + 1) * sizeof(char));
  for (i = 0; i < len; ++i)
    {
      tmp[i] = src[start + i];
    }
  tmp[i] = '\0';

  result = atoi(tmp);
  free(tmp);
  return result;
}

Q_PATH*
read_q_path(char* src, int* pos, CUSTOM_PARM* custom_params)
{
  int nr;
  Q_PATH* result;
  char c;

  consumespaces(src, pos);

  c = src[*pos];
  if (c == '\0')
    {
      /* end of input */
      return NULL;
    }
  else if (c != '(')
    {
      /* not a q_path */
      return NULL;
    }

  assert(c == '(');
  *pos = *pos + 1;

  result = (Q_PATH*) my_malloc(sizeof(Q_PATH));

  result->length = 0;
  result->labels = NULL;

  while (1)
    {
      consumespaces(src, pos);
      c = src[*pos];
      if (c == ')')
        {
          /* end of q_gram */
          break;
        }

      nr = read_int(src, pos);
      assert(nr <= string_index_max_capacity);
      result->length++;

      result->labels = (string_index_key_t*) my_realloc(result->labels,
          result->length * sizeof(string_index_key_t));
      result->labels[result->length - 1] = nr;

    }
  assert(c == ')');
  *pos = *pos + 1;

  return result;
}
int path_count(CUSTOM_PARM* custom_params)
{
  return custom_params->q_gram_length_max - custom_params->q_gram_length_min + 1;
}
int path_index(int q_path_length, CUSTOM_PARM* custom_params)
{
  return ((q_path_length + 1)/2) - custom_params->q_gram_length_min;
}

PATH_PROFILE* read_path_profile(char* src, int* pos, CUSTOM_PARM* custom_params)
{
  char c;
  int cnt = 0;
  int i;
  PATH_PROFILE* result;
  Q_PATH* q_path;

  consumespaces(src, pos);
  c = src[*pos];

  if (c == '\0')
    {
      /* end of input */
      return NULL;
    }
  else if (c != '(')
    {
      /* not a PATH_PROFILE */
      return NULL;
    }

  cnt = path_count(custom_params);

  result = (PATH_PROFILE*) my_malloc(sizeof(PATH_PROFILE));
  result->path_counts = (unsigned short*) my_malloc(cnt * sizeof(unsigned short));
  result->pathes = (Q_PATH***) my_malloc(cnt * sizeof(Q_PATH**));
  for (i = 0; i < cnt; ++i)
    {
      result->path_counts[i] = 0;
      result->pathes[i] = NULL;
    }

  while (1)
    {
      q_path = read_q_path(src, pos, custom_params);
      if (q_path == NULL)
        {
          break;
        }

      i = path_index(q_path->length, custom_params);

      assert(i < cnt);
      result->path_counts[i]++;
      result->pathes[i] = (Q_PATH**) my_realloc(result->pathes[i], result->path_counts[i] * sizeof(Q_PATH*));
      result->pathes[i][result->path_counts[i] - 1] = q_path;

    }

  return result;
}

char*
read_example_id(char* src, int* pos, CUSTOM_PARM* custom_params)
{
  char c;
  char* result;
  int i;
  switch (custom_params->custom_kernel_input_format)
    {
  case 't':
  case 'q':
    return NULL;
    break;
  case 'c':

    c = src[*pos];
    assert(c == '#');
    *pos = *pos + 1;

    i = *pos;

    while (src[i] != '\0')
      {
        i++;
      }

    result = (char*) my_malloc((i - *pos + 1) * sizeof(char));
    i = *pos;
    while (src[*pos] != '\0')
      {
        result[*pos - i] = src[*pos];
        *pos = *pos + 1;
      }
    result[*pos - i] = '\0';

    return result;

    break;
  default:
    perror("Unexpected custom_kernel_input_format value.");
    exit(-1);
    }
}


void free_spectrum_tree(SPECTRUM_TREE *spectrum_tree, CUSTOM_PARM* custom_params)
{
  if (spectrum_tree != NULL)
    {
      /*TODO: free other members*/
      free(spectrum_tree);
    }
}

string_index_key_t matrix_get(int i, int j, int i_len, string_index_key_t* d) {
  return d[(j*i_len)+i];
}
void matrix_set(int i, int j, string_index_key_t val, int i_len, string_index_key_t* d) {
  d[(j*i_len)+i] = val;
}

int levenshtein_similarity(string_index_key_t* s, string_index_key_t* t, int length) {
  return length - levenshtein_distance(s, length, t, length);
}
int i_levenshtein_similarity(string_index_key_t* s, string_index_key_t* t, int length, CUSTOM_PARM* custom_params) {
  return (((length -1)/2)* (I_DEPENDENCY_MISMATCH_COST + I_ENTITY_MISMATCH_COST)) - i_levenshtein_distance(s, length, t, length, custom_params);
}

int i_replace_cost(
    string_index_key_t dependency_type1,
    string_index_key_t node_label1,
    string_index_key_t dependency_type2,
    string_index_key_t node_label2,
    CUSTOM_PARM* custom_params) {
  int result = 0;
  if (dependency_type1 != dependency_type2) {
    result += I_DEPENDENCY_MISMATCH_COST;
  }
  if (node_label1 != node_label2) {
    if (node_label1 == OTHER_ENTITY_LABEL
        || node_label2 == OTHER_ENTITY_LABEL
        || node_label1 == THIS_OR_OTHER_ENTITY_LABEL
        || node_label2 == THIS_OR_OTHER_ENTITY_LABEL
    ) {
      /* this is very expensive */
      result += I_ENTITY_MISMATCH_COST;
    }
    else {
      result += I_NODE_LABEL_MISMATCH_COST;
    }
  }
  return result;
}

int i_ins_del_cost(
    string_index_key_t node_label,
    CUSTOM_PARM* custom_params) {
  int result = I_DEPENDENCY_MISMATCH_COST;
  if (node_label == OTHER_ENTITY_LABEL
      || node_label == THIS_OR_OTHER_ENTITY_LABEL
  ) {
    /* this is very expensive */
    result += I_ENTITY_MISMATCH_COST;
  }
  else {
    result += I_NODE_LABEL_MISMATCH_COST;
  }
  return result;
}

int i_levenshtein_distance(string_index_key_t* s, int m, string_index_key_t* t, int n, CUSTOM_PARM* custom_params) {
  string_index_key_t** d;
  int i,j, m_plus_1, n_plus_1, i_type, j_type;
  m_plus_1 = ((m-1)/2) +1;
  n_plus_1 = ((n-1)/2) +1;
  /*d is a table with m+1 rows and n+1 columns*/
  d = (string_index_key_t**) my_malloc(m_plus_1 * sizeof(string_index_key_t*));


  for (i = 0; i < m_plus_1; ++i) {
    d[i] = (string_index_key_t*) my_malloc(n_plus_1 * sizeof(string_index_key_t));
    d[i][0] = i;
  }

  for (j = 0; j < n_plus_1; ++j) {
    d[0][j] = j;
  }

  for (i = 1; i < m_plus_1; ++i) {
    for (j = 1; j < n_plus_1; ++j) {

      i_type = ((i-1)*2) + 1;
      j_type = ((j-1)*2) + 1;

      d[i][j] =
          min(
              d[i-1][j] + i_ins_del_cost(s[i_type+1], custom_params), /*deletion*/
              d[i][j-1] + i_ins_del_cost(s[j_type+1], custom_params), /*insertion*/
              d[i-1][j-1] + i_replace_cost(s[i_type], s[i_type+1], t[j_type], t[j_type+1], custom_params)
          );
    }
  }

  i = d[m_plus_1 - 1][n_plus_1 - 1];
  for (i = 0; i < m_plus_1; ++i) {
    free(d[i]);
  }
  free(d);
  return i;
}


int levenshtein_distance(string_index_key_t* s, int m, string_index_key_t* t, int n) {
  string_index_key_t* d;
  string_index_key_t cost;
  int i,j, m_plus_1;
  m_plus_1 = m+1;
  /*d is a table with m+1 rows and n+1 columns*/
  d = (string_index_key_t*) my_malloc((m_plus_1) * (n +1) * sizeof(string_index_key_t));

  for (i = 0; i <= m; ++i) {
    matrix_set(i, 0, i, m_plus_1, d);
  }

  for (j = 0; j <= n; ++j) {
    matrix_set(0, j, j, m_plus_1, d);
  }

  for (i = 1; i <= m; ++i) {
    for (j = 1; j <= n; ++j) {
      if (s[i -1] == t[j -1]) {
        cost = 0;
      }
      else {
        cost = 1;
      }

      matrix_set(i, j,
          min(
              matrix_get(i-1, j, m_plus_1, d) + 1,
              matrix_get(i, j-1, m_plus_1, d) +1,
              matrix_get(i-1, j-1, m_plus_1, d) +cost
          ),
          m_plus_1, d);
    }
  }

  i = matrix_get(m, n, m_plus_1, d);
  free(d);
  return i;
}

string_index_key_t min(string_index_key_t a, string_index_key_t b, string_index_key_t c) {
  if (a < b && a < c) {
    return a;
  }
  else if (b < a && b < c) {
    return b;
  }
  else {
    return c;
  }
}

