/* tree.h
 *
 * Written by Mark Johnson, 27th May 1997
 *         and Peter Palaga 2009-04-13
 */

#ifndef TREE_H
#define TREE_H
#define NDEBUG

#define tree_depth_max UCHAR_MAX
#define tree_depth_min 1
#define node_count_max USHRT_MAX
#define UNINITIALIZED_KERNEL_VALUE (-1)
#define UNINITIALIZED_KERNEL_VALUE_DBL (-1.0)

#define DEFAULT_DEPENDENCY_TYPE_SCORE 1
#define DEFAULT_DEPENDENCY_TYPE_MISMATCH_SCORE 0
#define DEFAULT_ENTITY_LABEL_MISMATCH_SCORE 0
#define DEFAULT_TOKEN_LABEL_MISMATCH_SCORE 0
#define DEFAULT_TOKEN_LABEL_SCORE 2
#define CANCELATION_SCORE -1

#define TOKEN_LABEL_ENTITY_1 0
#define TOKEN_LABEL_ENTITY_2 1
#define TOKEN_LABEL_ENTITY_12 2

#define THIS_ENTITY_LABEL 0
#define OTHER_ENTITY_LABEL 1
#define THIS_OR_OTHER_ENTITY_LABEL 2

#define I_DEPENDENCY_MISMATCH_COST 2
#define I_ENTITY_MISMATCH_COST 3
#define I_NODE_LABEL_MISMATCH_COST 1


#include "string_index.h"
#include "hashtable/hashtable.h"

typedef unsigned char tree_depth_t;

typedef struct custom_parm {
  unsigned char q_gram_length_max;
  unsigned char q_gram_length_min;
  unsigned char k_band;

  unsigned char custom_kernel_normalize;
  char custom_kernel_input_format;
  /* 'b' ... k-band path
   */
  char spectrum_tree_kernel_match_mode;
  /* 'e' ... exact
   * 't' ... tolerant - dependecy types must match, nodel labels need not
   * 'l' ... (naive) levenshtein similarity
   * 'i' ... (intelligent) levenshtein similarity
   */
  unsigned char entity_label_score;
  signed char entity_label_mismatch_score;
  unsigned char token_label_score;
  signed char token_label_mismatch_score;
  unsigned char dependency_type_score;
  signed char dependency_type_mismatch_score;
} CUSTOM_PARM;

typedef struct q_path {
  unsigned char length;
  string_index_key_t* labels;
} Q_PATH;

typedef struct path_profile {
  unsigned short* path_counts;
  Q_PATH*** pathes; /* 2-dimensional array of q_pathes
                       the first dimension size is
                       1 if custom_params->q_gram_length_mode == 'f'
                       (custom_params->q_gram_length - 1)/2 if custom_params->q_gram_length_mode == 'u'
                       the second dimension sizes are stored in path_counts array */
} PATH_PROFILE;

typedef unsigned short node_count_t;

typedef struct spectrum_tree {
  int norm_square_spectrum; /* = kernel(this, this) or UNINITIALIZED_KERNEL_VALUE */
  char* example_id;
  PATH_PROFILE* path_profile;
} SPECTRUM_TREE;

PATH_PROFILE* read_path_profile(char* src, int* pos, CUSTOM_PARM* custom_params);

int count_common_q_grams(SPECTRUM_TREE* spectrum_tree1, SPECTRUM_TREE* spectrum_tree2, CUSTOM_PARM* custom_params);
char* read_example_id(char* src, int* pos, CUSTOM_PARM* custom_params);
int path_count(CUSTOM_PARM* custom_params);
int path_index(int q_path_length, CUSTOM_PARM* custom_params);

void consumespaces(char* src, int* pos);
string_index_key_t parselabel(char* src, int* pos, unsigned char is_terminal, STRING_INDEX *string_index);

SPECTRUM_TREE * create_spectrum_tree(char *source, CUSTOM_PARM* custom_params);
void free_spectrum_tree(SPECTRUM_TREE *spectrum_tree, CUSTOM_PARM* custom_params);

int norm_square_spectrum(SPECTRUM_TREE* spectrum_tree, CUSTOM_PARM* custom_params);
int score_context_paths(SPECTRUM_TREE* spectrum_tree1, SPECTRUM_TREE* spectrum_tree2, CUSTOM_PARM* custom_params);

int score_pathes(Q_PATH* path1, Q_PATH* path2, CUSTOM_PARM* custom_params);
int score_tolerant(Q_PATH* path1, Q_PATH* path2, CUSTOM_PARM* custom_params);
int score_tolerant_values(string_index_key_t val1, string_index_key_t val2, int i, CUSTOM_PARM* custom_params);


Q_PATH *concat_q_path(Q_PATH* q_path, string_index_key_t next_element, unsigned char q_gram_length);
void print_q_path(Q_PATH* q_path);
Q_PATH *q_path_dup(Q_PATH* q_path, unsigned char q_gram_length);

Q_PATH *create_q_path(unsigned char q_path_capacity);
void free_q_path(Q_PATH* q_path);
void free_q_path_void(void* q_path);
Q_PATH * q_subpath(unsigned char start, unsigned char end, Q_PATH* q_path, unsigned char q_gram_length);


unsigned int q_gram_hash(void* q_gram);
unsigned int path_hash(void* path);
int q_gram_equal(void* q_gram1, void* q_gram2);
int path_equal(void* path1, void* path2);


int levenshtein_distance(string_index_key_t* s, int m, string_index_key_t* t, int n);
int levenshtein_similarity(string_index_key_t* s, string_index_key_t* t, int length);
string_index_key_t matrix_get(int i, int j, int i_len, string_index_key_t* d);
void matrix_set(int i, int j, string_index_key_t val, int i_len, string_index_key_t* d);
string_index_key_t min(string_index_key_t a, string_index_key_t b, string_index_key_t c);


int i_levenshtein_distance(string_index_key_t* s, int m, string_index_key_t* t, int n, CUSTOM_PARM* custom_params);
int i_ins_del_cost(string_index_key_t node_label, CUSTOM_PARM* custom_params);
int i_replace_cost(
    string_index_key_t dependency_type1,
    string_index_key_t node_label1,
    string_index_key_t dependency_type2,
    string_index_key_t node_label2,
    CUSTOM_PARM* custom_params);
int i_levenshtein_similarity(string_index_key_t* s, string_index_key_t* t, int length, CUSTOM_PARM* custom_params);

#endif




